Package: dbMatrix
Title: Database-Backed Matrix Classes and Operations
Version: 0.1.0
Authors@R: 
  c(person("Edward", "C. Ruiz", , "ecr7407@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-9174-5387")),
    person("Jiaji", "George Chen", email = "jiajic@bu.edu", role = c("aut")),
    person("Ruben", "Dries", email = "rubendries@gmail.com",role = c("aut")))
Description: Provides S4 classes and methods for storing dense and sparse
    matrices in 'DuckDB' databases. The package supports constructing
    database-backed matrices from base R and 'Matrix' objects, extracting
    slices and summaries, performing arithmetic and selected linear algebra
    operations, and materializing results for larger-than-memory workflows.
    It integrates with 'dbProject' to keep database paths, live connections,
    and lazy matrix tables synchronized across interactive analyses.
License: GPL-3
Encoding: UTF-8
URL: https://github.com/dbverse-org/dbmatrix-r,
        https://dbverse-org.github.io/dbmatrix-r/
BugReports: https://github.com/dbverse-org/dbmatrix-r/issues
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: Matrix (>= 1.6-5), MatrixGenerics (>= 1.12.3), methods, DBI,
        dplyr, dbplyr, duckdb (>= 1.4.0), data.table (>= 1.12.2), glue,
        bit64, cli, Rcpp, arrow, nanoarrow, dbProject, rlang
LinkingTo: Rcpp, RcppEigen, RSpectra, nanoarrow
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), irlba, crayon,
        R.utils, checkmate, reticulate, sparseMatrixStats, RSpectra
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-05-14 02:35:49 UTC; ecruiz
Author: Edward C. Ruiz [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9174-5387>),
  Jiaji George Chen [aut],
  Ruben Dries [aut]
Maintainer: Edward C. Ruiz <ecr7407@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-19 08:30:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-19 10:34:55 UTC; unix
Archs: dbMatrix.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.dbMatrix,Matrix)
S3method(as.dbMatrix,default)
S3method(as.dbMatrix,dgCMatrix)
S3method(as.dbMatrix,dgTMatrix)
S3method(as.dbMatrix,matrix)
S3method(as.matrix,dbDenseMatrix)
S3method(as.matrix,dbMatrix)
S3method(as.matrix,dbSparseMatrix)
S3method(compute,dbDenseMatrix)
S3method(compute,dbMatrix)
S3method(compute,dbSparseMatrix)
export("colnames<-.dbMatrix")
export("rownames<-.dbMatrix")
export(as.dbMatrix)
export(colnames.dbMatrix)
export(dbDenseMatrix)
export(dbMatrix)
export(dbMatrix_from_tbl)
export(dbSparseMatrix)
export(db_svd)
export(ncol.dbMatrix)
export(nrow.dbMatrix)
export(rownames.dbMatrix)
export(sim_dbDenseMatrix)
export(sim_dbSparseMatrix)
export(to_named_ijx_tbl)
exportClasses(dbDenseMatrix)
exportClasses(dbMatrix)
exportClasses(dbSparseMatrix)
exportMethods("%in%")
exportMethods("[")
exportMethods("[<-")
exportMethods("dimnames<-")
exportMethods(Arith)
exportMethods(Math)
exportMethods(Ops)
exportMethods(Summary)
exportMethods(colMeans)
exportMethods(colSds)
exportMethods(colSums)
exportMethods(colVars)
exportMethods(dbLoad)
exportMethods(dim)
exportMethods(dimnames)
exportMethods(head)
exportMethods(is.na)
exportMethods(length)
exportMethods(mean)
exportMethods(rowMeans)
exportMethods(rowSds)
exportMethods(rowSums)
exportMethods(rowVars)
exportMethods(t)
exportMethods(tail)
exportMethods(writeMM)
importClassesFrom(DBI,DBIConnection)
importClassesFrom(dbProject,dbData)
importFrom(Matrix,t)
importFrom(Matrix,writeMM)
importFrom(MatrixGenerics,colMeans)
importFrom(MatrixGenerics,colSds)
importFrom(MatrixGenerics,colSums)
importFrom(MatrixGenerics,colVars)
importFrom(MatrixGenerics,rowMeans)
importFrom(MatrixGenerics,rowSds)
importFrom(MatrixGenerics,rowSums)
importFrom(MatrixGenerics,rowVars)
importFrom(Rcpp,sourceCpp)
importFrom(data.table,":=")
importFrom(dbProject,"conn<-")
importFrom(dbProject,.check_con)
importFrom(dbProject,.check_name)
importFrom(dbProject,.check_overwrite)
importFrom(dbProject,.check_tbl)
importFrom(dbProject,conn)
importFrom(dbProject,dbList)
importFrom(dbProject,dbLoad)
importFrom(dbProject,dbReconnect)
importFrom(dbProject,to_view)
importFrom(dbProject,unique_table_name)
importFrom(dplyr,compute)
importFrom(methods,Arith)
importFrom(methods,Math)
importFrom(methods,Ops)
importFrom(methods,Summary)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,validObject)
importFrom(nanoarrow,as_nanoarrow_array_stream)
importFrom(rlang,abort)
importFrom(rlang,inform)
importFrom(rlang,warn)
importFrom(stats,rnorm)
importFrom(stats,setNames)
importFrom(utils,capture.output)
useDynLib(dbMatrix, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# dbMatrix

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/dbMatrix)](https://CRAN.R-project.org/package=dbMatrix)

<!-- badges: end -->

`{dbMatrix}` is a core package in the
[`dbverse`](https://dbverse-org.github.io/dbverse/) library.

The goal of `{dbMatrix}` is to support larger-than-memory matrix
computations.

**Note: Work in progress! Bugs or unexpected behavior are likely to
occur. We welcome user feedback and reporting issues on the [Github
page](https://github.com/dbverse-org/dbmatrix-r/).**

## Installation

You can install the development version of dbMatrix from Github like so:

``` r
# install.packages("pak", repos = sprintf("https://r-lib.github.io/p/pak/stable/%s/%s/%s", .Platform$pkgType, R.Version()$os, R.Version()$arch))
pak::pak("dbverse-org/dbmatrix-r")
```

# Package-wide global variables to avoid R CMD check NOTEs
# These are used in NSE contexts (dplyr, dbplyr)

utils::globalVariables(c(
  # Column names used in dbplyr/dplyr
  "i", "j", "x", "n",
  "i_orig", "j_orig",
  "new_i", "new_j", "new_value",
  "idx", ".I",
  
  # Summary statistics column names
  "sum_x", "sum_x2", "mean_x", "var_x", "sd_x",
  "scale_val",
  
  # Matrix names file columns
  "..mtx_colname_col_idx", "..mtx_rowname_col_idx",
  
  # Coercion variables
  "x.i", "x.x", "x.y",
  
  # Temporary lookup table columns
  ".col_name_tmp", ".row_name_tmp"
))

#' Perform Streaming SVD on a dbMatrix
#' @param dbm A dbSparseMatrix object
#' @param k Number of singular values to compute
#' @param center Logical, center rows (default TRUE)
#' @param scale Logical, scale rows (default FALSE)
#' @param center_rows Logical, center rows vs columns (default TRUE for standard PCA)
#' @param memory_limit Bytes for Fast Path. Default 8 GB.
#' @param return_format "svd" (d, u, v) or "pca" (eigenvalues, loadings, coords)
#' @return List with SVD or PCA components
#' @export
db_svd <- function(dbm, k = 10, center = TRUE, scale = FALSE, center_rows = NULL,
                   memory_limit = getOption("dbMatrix.svd_memory", 8 * 1024^3),
                   return_format = c("svd", "pca")) {
  
  if (!requireNamespace("nanoarrow", quietly = TRUE))
    stop("Package 'nanoarrow' is required")
  
  if (is.null(center_rows)) center_rows <- TRUE
  
  # Column-wise centering/scaling not yet supported in C++ operator

  if (!center_rows && (center || scale))
    stop("center_rows=FALSE with center=TRUE or scale=TRUE is not yet supported")
  
  dbm <- .castNumeric(dbm) # Ensure numeric type
  con <- get_con(dbm)
  dims <- dim(dbm)
  n_rows <- dims[1]
  n_cols <- dims[2]
  
  # Validate dimensions

  if (n_rows == 0 || n_cols == 0)
    stop("Matrix must have at least one row and one column")
  
  # Validate k (Spectra requires k <= n_rows - 1 for the Gram operator)
  k <- as.integer(k)
  max_k <- min(n_rows - 1L, n_cols)
  if (k < 1) stop("k must be at least 1")
  if (max_k < 1) stop("Matrix too small for SVD (need at least 2 rows)")
  if (k > max_k) {
    warning(sprintf("k=%d exceeds max allowed (%d); reducing", k, max_k))
    k <- max_k
  }
  
  # Compute row means/sds for normalization (center_rows=TRUE when center||scale)
  means <- if (center) as.numeric(rowMeans(dbm)) else numeric(0)
  sds <- numeric(0)
  if (scale) {
    sds <- as.numeric(rowSds(dbm))
    sds[is.na(sds) | sds == 0] <- 1
  }

  tbl <- dbm[]
  final_sql <- as.character(dbplyr::sql_render(tbl))

  # Estimate data size for path selection
  nnz_df <- dplyr::collect(tbl |> dplyr::count())
  nnz <- as.numeric(nnz_df$n[1])
  if (is.null(nnz) || is.na(nnz) || nnz == 0) nnz <- n_rows * n_cols

  data_size <- nnz * 16
  use_cache <- memory_limit > 0 && data_size <= memory_limit

  if (getOption("dbMatrix.verbose", TRUE)) {
    if (!use_cache && memory_limit > 0) {
      cli::cli_alert_info("Data size ({format(structure(data_size, class='object_size'), units='auto')}) exceeds limit - using streaming iteration")
    }
  }

  stream_factory <- function() {
    res <- DBI::dbSendQuery(con, final_sql, arrow = TRUE)
    reader <- duckdb::duckdb_fetch_arrow(res, chunk_size = 1000000)
    nanoarrow::as_nanoarrow_array_stream(reader)
  }

  if (!is(dbm, "dbSparseMatrix"))
    stop("db_svd currently only supports dbSparseMatrix objects")

  # Determine row_offset, row_scale, col_scale for implicit normalization
  # A_norm = diag(row_scale) * A * diag(col_scale) + row_offset
  row_offset <- if (center) -means else rep(0, n_rows)
  row_scale <- rep(1, n_rows)
  col_scale <- rep(1, n_cols)

  if (scale) {
    inv_sds <- 1 / sds
    row_scale <- as.numeric(inv_sds)
    # Adjust offset: centering after scaling means offset = -mean/sd
    if (center) row_offset <- -means * inv_sds
  }
  
  if (use_cache) {
    # Fast Path: Arrow -> Eigen CSC
    if (getOption("dbMatrix.verbose", TRUE)) {
      cli::cli_alert_info("Data size ({format(structure(data_size, class='object_size'), units='auto')}) - Arrow/Eigen (CSC)")
    }
    
    result <- .compute_op_svd_arrow_cpp(
      stream_factory = stream_factory,
      n_rows = as.integer(n_rows),
      n_cols = as.integer(n_cols),
      row_offset = as.numeric(row_offset),
      row_scale = as.numeric(row_scale),
      col_scale = as.numeric(col_scale),
      k = as.integer(k)
    )
  } else {
    stop("Estimated SVD input exceeds memory_limit; increase memory_limit to use the in-memory Arrow/Eigen path.")
  }
  
  # Format results
  u_mat <- result$u
  v_mat <- result$v
  d_vec <- result$d
  
  if (!is.null(rownames(dbm)) && nrow(u_mat) == nrow(dbm))
    rownames(u_mat) <- rownames(dbm)
  if (!is.null(colnames(dbm)) && nrow(v_mat) == ncol(dbm))
    rownames(v_mat) <- colnames(dbm)
  
  return_format <- match.arg(return_format)
  
  if (return_format == "pca") {
    coords <- sweep(v_mat, 2, d_vec, "*")
    colnames(coords) <- paste0("Dim.", seq_len(k))
    colnames(u_mat) <- paste0("Dim.", seq_len(k))
    list(eigenvalues = d_vec^2, loadings = u_mat, coords = coords)
  } else {
    list(d = d_vec, u = u_mat, v = v_mat)
  }
}

# internal functions ####
#' get_tblName
#'
#' @param dbMatrix A database-backed object inheriting from `dbData`.
#' @return A character scalar giving the DuckDB table name associated with the
#'   object.
#'
#' @keywords internal
get_tblName <- function(dbMatrix) {
  # check if dbData is a dbData object
  if (!inherits(dbMatrix, "dbData")) {
    stopf("dbMatrix must be a dbData object")
  }

  tblName <- dbMatrix@name

  return(tblName)
}

#' get_dbdir
#'
#' @param dbMatrix A database-backed object inheriting from `dbData`.
#' @return A character scalar giving the DuckDB database directory used by the
#'   object.
#'
#' @keywords internal
#' string of the path to the database directory
get_dbdir <- function(dbMatrix) {
  # check if dbData is a dbData object
  if (!inherits(dbMatrix, "dbData")) {
    stopf("dbMatrix must be a dbData object")
  }

  dbdir <- dbMatrix@value[[1]]$con@driver@dbdir

  return(dbdir)
}


#' get_con
#'
#' @param dbMatrix A database-backed object inheriting from `dbData`.
#' @return A live DBI connection associated with the database-backed object.
#'
#' @keywords internal
#'
get_con <- function(dbMatrix) {
  # check if dbData is a dbData object
  if (!inherits(dbMatrix, "dbData")) {
    stopf("dbMatrix must be a dbData object")
  }

  # Use dbProject's connection accessor for consistency and auto-reconnection
  con <- dbProject::conn(dbMatrix)

  # Fallback to direct access if dbProject method doesn't work
  if (is.null(con)) {
    con <- dbMatrix@value[[1]]$con
  }

  return(con)
}

#' @importFrom MatrixGenerics colMeans colSums rowMeans rowSums colSds rowSds rowVars colVars
#' @importFrom dplyr compute
#' @importFrom Rcpp sourceCpp
#' @importFrom methods as is new validObject setClass setGeneric setMethod Arith Ops Math Summary
#' @importFrom stats rnorm setNames
#' @importFrom utils capture.output
#' @importFrom rlang abort warn inform
#' @importClassesFrom DBI DBIConnection
#' @importFrom nanoarrow as_nanoarrow_array_stream
#' @useDynLib dbMatrix, .registration = TRUE
#' @importFrom dbProject to_view conn conn<- dbReconnect dbList dbLoad
#' @importFrom dbProject .check_con .check_name .check_overwrite .check_tbl unique_table_name
NULL

#' @importFrom data.table ":="
NULL

# Print Formatting ####

#' @title Wrap message
#' @param ... additional strings and/or elements to pass to wrap_txt
#' @param sep how to join elements of string (default is one space)
#' @keywords internal
#' @noRd
wrap_msg <- function(..., sep = ' ') {
  message(wrap_txt(..., sep = sep))
}

#' @title Wrap text
#' @param ... additional params to pass
#' @param sep how to join elements of string (default is one space)
#' @param strWidth externally set wrapping width. (default value of 100 is not effected)
#' @param errWidth default = FALSE. Set strWidth to be compatible with error printout
#' @keywords internal
#' @noRd
wrap_txt <- function(..., sep = ' ', strWidth = 100, errWidth = FALSE) {
  custom_width <- ifelse(is.null(match.call()$strWidth), yes = FALSE, no = TRUE)
  if (!isTRUE(custom_width)) {
    if (isTRUE(errWidth)) strWidth <- getOption('width') - 6
  }

  cat(..., sep = sep) |>
    capture.output() |>
    strwrap(
      prefix = ' ',
      initial = '', # indent later lines, no indent first line
      width = min(80, getOption("width"), strWidth)
    ) |>
    paste(collapse = '\n')
}


# Custom stop function
stopf <- function(...) {
  wrap_txt('dbMatrix: ', ..., errWidth = TRUE) |>
    stop(call. = FALSE)
}


# From a vector, generate a string with the pattern 'item1', 'item2'
#' @keywords internal
#' @noRd
vector_to_string <- function(x) {
  toString(sprintf("'%s'", x))
}

#' @title Generate array for pretty printing of matrix values
#' @param i,j,x matched vectors of integers in i and j, with value in x
#' @param dims dimensions of the array (integer vector of 2)
#' @param fill fill character
#' @param digits default = 5. If numeric, round to this number of digits
#' @return No return value. Called for its side effect of printing a formatted
#'   matrix preview to the console.
#' @keywords internal
print_array <- function(
  i = NULL,
  j = NULL,
  x = NULL,
  dims,
  rownames = rep('', dims[1]),
  class = c('sparse', 'dense'),
  fill = '.',
  digits = 5L
) {
  total_len <- prod(dims)

  # pre-generate filler values
  if (class == "dense") {
    n_digits <- getOption('dbMatrix.digits', default = 7)
    a_vals <- rep(format(round(0, n_digits), nsmall = n_digits), total_len)
  } else {
    a_vals <- rep('.', total_len)
  }

  ijx_nargs <- sum(!is.null(i), !is.null(j), !is.null(x))
  if (ijx_nargs < 3 && ijx_nargs > 1) {
    stopf('All values for i, j, and x must be given when printing')
  }
  if (ijx_nargs == 3) {
    # format numeric
    if (is.numeric(x)) {
      ifelse(
        x < 1e4,
        format(x, digits = digits),
        format(x, digits = digits, scientific = TRUE)
      )
    }
    # populate sparse values by replace nth elements
    # note that i and j index values must be determined outside of this function
    # since the colnames are not known in here
    for (n in seq_along(x)) {
      a_vals[ij_array_map(i = i[n], j = j[n], dims = dims)] <- x[n]
    }
  }

  # print array
  # Ensure max.print is large enough to show the entire preview array
  # otherwise capture.output will truncate and cause NAs in the show method
  op <- options(max.print = total_len + 1000L)
  on.exit(options(op))

  array(a_vals, dims, dimnames = list(rownames, rep('', dims[2]))) |>
    print(quote = FALSE, right = TRUE)
}

# Map row (i) and col (j) indices to nth value of an array vector
#' @keywords internal
#' @noRd
#' @return integer position in array vector the i and j map to
ij_array_map <- function(i, j, dims) {
  # arrays map vector values first by row then by col
  (j - 1) * dims[1] + i
}

# DBI ####

# ## dbDisconnect ####
# #' @title dbDisconnect
# #' @rdname DBI
# #' @export
# setMethod('dbDisconnect', signature(x = 'dbMatrix'),
#           function(x, ...){
#             con <- get_con(x)
#             DBI::dbDisconnect(conn = con, shutdown = TRUE)
#           })
#
# ## dbListTables ####
# #' @title dbListTables
# #' @rdname DBI
# #' @export
# setMethod('dbListTables', signature(x = 'dbMatrix'),
#           function(x, ...){
#             con <- get_con(x)
#             DBI::dbListTables(conn = con)
#           })

# dbMatrix ####

## dbLoad ####
#' Create a dbMatrix object computed in a database
#' @param conn DBIConnection object
#' @param name valid name value (character)
#' @param class character, class of the dbMatrix object (e.g. "dbDenseMatrix" or "dbSparseMatrix")
#' @export
#' @rdname dbMatrix-methods
setMethod(
  'dbLoad',
  signature(conn = 'DBIConnection'),
  function(conn, name, class) {
    .check_con(conn = conn)
    if (!name %in% DBI::dbListTables(conn)) {
      stopf("'name' not found in database")
    }
    if (!class %in% c('dbDenseMatrix', 'dbSparseMatrix')) {
      stopf("Class must be 'dbDenseMatrix' or 'dbSparseMatrix'")
    }

    dim_names <- c(
      paste0("__", name, "_rownames"),
      paste0("__", name, "_colnames")
    )

    # check if rownames and colnames exist
    if (!all(dim_names %in% DBI::dbListTables(conn))) {
      stopf("Dimension names not found. Did you save with dbMatrix::compute?")
    }

    # load values saved from dbMatrix::compute()
    # IMPORTANT: must sort by i/j to ensure correct ordering.
    # DuckDB table scan order is not guaranteed to match insertion order.
    rownames <- dplyr::tbl(conn, dim_names[1]) |>
      dplyr::arrange(i) |>
      dplyr::pull('rownames')
    colnames <- dplyr::tbl(conn, dim_names[2]) |>
      dplyr::arrange(j) |>
      dplyr::pull('colnames')
    dim_names <- list(as.factor(rownames), as.factor(colnames))
    dims <- c(length(rownames), length(colnames))
    value <- dplyr::tbl(conn, name)

    x <- dbMatrix::dbMatrix(
      value = value,
      class = class,
      con = conn,
      name = name,
      dim_names = dim_names,
      dims = dims,
      overwrite = 'PASS'
    )

    return(x)
  }
)

# dbplyr ####

#' Generate table names
#' @details
#' based on dbplyr::unique_table_name
#'
#' @noRd
#' @keywords internal
unique_table_name <- function(prefix = "dbMatrix") {
  vals <- c(letters, LETTERS, 0:9)
  name <- paste0(sample(vals, 10, replace = TRUE), collapse = "")
  paste0(prefix, "_", name)
}

# Helper to check memory limit
#' @keywords internal
#' @noRd
.check_mem_limit <- function(x) {
  limit <- getOption("dbMatrix.max_mem_convert", default = 8 * 1024^3) # 8GB default

  # Estimate size: rows * cols * 8 bytes (double)
  # This is a conservative estimate for dense matrices.
  # For sparse, it might be an overestimate, but safer.
  dims <- dim(x)
  est_size <- as.numeric(dims[1]) * as.numeric(dims[2]) * 8

  if (est_size > limit) {
    stop(sprintf(
      "dbMatrix: Implicit conversion to in-memory matrix blocked.\nEstimated size: %s\nLimit: %s\nIncrease 'dbMatrix.max_mem_convert' option to override.",
      format(structure(est_size, class = "object_size"), units = "auto"),
      format(structure(limit, class = "object_size"), units = "auto")
    ))
  }

  if (getOption("dbMatrix.verbose", default = TRUE)) {
    cli::cli_alert_info(
      "Coercing dbMatrix to in-memory matrix (est. size: {format(structure(est_size, class = 'object_size'), units = 'auto')}). Control with 'dbMatrix.max_mem_convert'. See ?dbMatrix-options for more info."
    )
  }
}

# Temp table cleanup ####

#' Extract temp table names from SQL query
#'
#' Finds all __dbM_* prefixed table names in a SQL query string.
#' These are temporary tables created during operations like subsetting.
#'
#' @param sql Character string of SQL query
#' @return Character vector of temp table names found
#' @keywords internal
#' @noRd
.extract_temp_tables <- function(sql) {
  # Match __dbM_ followed by word characters (alphanumeric + underscore)
  pattern <- "__dbM_[a-zA-Z0-9_]+"
  matches <- regmatches(sql, gregexpr(pattern, sql))[[1]]
  unique(matches)
}

#' Cleanup temp tables from a connection
#'
#' Unregisters and drops temp tables created during dbMatrix operations.
#'
#' @param con DuckDB connection
#' @param tables Character vector of table names to clean up
#' @param verbose Logical, whether to print cleanup messages
#' @keywords internal
#' @noRd
.cleanup_temp_tables <- function(
  con,
  tables,
  verbose = getOption("dbMatrix.verbose", TRUE)
) {
  if (length(tables) == 0) {
    return(invisible(NULL))
  }

  for (tbl in tables) {
    # Try to unregister (for registered views)
    try(duckdb::duckdb_unregister(con, tbl), silent = TRUE)
    # Try to drop table (for materialized temps)
    try(
      DBI::dbExecute(con, paste0("DROP TABLE IF EXISTS \"", tbl, "\"")),
      silent = TRUE
    )
  }

  if (verbose && length(tables) > 0) {
    cli::cli_alert_success("Cleaned up {length(tables)} temp table(s).")
  }

  invisible(tables)
}

# Named long format conversion ####

#' Convert dbMatrix to named ijx table
#'
#' Converts a `dbMatrix` to a lazy long table where row and column indices are
#' replaced by dimension names.
#'
#' @param x A dbMatrix object (dbSparseMatrix or dbDenseMatrix)
#' @param row_col Name for the row-name column (default: "row_name")
#' @param col_col Name for the column-name column (default: "col_name")
#' @param compute Whether to materialize as temp table (default: FALSE)
#' @return A lazy tbl with columns: row_col, col_col, x
#' @concept dbMatrix
#' @export
to_named_ijx_tbl <- function(
  x,
  row_col = "row_name",
  col_col = "col_name",
  compute = FALSE
) {
  if (!inherits(x, "dbMatrix")) {
    stopf("Input must be a dbMatrix object")
  }

  # Get connection and dimension names
  con <- dbplyr::remote_con(x[])
  row_names <- rownames(x)
  col_names <- colnames(x)

  # Build lookup data.frames
  row_lookup <- data.frame(
    i = seq_along(row_names),
    .row_name_tmp = row_names,
    stringsAsFactors = FALSE
  )
  col_lookup <- data.frame(
    j = seq_along(col_names),
    .col_name_tmp = col_names,
    stringsAsFactors = FALSE
  )

  # Copy to database as temporary tables
  row_tbl <- dplyr::copy_to(
    con,
    row_lookup,
    name = unique_table_name("row_lookup"),
    temporary = TRUE,
    overwrite = TRUE
  )
  col_tbl <- dplyr::copy_to(
    con,
    col_lookup,
    name = unique_table_name("col_lookup"),
    temporary = TRUE,
    overwrite = TRUE
  )

  # Join and rename
  result <- x[] |>
    dplyr::left_join(row_tbl, by = "i") |>
    dplyr::left_join(col_tbl, by = "j") |>
    dplyr::select(
      !!rlang::sym(row_col) := .row_name_tmp,
      !!rlang::sym(col_col) := .col_name_tmp,
      x
    )

  if (compute) {
    result <- dplyr::compute(
      result,
      name = unique_table_name("named_long"),
      temporary = TRUE,
      overwrite = TRUE
    )
  }

  result
}

#' @keywords internal
#' @noRd
.to_named_long <- function(
  x,
  row_col = "row_name",
  col_col = "col_name",
  compute = FALSE
) {
  to_named_ijx_tbl(
    x = x,
    row_col = row_col,
    col_col = col_col,
    compute = compute
  )
}

# callGeneric does not seem to work in dplyr chains
# will need to write them out
# setMethod('Ops', signature(e1 = 'dbMatrix', e2 = 'ANY'), function(e1, e2)
# {
#   e1[] = e1[] |> dplyr::mutate(x = callGeneric(e1 = x, e2 = e2))
#   e1
# })

# Math Ops Helpers ####

#' @noRd
ops_ordered_args_vect <- function(dbm_narg, a, b) {
  switch(
    dbm_narg,
    paste0('`(', a, ', ', b, '))'),
    paste0('`(', b, ', ', a, '))')
  )
}

#' @noRd
arith_call_dbm <- function(dbm_narg, dbm, num_vect, generic_char) {
  # order matters
  ordered_args <- ops_ordered_args_vect(dbm_narg, 'x', 'num_vect')

  # helper functions
  .skip_computation <- function(num_vect, generic_char) {
    if ((generic_char %in% c('*', '/') && all(num_vect == 1))) {
      return(TRUE)
    }

    if ((generic_char %in% c('+', '-') && all(num_vect == 0))) {
      return(TRUE)
    }

    return(FALSE)
  }

  .do_vect_multi <- function(num_vect, generic_char, dbm) {
    if (length(num_vect) > 1) {
      return(TRUE)
    }

    if (is(dbm, 'dbDenseMatrix')) {
      return(FALSE)
    }

    if (is.infinite(num_vect) && generic_char %in% c('*')) {
      return(TRUE)
    }

    if (num_vect != 0 && generic_char %in% c('-', '+')) {
      return(TRUE)
    }

    if (num_vect == 0 && generic_char %in% c('/')) {
      return(TRUE)
    }

    return(FALSE)
  }

  .is_additive_ops <- function(num_vect) {
    if (num_vect == "+" || num_vect == "-") {
      return(TRUE)
    } else {
      return(FALSE)
    }
  }

  .do_densification <- function(num_vect, generic_char, dbm) {
    if (
      is(dbm, "dbSparseMatrix") &&
        length(num_vect) == 1 &&
        .is_additive_ops(generic_char)
    ) {
      return(TRUE)
    }

    if (
      is(dbm, "dbSparseMatrix") &&
        length(num_vect) == 1 &&
        (generic_char == "^" || generic_char == "%%") &&
        num_vect == 0
    ) {
      return(TRUE)
    }

    return(FALSE)
  }

  # Main function
  if (all(is.nan(num_vect))) {
    if (is(dbm, 'dbSparseMatrix')) {
      dbm <- .to_db_dense(x = dbm)
    }

    dbm[] <- dbm[] |> dplyr::mutate(x = 0.0 / 0.0) # generate NaNs

    return(dbm)
  }

  if (.skip_computation(num_vect, generic_char)) {
    return(dbm)
  }

  if (.do_densification(num_vect, generic_char, dbm)) {
    # Densify sparse matrix for operations that require it
    if (is(dbm, "dbSparseMatrix")) {
      dbm <- .to_db_dense(x = dbm)
    }
  }

  if (.do_vect_multi(num_vect, generic_char, dbm)) {
    return(arith_call_dbm_vect_multi(dbm, num_vect, generic_char, ordered_args))
  }

  if (generic_char == "^" && num_vect == Inf) {
    # edge case 1
    build_call <- paste0(
      "dbm[] |> ",
      "dplyr::mutate(",
      "x = dplyr::case_when(",
      "x < 0 ~ 0/0, ", # workaround to generate NaNs (0/0)
      "TRUE ~ `",
      generic_char,
      ordered_args,
      ") "
    )
  } else if (generic_char == "%/%" && num_vect == 0) {
    # edge case 2
    dbm <- .to_db_dense(x = dbm)
    dbm[] <- dbm[] |>
      dplyr::mutate(
        x = dplyr::case_when(
          x < 0 ~ -Inf,
          x > 0 ~ Inf,
          x == 0 ~ 0 / 0,
          TRUE ~ x
        )
      )
    return(dbm)
  } else if (generic_char == "%/%" && num_vect == Inf) {
    # edge case 3
    dbm[] <- dbm[] |>
      dplyr::mutate(
        x = dplyr::case_when(
          x < 0 ~ -1,
          x > 0 ~ 0,
          TRUE ~ x
        )
      )
    return(dbm)
  } else if (generic_char == "%%" && num_vect == Inf) {
    # edge case 4
    dbm[] <- dbm[] |>
      dplyr::mutate(x = dplyr::case_when(x < 0 ~ Inf, TRUE ~ x))

    return(dbm)
  } else {
    # general case
    build_call <- paste0(
      'dbm[] |> dplyr::mutate(x = `',
      generic_char,
      ordered_args
    )
  }

  dbm[] <- eval(str2lang(build_call))
  return(dbm)
}

#' @noRd
arith_call_dbm_vect_multi <- function(
  dbm,
  num_vect,
  generic_char,
  ordered_args
) {
  # get inputs
  con <- dbplyr::remote_con(dbm[])

  # handle order
  if (ordered_args == '`(x, num_vect))') {
    swap_arith_order <- FALSE
  } else {
    swap_arith_order <- TRUE
  }

  # Create dbDenseMatrix with vec_matrix
  dbv <- .as_vector_matrix(vector = num_vect, con = con)
  dbm <- .join_dbm_vect(
    dbm = dbm,
    vec_matrix = dbv,
    op = generic_char,
    swap_arith_order = swap_arith_order
  )
  return(dbm)
}

#' @keywords internal
#' @noRd
.as_vector_matrix <- function(vector, con) {
  ijx <- dplyr::tibble(i = seq_along(vector), j = 1, x = vector) |>
    dplyr::copy_to(
      dest = con,
      name = unique_table_name('vec_matrix'),
      temporary = TRUE,
      overwrite = TRUE
    )

  res <- new(
    Class = "dbDenseMatrix",
    value = ijx,
    name = NA_character_,
    init = TRUE,
    dims = c(length(vector), 1L),
    dim_names = list(paste0('row', seq_along(vector)), c('col1'))
  )

  return(res)
}

#' @title Evaluate if a dbSparseMatrix should be densified
#' @param generic_char A character string representing the operation to be performed.
#' @param vec_matrix A \code{dbMatrix} object with 1D row or col.
#' @details
#' Evaluates if a \code{dbSparseMatrix} should be
#' densified for `[Arith]` operations and specific scalar values for
#' operations in the order of dbSparseMatrix, vector
#'
#'
#' @return `TRUE` if the operation should densify the sparse matrix before
#'   evaluation, otherwise `FALSE`.
#' @keywords internal
.eval_op_densify <- function(generic_char, vec_matrix) {
  if (generic_char == "+" || generic_char == "-") {
    if (suppressWarnings(all(vec_matrix == 0))) {
      return(FALSE)
    } else {
      return(TRUE)
    }
  } else if (generic_char == "*") {
    if (
      suppressWarnings(any(vec_matrix == Inf)) ||
        suppressWarnings(any(vec_matrix == NaN))
    ) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  } else if (
    generic_char == "/" ||
      generic_char == "%%" ||
      generic_char == "%/%" ||
      generic_char == "^"
  ) {
    if (
      suppressWarnings(any(vec_matrix == 0)) ||
        suppressWarnings(any(vec_matrix == NaN))
    ) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  } else {
    stopf("Invalid operator: %s", generic_char)
  }
}

#' @title Join a \code{dbSparseMatrix} with a \code{dbMatrix} object
#' @param dbm A \code{dbSparseMatrix} object.
#' @param vec_matrix A \code{dbMatrix} object with 1D row or col.
#' @param op A character string representing the operation to be performed.
#' @param swap_arith_order order of the arguments for the operation. default: NULL
#' @return A [`dbMatrix`] object containing the result of applying `op` between
#'   `dbm` and `vec_matrix`.
#' @keywords internal
.join_dbm_vect <- function(dbm, vec_matrix, op, swap_arith_order = FALSE) {
  # check inputs
  con <- dbplyr::remote_con(vec_matrix[])
  n_rows <- bit64::as.integer64.integer(dim(dbm)[1])
  n_cols <- bit64::as.integer64.integer(dim(dbm)[2])
  length <- bit64::as.integer64.integer(length(vec_matrix))
  total_dims <- n_rows * n_cols

  # validate dimensions
  rep <- ceiling(total_dims / length) #FIXME: R cannot do math with int64
  if (is.na(rep)) {
    stopf("total_dims must be a positive integer")
  }

  # check for length mismatch
  remainder <- total_dims %% length
  if (remainder > 0) {
    cli::cli_alert_warning(
      wrap = FALSE,
      "Longer object length [{total_dims}] is not a multiple of shorter
     object length [{length}]"
    )
  }

  # helper functions
  .perform_dense_join <- function(dbm, vec_matrix, op) {
    # get precomputed matrix
    precomp <- .initialize_precompute_matrix(con, n_rows, n_cols)

    # Prepare precomp with v_idx
    # We use idx (0-based) which is explicitly present in precomp table
    precomp_aug <- precomp |>
      dplyr::mutate(
        v_idx = dplyr::sql(glue::glue(
          "(((j - 1) * {n_rows} + (i - 1)) % {length}) + 1"
        ))
      )

    # Prepare vec_matrix (v)
    v_tbl <- vec_matrix[] |> dplyr::select(v_i = i, v_x = x)

    # Prepare dbm (m)
    # Join on i, j directly to avoid calculating idx
    m_tbl <- dbm[] |> dplyr::select(m_i = i, m_j = j, m_x = x)

    # Join everything
    # Join precomp (i, j) with dbm (m_i, m_j)
    res <- precomp_aug |>
      dplyr::inner_join(m_tbl, by = c("i" = "m_i", "j" = "m_j")) |>
      dplyr::left_join(v_tbl, by = c("v_idx" = "v_i"))

    # Calculate x
    if (swap_arith_order) {
      res <- res |> dplyr::mutate(x = dplyr::sql(glue::glue("v_x {op} m_x")))
    } else {
      res <- res |> dplyr::mutate(x = dplyr::sql(glue::glue("m_x {op} v_x")))
    }

    return(res |> dplyr::select(i, j, x))
  }

  .perform_sparse_join <- function(dbm, vec_matrix, op) {
    # Prepare tables
    m_tbl <- dbm[] |> dplyr::select(i, j, m_x = x)
    v_tbl <- vec_matrix[] |> dplyr::select(v_i = i, v_x = x)

    # Join
    res <- m_tbl |>
      dplyr::left_join(v_tbl, by = c("i" = "v_i"))

    # Calculate x
    if (swap_arith_order) {
      res <- res |> dplyr::mutate(x = dplyr::sql(glue::glue("v_x {op} m_x")))
    } else {
      res <- res |> dplyr::mutate(x = dplyr::sql(glue::glue("m_x {op} v_x")))
    }

    return(res |> dplyr::select(i, j, x))
  }

  # Column-wise join (for vector length = ncol)
  .perform_col_join <- function(dbm, vec_matrix, op) {
    # Prepare tables
    m_tbl <- dbm[] |> dplyr::select(i, j, m_x = x)
    v_tbl <- vec_matrix[] |> dplyr::select(v_j = i, v_x = x) # Note: vec_matrix uses 'i' as index

    # Join on column index
    res <- m_tbl |>
      dplyr::left_join(v_tbl, by = c("j" = "v_j"))

    # Calculate x
    if (swap_arith_order) {
      res <- res |> dplyr::mutate(x = dplyr::sql(glue::glue("v_x {op} m_x")))
    } else {
      res <- res |> dplyr::mutate(x = dplyr::sql(glue::glue("m_x {op} v_x")))
    }

    return(res |> dplyr::select(i, j, x))
  }

  # main function
  if (is(dbm, "dbSparseMatrix")) {
    if (.eval_op_densify(generic_char = op, vec_matrix = vec_matrix)) {
      dense_dbm <- .to_db_dense(dbm)
      dense_dbm[] <- .perform_dense_join(dense_dbm, vec_matrix, op)
      return(dense_dbm)
    } else {
      if (nrow(dbm) == length(vec_matrix)) {
        # Row-wise operation (vector length = nrow)
        dbm[] <- .perform_sparse_join(dbm, vec_matrix, op)
        return(dbm)
      } else if (ncol(dbm) == length(vec_matrix)) {
        # Column-wise operation (vector length = ncol)
        dbm[] <- .perform_col_join(dbm, vec_matrix, op)
        return(dbm)
      } else {
        # General recycling case
        dbm[] <- .perform_dense_join(dbm, vec_matrix, op)
        return(dbm)
      }
    }
  } else {
    # dbDenseMatrix
    if (nrow(dbm) == length(vec_matrix)) {
      # Row-wise operation (vector length = nrow)
      dbm[] <- .perform_sparse_join(dbm, vec_matrix, op)
      return(dbm)
    } else if (ncol(dbm) == length(vec_matrix)) {
      # Column-wise operation (vector length = ncol)
      dbm[] <- .perform_col_join(dbm, vec_matrix, op)
      return(dbm)
    } else {
      # General recycling case
      dbm[] <- .perform_dense_join(dbm, vec_matrix, op)
      return(dbm)
    }
  }
}

# Math Ops ####

## Arith: dbm_e2 ####
#' Arith dbMatrix, e2
#' @description
#' See `methods::Arith` for more details.
#' @param e1 First operand.
#' @param e2 Second operand.
#' @return \itemize{
#'   \item Arithmetic and logical group methods return a [`dbMatrix`] object of
#'   the appropriate dense or sparse subclass, with the same dimensions as the
#'   input and transformed values stored in DuckDB.
#'   \item `dbLoad()` returns a [`dbDenseMatrix`] or [`dbSparseMatrix`] pointing
#'   to an existing DuckDB table.
#'   \item `writeMM()` writes a Matrix Market file to `file` and returns
#'   `invisible(TRUE)` on success.
#' }
#' @rdname dbMatrix-methods
#' @export
#' @usage \S4method{Arith}{dbMatrix,ANY}(e1, e2)
setMethod('Arith', signature(e1 = 'dbMatrix', e2 = 'ANY'), function(e1, e2) {
  dbm <- .castNumeric(e1)

  num_vect <- if (typeof(e2) != 'double') {
    as.numeric(e2)
  } else {
    e2
  }

  arith_call_dbm(
    dbm_narg = 1L,
    dbm = dbm,
    num_vect = num_vect,
    generic_char = as.character(.Generic)
  )
})

## Arith: e1_dbm ####
#' Arith e1, dbMatrix
#' @description
#' See `methods::Arith` for more details.
#' @rdname dbMatrix-methods
#' @export
#' @usage \S4method{Arith}{ANY,dbMatrix}(e1, e2)
setMethod('Arith', signature(e1 = 'ANY', e2 = 'dbMatrix'), function(e1, e2) {
  dbm <- .castNumeric(e2)

  num_vect <- if (typeof(e1) != 'double') {
    as.numeric(e1)
  } else {
    e1
  }

  arith_call_dbm(
    dbm_narg = 2L,
    dbm = dbm,
    num_vect = num_vect,
    generic_char = as.character(.Generic)
  )
})

## Arith: dbm_dbm ####
#' Arith dbMatrix, e2
#' @description
#' See `methods::Arith` for more details.
#' @rdname dbMatrix-methods
#' @export
setMethod(
  'Arith',
  signature(e1 = 'dbMatrix', e2 = 'dbMatrix'),
  function(e1, e2) {
    generic_char <- as.character(.Generic)
    dim1 <- dim(e1)
    dim2 <- dim(e2)

    # Check for arith operations if neither is a vec_matrix
    if (!1 %in% c(dim1, dim2)) {
      if (generic_char %in% c('/', '^', '%%', '%/%')) {
        # Division for dbSparseMatrix: only safe if all numerator positions
        # are covered by denominator positions (support-preserving)
        if (
          generic_char == '/' &&
            is(e1, "dbSparseMatrix") &&
            is(e2, "dbSparseMatrix")
        ) {
          # Check for uncovered positions (would produce Inf from 0 denominator)
          uncovered <- e1[] |>
            dplyr::select(i, j) |>
            dplyr::anti_join(e2[] |> dplyr::select(i, j), by = c('i', 'j')) |>
            dplyr::collect()

          if (nrow(uncovered) > 0) {
            stopf(
              "dbSparseMatrix division: %d numerator positions have implicit
              zeros in denominator (would produce Inf). Ensure denominator
              covers all numerator positions, or convert to dbDenseMatrix.",
              nrow(uncovered)
            )
          }

          # Safe to proceed with sparse-preserving inner_join division
          e1[] <- e1[] |>
            dplyr::inner_join(e2[], by = c('i', 'j')) |>
            dplyr::mutate(x = x.x / x.y) |>
            dplyr::select(i, j, x)
          e1@name <- NA_character_
          return(e1)
        }
        # Allow division for dbDenseMatrix (no implicit zeros issue)
        if (
          generic_char == '/' &&
            is(e1, "dbDenseMatrix") &&
            is(e2, "dbDenseMatrix")
        ) {
          # Continue to full_join logic below
        } else if (generic_char != '/') {
          # Other ops (^, %%, %/%) still blocked for now
          stopf(
            "Arith operations with '^', '%%', '%/%' are not yet supported
              between dbMatrix objects."
          )
        } else {
          # Mixed sparse/dense division - block for now
          stopf(
            "Division between mixed dbSparseMatrix/dbDenseMatrix not supported.
              Convert both to same type first."
          )
        }
      }
    }

    # Case 1: vec_matrix, vec_matrix
    if (1 %in% dim2 && 1 %in% dim1) {
      if (!all(dim1 == dim2)) {
        stopf("vec_matrix-vec_matrix recycling not yet supported.")
      }
    } else if (1 %in% dim2) {
      # Case 2: dbMatrix, vec_matrix
      res <- .join_dbm_vect(dbm = e1, vec_matrix = e2, op = generic_char)
      return(res)
    } else if (1 %in% dim1) {
      # Case 3: vec_matrix, dbMatrix
      res <- .join_dbm_vect(
        dbm = e2,
        vec_matrix = e1,
        op = generic_char,
        swap_arith_order = TRUE
      )
      return(res)
    } else {
      # Case 4: Both inputs are dbMatrix
      if (!all(dim1 == dim2)) {
        stopf('non-conformable matrix dimensions')
      }

      # Perform full join operation on dbMatrix
      # TODO: avoid full join
      build_call <- glue::glue(
        "e1[] |>
         dplyr::full_join(e2[], by = c('i', 'j')) |>
         dplyr::mutate(x = `{generic_char}`(dplyr::coalesce(x.x, 0),
                       dplyr::coalesce(x.y, 0))) |>
         dplyr::select(i, j, x) |>
         dplyr::arrange(i)"
      )
      e1[] <- eval(str2lang(build_call))
      e1@name <- NA_character_
      return(e1)
    }
  }
)

## Ops: dbm_e2 ####
#' Ops dbMatrix, e2
#' @description
#' See `methods::Ops` for more details.
#' @rdname dbMatrix-methods
#' @export
#' @usage \S4method{Ops}{dbMatrix,ANY}(e1, e2)
setMethod('Ops', signature(e1 = 'dbMatrix', e2 = 'ANY'), function(e1, e2) {
  op <- as.character(.Generic)
  compare_ops <- c(">", "<", ">=", "<=", "==", "!=")

  if (is.na(e2) && op == "==") {
    build_call <- glue::glue('e1[] |> dplyr::mutate(x = as.numeric(is.na(x)))')
  } else if (is.na(e2) && op == "!=") {
    build_call <- glue::glue('e1[] |> dplyr::mutate(x = as.numeric(!is.na(x)))')
  } else if (op %in% compare_ops) {
    # Comparison operators return BOOLEAN - cast to numeric for type safety
    build_call <- glue::glue(
      'e1[] |> dplyr::mutate(x = as.numeric(`',
      op,
      '`(x, e2)))'
    )
  } else {
    build_call <- glue::glue('e1[] |> dplyr::mutate(x = `', op, '`(x, e2))')
  }

  e1[] <- eval(str2lang(build_call))
  e1@name <- NA_character_
  return(e1)
})

## Ops: e1_dbm ####
#' Ops e1, dbMatrix
#' @description
#' See `methods::Ops` for more details.
#' @rdname dbMatrix-methods
#' @export
#' @usage \S4method{Ops}{ANY,dbMatrix}(e1, e2)
setMethod('Ops', signature(e1 = 'ANY', e2 = 'dbMatrix'), function(e1, e2) {
  op <- as.character(.Generic)
  compare_ops <- c(">", "<", ">=", "<=", "==", "!=")

  if (op %in% compare_ops) {
    # Comparison operators return BOOLEAN - cast to numeric for type safety
    build_call <- glue::glue(
      'e2[] |> dplyr::mutate(x = as.numeric(`',
      op,
      '`(e1, x)))'
    )
  } else {
    build_call <- glue::glue('e2[] |> dplyr::mutate(x = `', op, '`(e1, x))')
  }

  e2[] <- eval(str2lang(build_call))
  e2@name <- NA_character_
  e2
})

## Ops: dbm_dbm ####
#' Ops dbMatrix, dbMatrix
#' @description
#' See `methods::Ops` for more details.
#' @rdname dbMatrix-methods
#' @export
#' @usage \S4method{Ops}{dbMatrix,dbMatrix}(e1, e2)
setMethod('Ops', signature(e1 = 'dbMatrix', e2 = 'dbMatrix'), function(e1, e2) {
  if (!any(e1@dims %in% e2@dims)) {
    stopf('non-conformable matrix dimensions')
  }

  op <- as.character(.Generic)
  compare_ops <- c(">", "<", ">=", "<=", "==", "!=")

  if (op %in% compare_ops) {
    # Comparison operators return BOOLEAN - cast to numeric for type safety
    build_call <- glue::glue(
      "
      e1[] |>
      dplyr::left_join(e2[], by = c('i', 'j'), suffix = c('', '.y')) |>
      dplyr::mutate(x = as.numeric(`",
      op,
      "`(x, x.y))) |>
      dplyr::select(c('i', 'j', 'x'))
    "
    )
  } else {
    build_call <- glue::glue(
      "
      e1[] |>
      dplyr::left_join(e2[], by = c('i', 'j'), suffix = c('', '.y')) |>
      dplyr::mutate(x = `",
      op,
      "`(x, x.y)) |>
      dplyr::select(c('i', 'j', 'x'))
    "
    )
  }

  e1[] <- eval(str2lang(build_call))
  e1@name <- NA_character_
  e1
})

## %in% operator ####
#' @title Value Matching
#'
#' @description
#' Implements the `%in%` operator for dbMatrix objects. This operator checks if
#' elements from the left operand are contained in the right operand, returning
#' a logical vector.
#'
#' @param x A dbMatrix object or any other object
#' @param table Any object or a dbMatrix object
#'
#' @details
#' This is a method for the standard `%in%` operator for dbMatrix objects.
#' It follows R's standard behavior for the `%in%` operator:
#'
#' - When `x` is a dbDenseMatrix, it returns a logical vector with the same length as the
#'   total number of elements in the matrix.
#' - When `table` is a dbDenseMatrix, it allows checking if elements in `x` are in the matrix.
#' - For dbSparseMatrix objects, it throws an error to match the behavior of dgCMatrix.
#'
#' @return
#' A logical vector of the same length as `x`, indicating which elements of `x` are in `table`.
#'
#' @examples
#' con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
#' mat <- matrix(1:9, nrow = 3, ncol = 3)
#' dbmat <- dbMatrix(
#'   value = mat,
#'   con = con,
#'   name = "example_matrix",
#'   class = "dbDenseMatrix",
#'   overwrite = TRUE
#' )
#'
#' dbmat %in% c(1, 3, 5, 7, 9)
#'
#' c(1, 3, 5, 7, 9) %in% dbmat
#' DBI::dbDisconnect(con, shutdown = TRUE)
#' @rdname percent-in
#' @concept transform
#' @export
setMethod(
  "%in%",
  signature(x = "dbDenseMatrix", table = "ANY"),
  function(x, table) {
    if (length(table) == 0) {
      return(rep(FALSE, length(x)))
    }

    db_vector <- x[] |>
      dplyr::arrange(j, i) |>
      dplyr::pull(x)

    return(db_vector %in% table)
  }
)


#' @rdname percent-in
#' @concept transform
#' @export
setMethod(
  "%in%",
  signature(x = "ANY", table = "dbDenseMatrix"),
  function(x, table) {
    if (length(x) == 0) {
      return(logical(0))
    }

    table_data <- table[] |>
      dplyr::arrange(j, i) |>
      dplyr::pull(x)

    result <- x %in% table_data
    return(result)
  }
)


#' @rdname percent-in
#' @concept transform
#' @export
setMethod(
  "%in%",
  signature(x = "dbSparseMatrix", table = "ANY"),
  function(x, table) {
    stop("'match' requires vector arguments")
  }
)

# Math Summary Ops ####
## rowSums dbdm ####
#' Row (column) sums for [`dbMatrix`] objects
#' @inherit MatrixGenerics::rowSums description
#' @inheritParams MatrixGenerics::rowSums
#' @param na.rm Always TRUE for [`dbMatrix`]. Included for compatibility
#' with the generic.
#' @param dims Always 1 for [`dbMatrix`] queries. Included for compatibility with
#' the generic.
#' @param memory logical. If FALSE (default), results returned as dbDenseMatrix. This is recommended
#' for large computations. Set to TRUE to return the results as a vector.
#' @return A named numeric vector containing one sum per row or column of `x`.
#' @concept summary
#' @rdname row_col_sums
#' @export
#' @usage \S4method{rowSums}{dbDenseMatrix}(x, na.rm = FALSE, dims = 1, ..., memory = FALSE)
setMethod(
  'rowSums',
  signature(x = 'dbDenseMatrix'),
  function(x, na.rm = FALSE, dims = 1, ..., memory = FALSE) {
    x <- .castNumeric(x)
    n_rows <- nrow(x)
    rn <- rownames(x)

    # calculate rowSums - collect with row indices
    res_df <- x[] |>
      dplyr::group_by(i) |>
      dplyr::summarise(sum_x = sum(x, na.rm = TRUE)) |>
      dplyr::collect()

    # Build result vector with correct ordering using i indices
    res <- numeric(n_rows)
    res[res_df$i] <- res_df$sum_x
    names(res) <- rn
    return(res)
  }
)

## rowSums dbsm ####
#' @concept summary
#' @rdname row_col_sums
#' @export
setMethod(
  'rowSums',
  signature(x = 'dbSparseMatrix'),
  function(x, ...) {
    x <- .castNumeric(x)
    num_row <- nrow(x)

    res_sparse <- x[] |>
      dplyr::group_by(i) |>
      dplyr::summarise(sum_x = sum(x, na.rm = TRUE)) |>
      dplyr::collect()

    res <- numeric(num_row)
    res[res_sparse$i] <- res_sparse$sum_x
    names(res) <- rownames(x)
    return(res)
  }
)

## colSums dbdm####
#' @rdname row_col_sums
#' @export
setMethod(
  'colSums',
  signature(x = 'dbDenseMatrix'),
  function(x, ...) {
    x <- .castNumeric(x)

    # calculate colSums
    res <- x[] |>
      dplyr::group_by(j) |>
      dplyr::summarise(sum_x = sum(x, na.rm = TRUE)) |>
      dplyr::collapse() |>
      dplyr::arrange(j) |>
      dplyr::pull(sum_x)

    names(res) <- colnames(x)
    return(res)
  }
)

## colSums dbsm ####
#' @concept summary
#' @rdname row_col_sums
#' @export
setMethod(
  'colSums',
  signature(x = 'dbSparseMatrix'),
  function(x, ...) {
    x <- .castNumeric(x)
    num_col <- ncol(x)

    res_sparse <- x[] |>
      dplyr::group_by(j) |>
      dplyr::summarise(sum_x = sum(x, na.rm = TRUE)) |>
      dplyr::collect()

    res <- numeric(num_col)
    res[res_sparse$j] <- res_sparse$sum_x
    names(res) <- colnames(x)
    return(res)
  }
)

## rowMeans dbm ####
#' Row (column) means for [`dbMatrix`] objects
#' @inheritParams MatrixGenerics::rowMeans
#' @inherit MatrixGenerics::rowMeans description
#' @param na.rm Always TRUE for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @param dims Always 1 for [`dbMatrix`] queries. Included for compatibility with
#' the generic.
#' @return A named numeric vector containing one mean per row or column of `x`.
#' @concept summary
#' @rdname row_col_means
#' @export
setMethod(
  'rowMeans',
  signature(x = 'dbMatrix'),
  function(x, ...) {
    x <- .castNumeric(x)
    n_cols <- ncol(x)

    # Calculate rowMeans directly: sum(x) per row / n_cols
    res <- rowSums(x) / n_cols
    names(res) <- rownames(x)
    return(res)
  }
)

## colMeans dbm ####
#' @rdname row_col_means
#' @export
setMethod(
  'colMeans',
  signature(x = 'dbMatrix'),
  function(x, ...) {
    x <- .castNumeric(x)
    n_rows <- nrow(x)

    # Calculate colMeans directly: sum(x) per col / n_rows
    res <- colSums(x) / n_rows
    names(res) <- colnames(x)
    return(res)
  }
)

## colSds dbdm ####
#' Row (column) standard deviations for [`dbMatrix`] objects
#' @inherit MatrixGenerics::rowSds description
#' @param x A [`dbMatrix`] object.
#' @param rows Always NULL for [`dbMatrix`] queries. TODO
#' @param cols Always NULL for [`dbMatrix`] queries. TODO
#' @param na.rm Always TRUE for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @param center Always NULL for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @param ... Additional arguments (not used, but included for compatibility with the generic).
#' @param useNames Always TRUE for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @return A named numeric vector containing one sample standard deviation per
#'   row or column of `x`.
#' @concept summary
#' @rdname row_col_sds
#' @export
setMethod(
  "colSds",
  signature(x = "dbDenseMatrix"),
  function(x, ..., useNames = TRUE) {
    x <- .castNumeric(x)
    m <- nrow(x)

    col_stats <- x[] |>
      dplyr::group_by(j) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = TRUE),
        sum_x2 = sum(x * x, na.rm = TRUE),
        .groups = "drop"
      ) |>
      dplyr::mutate(
        sd_x = if (m <= 1) {
          NA_real_
        } else {
          sqrt((sum_x2 - (sum_x * sum_x) / !!m) / (!!m - 1))
        }
      )

    res <- col_stats |>
      dplyr::arrange(j) |>
      dplyr::pull(sd_x)
    if (useNames) {
      names(res) <- colnames(x)
    }
    return(res)
  }
)

## colSds dbsm ####
#' @concept summary
#' @rdname row_col_sds
#' @export
setMethod(
  "colSds",
  signature(x = "dbSparseMatrix"),
  function(x, ..., useNames = TRUE) {
    x <- .castNumeric(x)
    m <- nrow(x)
    num_col <- ncol(x)

    res_sparse <- x[] |>
      dplyr::group_by(j) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = TRUE),
        sum_x2 = sum(x * x, na.rm = TRUE),
        .groups = "drop"
      ) |>
      dplyr::collect()

    # Initialize with 0 if m > 1, else NA
    default_val <- if (m <= 1) NA_real_ else 0
    res <- rep(default_val, num_col)

    # Calculate sd for non-zeros
    if (m > 1) {
      sds <- sqrt(
        (res_sparse$sum_x2 - (res_sparse$sum_x * res_sparse$sum_x) / m) /
          (m - 1)
      )
      res[res_sparse$j] <- sds
    }

    if (useNames) {
      names(res) <- colnames(x)
    }
    return(res)
  }
)

## rowSds dbdm ####
#' @concept summary
#' @rdname row_col_sds
#' @export
setMethod(
  "rowSds",
  signature(x = "dbDenseMatrix"),
  function(
    x,
    rows = NULL,
    cols = NULL,
    na.rm = TRUE,
    center = NULL,
    ...,
    useNames = TRUE
  ) {
    x <- .castNumeric(x)
    k <- ncol(x)

    row_stats <- x[] |>
      dplyr::group_by(i) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = na.rm),
        sum_x2 = sum(x * x, na.rm = na.rm),
        .groups = "drop"
      ) |>
      dplyr::mutate(
        sd_x = if (k <= 1) {
          NA_real_
        } else {
          sqrt((sum_x2 - (sum_x * sum_x) / !!k) / (!!k - 1))
        }
      )

    res <- row_stats |>
      dplyr::arrange(i) |>
      dplyr::pull(sd_x)
    if (useNames) {
      names(res) <- rownames(x)
    }
    return(res)
  }
)

## rowSds dbsm ####
#' @concept summary
#' @rdname row_col_sds
#' @export
setMethod(
  "rowSds",
  signature(x = "dbSparseMatrix"),
  function(
    x,
    rows = NULL,
    cols = NULL,
    na.rm = TRUE,
    center = NULL,
    ...,
    useNames = TRUE
  ) {
    x <- .castNumeric(x)
    k <- ncol(x)
    num_row <- nrow(x)

    res_sparse <- x[] |>
      dplyr::group_by(i) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = na.rm),
        sum_x2 = sum(x * x, na.rm = na.rm),
        .groups = "drop"
      ) |>
      dplyr::collect()

    default_val <- if (k <= 1) NA_real_ else 0
    res <- rep(default_val, num_row)

    if (k > 1) {
      sds <- sqrt(
        (res_sparse$sum_x2 - (res_sparse$sum_x * res_sparse$sum_x) / k) /
          (k - 1)
      )
      res[res_sparse$i] <- sds
    }

    if (useNames) {
      names(res) <- rownames(x)
    }
    return(res)
  }
)

# MatrixGenerics ####
## rowVars dbdm ####
#' Row (column) variances for [`dbMatrix`] objects
#' @inherit MatrixGenerics::rowVars description
#' @param x A [`dbMatrix`] object.
#' @param rows Always NULL for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @param cols Always NULL for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @param na.rm Always TRUE for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @param center Always NULL for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @param ... Additional arguments (not used, but included for compatibility with the generic).
#' @param useNames Always TRUE for [`dbMatrix`] queries. Included for compatibility
#' with the generic.
#' @return A named numeric vector containing one sample variance per row or
#'   column of `x`.
#' @concept summary
#' @rdname row_col_vars
#' @export
setMethod(
  "rowVars",
  signature(x = "dbDenseMatrix"),
  function(
    x,
    rows = NULL,
    cols = NULL,
    na.rm = TRUE,
    center = NULL,
    ...,
    useNames = TRUE
  ) {
    x <- .castNumeric(x)
    k <- ncol(x)

    row_stats <- x[] |>
      dplyr::group_by(i) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = na.rm),
        sum_x2 = sum(x * x, na.rm = na.rm),
        .groups = "drop"
      ) |>
      dplyr::mutate(
        var_x = (sum_x2 - (sum_x * sum_x) / !!k) / (!!k - 1)
      )

    v <- row_stats |>
      dplyr::arrange(i) |>
      dplyr::pull(var_x)
    if (useNames) {
      names(v) <- rownames(x)
    }
    return(v)
  }
)

## rowVars dbsm ####
#' @concept summary
#' @rdname row_col_vars
#' @export
setMethod(
  'rowVars',
  signature(x = 'dbSparseMatrix'),
  function(
    x,
    rows = NULL,
    cols = NULL,
    na.rm = TRUE,
    center = NULL,
    ...,
    useNames = TRUE
  ) {
    x <- .castNumeric(x)
    k <- ncol(x)
    num_row <- nrow(x)

    res_sparse <- x[] |>
      dplyr::group_by(i) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = na.rm),
        sum_x2 = sum(x * x, na.rm = na.rm),
        .groups = "drop"
      ) |>
      dplyr::collect()

    default_val <- if (k <= 1) NA_real_ else 0
    res <- rep(default_val, num_row)

    if (k > 1) {
      vars <- (res_sparse$sum_x2 -
        (res_sparse$sum_x * res_sparse$sum_x) / k) /
        (k - 1)
      res[res_sparse$i] <- vars
    }

    if (useNames) {
      names(res) <- rownames(x)
    }
    return(res)
  }
)

## colVars dbdm ####
#' @rdname row_col_vars
#' @export
setMethod(
  "colVars",
  signature(x = "dbDenseMatrix"),
  function(
    x,
    rows = NULL,
    cols = NULL,
    na.rm = TRUE,
    center = NULL,
    ...,
    useNames = TRUE
  ) {
    x <- .castNumeric(x)
    m <- nrow(x) # total rows

    col_stats <- x[] |>
      dplyr::group_by(j) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = na.rm),
        sum_x2 = sum(x * x, na.rm = na.rm),
        .groups = "drop"
      ) |>
      dplyr::mutate(
        var_x = (sum_x2 - (sum_x * sum_x) / !!m) / (!!m - 1)
      )

    v <- col_stats |>
      dplyr::arrange(j) |>
      dplyr::pull(var_x)
    if (useNames) {
      names(v) <- colnames(x)
    }
    return(v)
  }
)

## colVars dbsm ####
#' @concept summary
#' @rdname row_col_vars
#' @export
setMethod(
  "colVars",
  signature(x = "dbSparseMatrix"),
  function(
    x,
    rows = NULL,
    cols = NULL,
    na.rm = TRUE,
    center = NULL,
    ...,
    useNames = TRUE
  ) {
    x <- .castNumeric(x)
    m <- nrow(x)
    num_col <- ncol(x)

    res_sparse <- x[] |>
      dplyr::group_by(j) |>
      dplyr::summarise(
        sum_x = sum(x, na.rm = na.rm),
        sum_x2 = sum(x * x, na.rm = na.rm),
        .groups = "drop"
      ) |>
      dplyr::collect()

    default_val <- if (m <= 1) NA_real_ else 0
    res <- rep(default_val, num_col)

    if (m > 1) {
      vars <- (res_sparse$sum_x2 -
        (res_sparse$sum_x * res_sparse$sum_x) / m) /
        (m - 1)
      res[res_sparse$j] <- vars
    }

    if (useNames) {
      names(res) <- colnames(x)
    }
    return(res)
  }
)

## mean dbdm####

#' Arithmetic Mean for [`dbMatrix`] objects
#' @inheritParams base::mean
#' @inherit base::mean description
#' @param x [`dbMatrix`] object
#' @return A length-one numeric vector giving the arithmetic mean of all entries
#'   in `x`.
#' @concept summary
#' @rdname mean
#' @export
setMethod('mean', signature(x = 'dbDenseMatrix'), function(x, ...) {
  x <- .castNumeric(x)

  res <- x[] |>
    dplyr::summarise(mean_x = mean(x, na.rm = TRUE)) |>
    dplyr::pull(mean_x)

  return(res)
})

## mean dbsm####
#' @concept summary
#' @rdname mean
#' @export
setMethod('mean', signature(x = 'dbSparseMatrix'), function(x, ...) {
  x <- .castNumeric(x)

  dim <- dim(x)
  n <- dim[1] * dim[2]

  res <- x[] |>
    dplyr::summarise(sum_x = sum(x, na.rm = TRUE)) |>
    dplyr::pull(sum_x)

  res <- res / n

  return(res)
})

# Math ####
## abs ####
## sign ####
## sqrt ####
## ceiling ####
## floor ####
## trunc ####
## cummax ####
## cummin ####
## cumprod ####
## cumsum ####
## log ####
## log10 ####
## log2 ####
## log1p ####
## acos ####
## acosh ####
## asin ####
## asinh ####
## atan ####
## atanh ####
## exp ####
## expm1 ####
## cos ####
## cosh ####
## cospi ####
## sin ####
## sinh ####
## sinpi ####
## tan ####
## tanh ####
## tanpi ####
## gamma ####
## lgamma ####
## digamma ####
## trigamma ####
#' Math Operations for [`dbMatrix`] Objects
#'
#' @description
#' Implements the `Math` [`S4groupGeneric`] functions
#' for [`dbMatrix`] objects. This includes various mathematical operations such as
#' logarithms, exponentials, trigonometric functions, and other transformations.
#'
#' @param x A [`dbMatrix`] object.
#'
#' @details
#' This method provides implementations for the following Math functions:
#'
#' *Arithmetic and rounding*:
#' * `abs()`, `sign()`, `sqrt()`, `ceiling()`, `floor()`, `trunc()`
#'
#' *Cumulative operations*:
#' * `cummax()`, `cummin()`, `cumprod()`, `cumsum()`
#' * **Note: `cumprod()` is not supported**
#'
#' *Logarithmic*:
#' * `log()`, `log10()`, `log2()`, `log1p()`
#'
#' **DuckDB Log Function Mappings**:
#' | R Function | DuckDB Function | Notes |
#' |------------|-----------------|-------|
#' | `log(x)` | `LN(x)` | Natural logarithm |
#' | `log10(x)` | `LOG10(x)` | Base-10 logarithm |
#' | `log2(x)` | `LOG2(x)` | Base-2 logarithm |
#' | `log1p(x)` | `LN(x + 1)` | log(1+x), computed as LN |
#'
#' **Sparsity-Preserving Log**:
#' For `dbSparseMatrix` with pending operations, `log(x + 1)` operations preserve sparsity
#' since `log(0 + 1) = 0`. The multiplicative component is applied first,
#' then the log transformation is applied to sparse values only.
#'
#' *Trigonometric*:
#' * `cos()`, `sin()`, `tan()`, `acos()`, `asin()`, `atan()`
#' * `cosh()`, `sinh()`, `tanh()`, `acosh()`, `asinh()`, `atanh()`
#' * `cospi()`, `sinpi()`, `tanpi()`
#' * **Note: `acosh()` `asinh()` `atanh()` are not supported**
#'
#' *Exponential*:
#' * `exp()`, `expm1()`
#' * **Note: `expm1()` is not supported**
#'
#' *Special functions*:
#' * `gamma()`, `lgamma()`, `digamma()`, `trigamma()`
#' * **Note: `digamma()` `trigamma()` are not supported**
#'
#' The function applies the specified mathematical operation to each element
#' of the [`dbMatrix`] object.
#'
#' @return
#' A [`dbMatrix`] object with the mathematical operation applied to each element.
#'
#' @examples
#' mat <- matrix(1, nrow = 3, ncol = 3)
#' dbmat <- as.dbMatrix(mat)
#' log(dbmat)
#' sqrt(dbmat)
#' sin(dbmat)
#'
#' @concept transform
#' @export
setMethod('Math', signature(x = 'dbMatrix'), function(x) {
  op_name <- as.character(.Generic)

  # SQL translations for R functions that DuckDB doesn't have natively
  # These functions need to be translated to equivalent SQL expressions
  sql_translations <- list(
    log1p = "LN(x + 1)", # log(1+x)
    expm1 = "EXP(x) - 1", # exp(x) - 1
    log = "LN(x)", # R's log() is natural log
    log10 = "LOG10(x)",
    log2 = "LOG2(x)"
  )

  if (op_name %in% names(sql_translations)) {
    # Use the SQL translation
    sql_expr <- sql_translations[[op_name]]
    x[] <- x[] |> dplyr::mutate(x = dplyr::sql(!!sql_expr))
  } else {
    # Standard path: apply function directly
    build_call <- glue::glue(
      "x[] |>
       dplyr::mutate(x = `",
      op_name,
      "`(x))"
    )
    x[] <- eval(str2lang(build_call))
  }

  x@name <- NA_character_
  return(x)
})

# is.na ####
#' Element-wise is.na for dbMatrix
#'
#' @description
#' Returns a dbMatrix with numeric values indicating NA positions (1 = NA, 0 = not NA).
#'
#' @param x A dbMatrix object.
#' @return A dbMatrix with same dimensions, containing 1 where the original
#'   value was NA and 0 otherwise.
#'
#' @examples
#' mat <- matrix(c(1, NA, 3, NA), nrow = 2)
#' dbmat <- as.dbMatrix(mat)
#' is.na(dbmat)
#'
#' @concept transform
#' @export
setMethod('is.na', signature(x = 'dbMatrix'), function(x) {
  x[] <- x[] |>
    dplyr::mutate(x = as.numeric(is.na(x)))
  x@name <- NA_character_
  return(x)
})

# Summary ####
## max ####
## min ####
## range ####
## prod ####
## sum ####
## any ####
## all ####
#' Summary Methods for [`dbMatrix`] Objects
#'
#' @description
#' Implements the [`S4groupGeneric`] group generic functions for dbMatrix objects.
#'
#' @param x A dbMatrix object.
#' @param ... Additional arguments (not used, but included for compatibility with the generic).
#' @param na.rm Logical. If TRUE, remove NA values before computation. Always set to TRUE for this implementation.
#' @concept summary
#' @details
#' This method provides implementations for the following [`S4groupGeneric`] functions:
#' * `max()`: Maximum value
#' * `min()`: Minimum value
#' * `range()`: *Not supported*
#' * `prod()`: Product of all values
#' * `sum()`: Sum of all values
#' * `any()`: Returns TRUE if any value is TRUE
#' * `all()`: Returns TRUE if all values are TRUE
#'
#'
#' @return
#' The result of applying the respective summary function to the dbMatrix object.
#' The type of the return value depends on the specific function called.
#'
#' @examples
#' mat <- matrix(1, nrow = 3, ncol = 3)
#' dbmat <- as.dbMatrix(mat)
#' max(dbmat)
#' min(dbmat)
#' prod(dbmat)
#' sum(dbmat)
#' any(dbmat > 0)
#' all(dbmat > 0)
#'
#' @export
setMethod('Summary', signature(x = 'dbMatrix'), function(x, ..., na.rm = TRUE) {
  # Check to see if 'range' or 'all' is equal to as.character(.Generic)
  if (as.character(.Generic) == 'range') {
    stopf(paste0("range() is not yet supported for dbMatrix objects."))
  }

  if (as.character(.Generic) == 'any' || as.character(.Generic) == 'all') {
    # Always cast to logical for any/all to be safe and avoid eager evaluation
    x[] <- x[] |> dplyr::mutate(x = as.logical(x))
    # warning("coercing argument of type 'double' to logical")
  }

  build_call <- glue::glue(
    "x[] |>
     dplyr::summarise(res := `",
    as.character(.Generic),
    "`(x, na.rm = TRUE)) |>
     dplyr::pull(res)"
  )
  res <- suppressWarnings(eval(str2lang(build_call)))

  # avoid using dplyr::pull()
  # build_call = glue::glue(
  #   "x[] |>
  #    dplyr::summarise(res := `", as.character(.Generic),"`(x, na.rm = TRUE)) |>
  #    dbplyr::sql_render()"
  # )
  # sql <- suppressWarnings(eval(str2lang(build_call)))
  # res <- DBI::dbGetQuery(dbplyr::remote_con(x[]), sql)[[1]] # get value

  if (is.na(res)) {
    #FIXME
    res <- TRUE
  }

  return(res)
})

# General Ops ####

### t ####

#' Matrix Transpose
#' @description Given a [`dbMatrix`] `x`, `t` returns the transpose of `x`.
#' @param x [`dbMatrix`] object
#' @importFrom Matrix t
#' @return [`dbMatrix`] object
#' @concept transform
#' @export
#' @rdname t-dbMatrix
setMethod('t', signature(x = 'dbMatrix'), function(x) {
  x@value <- x@value |> dplyr::select(i = j, j = i, x)
  x@dims <- c(x@dims[[2L]], x@dims[[1L]])
  x@dim_names <- list(x@dim_names[[2L]], x@dim_names[[1L]])

  return(x)
})

### nrow ####

#' The Number of Rows/Columns of a dbMatrix Object
#' @description `nrow` and `ncol` return the number of rows or columns present in `x`.
#' @param x [`dbMatrix`] object
#' @return A length-one integer giving the number of rows or columns in `x`.
#' @concept matrix_props
#' @rdname nrow_ncol
#' @export
nrow.dbMatrix <- function(x) {
  if (is.na(x@dims[1L])) {
    conn <- get_con(x)
    res <- DBI::dbGetQuery(
      conn = conn,
      sprintf('SELECT DISTINCT i from %s', dbplyr::remote_name(x[]))
    )
  } else {
    return(x@dims[1L])
  }

  return(base::nrow(res))
}

### ncol ####

#' @concept matrix_props
#' @rdname nrow_ncol
#' @export
ncol.dbMatrix <- function(x) {
  if (is.na(x@dims[2L])) {
    conn <- get_con(x)
    res <- DBI::dbGetQuery(
      conn = conn,
      sprintf('SELECT DISTINCT j from %s', dbplyr::remote_name(x[]))
    )
  } else {
    return(x@dims[2L])
  }
  return(base::nrow(res))
}

### dim ####

#' Dimensions of an Object
#' @description Retrieve the dimension of an object.
#' @param x [`dbMatrix`] object
#' @return An integer vector of length 2 giving the number of rows and columns
#'   in `x`.
#' @concept matrix_props
#' @export
setMethod('dim', signature(x = 'dbMatrix'), function(x) {
  if (anyNA(x@dims)) {
    return(c(nrow(x), ncol(x)))
  } else {
    res <- x@dims
  }
})

### head ####
#' Return the First or Last Parts of an Object
#' @inherit utils::head description
#' @inheritParams utils::head
#' @return A [`dbMatrix`] object containing the first or last `n` rows of `x`,
#'   with updated dimensions and row names.
#' @concept matrix_props
#' @rdname head_tail
#' @export
setMethod('head', signature(x = 'dbMatrix'), function(x, n = 6L, ...) {
  n_subset <- 1:n
  x[] <- x[] |> dplyr::filter(i %in% n_subset)

  x@dims[1L] <- min(nrow(x), as.integer(n))
  rownames(x) <- head(rownames(x))

  return(x)
})

### tail ####
#' @concept matrix_props
#' @rdname head_tail
#' @export
setMethod('tail', signature(x = 'dbMatrix'), function(x, n = 6L, ...) {
  tail_rownames <- tail(rownames(x))
  n_subset <- (nrow(x) - n):nrow(x)
  x[] <- x[] |> dplyr::filter(i %in% n_subset)

  x@dims[1L] <- min(nrow(x), as.integer(n))
  rownames(x) <- tail_rownames

  return(x)
})

### length ####
# TODO: add value param once setter is created
#' Length of a [`dbMatrix`] Object
#' @concept matrix_props
#' @description Get or set the length of vectors (including lists) and factors,
#' and of any other R object for which a method has been defined.
#' @param x [`dbMatrix`] object
#' @return A length-one integer giving the number of stored elements in `x`.
#' @rdname length
#' @export
setMethod('length', signature(x = 'dbMatrix'), function(x) {
  #FIXME modify for vec_matrix class
  if (!1 %in% dim(x)) {
    # dbMatrix
    res <- prod(dim(x))
    return(res)
  }
  res <- nrow(x)
  return(res)
})

# Column data types ####

#' @name colTypes
#' @title Column data types of dbData objects
#' @description
#' Get the column data types of objects that inherit from \code{'dbData'}
#' @param x dbData data object
#' @param ... additional params to pass
#' @noRd
#' @keywords internal
.colTypes <- function(x, ...) {
  con <- dbplyr::remote_con(x[])
  name <- dbplyr::remote_name(x[])

  # Handle lazy tables
  is_lazy_table <- is.null(name)
  if (is_lazy_table) {
    name <- unique_table_name(prefix = "tmp")
    to_view(x[], name = name)
  }

  # Get schema
  sql <- glue::glue("SELECT column_name, column_type FROM (DESCRIBE {name})")
  dat <- DBI::dbGetQuery(con, sql)
  res <- setNames(tolower(dat$column_type), dat$column_name)

  ## Clean up
  if (is_lazy_table) {
    DBI::dbExecute(con, glue::glue("DROP VIEW IF EXISTS {name}"))
  }

  return(res)
}

## .castNumeric ####

#' @title Set a column to numeric
#' @description
#' Sets a column to numeric type if not already
#' This precaution is to avoid truncation of values
#' @param x dbData object
#' @param col column to cast to numeric (optional for dbMatrix)
#' @param ... additional params to pass
#' @noRd
#' @keywords internal
.castNumeric <- function(x, col = NULL, ...) {
  # For dbMatrix objects, cast the 'x' column
  if (is(x, 'dbMatrix')) {
    if (.colTypes(x)['x'] != 'double') {
      x[] <- x[] |> dplyr::mutate(x := as.double(x))
    }
    return(x)
  }

  # For other dbData objects with specified column
  if (!is.null(col)) {
    if (.colTypes(x)[col] != 'double') {
      sym_col <- dplyr::sym(col)
      x[] <- x[] |> dplyr::mutate(!!sym_col := as.numeric(!!sym_col))
    }
  }

  return(x)
}

#' @describeIn simulate_objects Simulate a duckdb connection dplyr tbl_Pool in memory
#' @keywords internal
sim_duckdb <- function(
  value = datasets::iris,
  name = 'test',
  con = NULL,
  memory = TRUE
) {
  # setup in-memory db
  if (is.null(con)) {
    if (memory) {
      drv <- duckdb::duckdb(dbdir = ':memory:')
    } else {
      # create temporary temp.db file
      temp_file <- tempfile(fileext = '.duckdb')
      drv <- duckdb::duckdb(dbdir = temp_file)
    }

    # create connection
    con <- DBI::dbConnect(drv)
  }

  # check to see if table already exists and if so remove it
  if (DBI::dbExistsTable(con, name)) {
    DBI::dbRemoveTable(con, name)
  }

  # add table to database
  DBI::dbWriteTable(con, name, value, temporary = TRUE)

  # show
  dplyr::tbl(con, name)
}

#' @describeIn simulate_objects Simulate a dgcMatrix
#' @title sim_dgc
#' @keywords internal
#' @details
#' This function generates a simulated sparse matrix (dgCMatrix) with number
#' of rows and columns and sets n_vals random values to a non-zero value.
sim_dgc <- function(num_rows = 50, num_cols = 50, n_vals = 50) {
  # create simulated dgc matrix
  data <- matrix(0, nrow = num_rows, ncol = num_cols)

  # Set n random values to non-zero
  non_zero_indices <- sample.int((num_rows * num_cols), n_vals)
  data[non_zero_indices] <- rnorm(n_vals)

  # Create dumby sparse dgc matrix (column-major)
  mat <- as(data, "dgCMatrix")

  return(mat)
}

#' @describeIn simulate_objects Simulate a dense matrix
#' @title sim_dgc
#' @keywords internal
#' @details
#' This function generates a simulated dense matrix object with a specified number
#' of rows and columns.
sim_denseMat <- function(num_rows = 50, num_cols = 50) {
  # create simulated dense matrix
  mat <- matrix(rnorm(num_rows * num_cols), nrow = num_rows, ncol = num_cols)

  return(mat)
}

#' @describeIn simulate_objects Simulate a duckdb connection dplyr tbl_Pool in memory
#' @title sim_ijx_matrix
#'
#' @details
#' This function generates an ijx representation of a simulated dgCMatrix object
#' with a specified number of rows and columns and sets 50 random values to
#' a non-zero value.
#'
#' @param num_rows The number of rows in the matrix (default: 50)
#' @param num_cols The number of columns in the matrix (default: 50)
#' @param seed_num The seed number for reproducibility (default: 42)
#'
#' @return A dgCMatrix object
#' @keywords internal
sim_ijx_matrix <- function(
  mat_type = NULL,
  num_rows = 50,
  num_cols = 50,
  seed_num = 42
) {
  # check if mat_type is empty
  if (is.null(mat_type)) {
    stop("mat_type must be either 'dense' or 'sparse'")
  }

  # check if num_rows and num_cols are both greater than or equal to 10
  if (num_rows < 10 | num_cols < 10) {
    stop("Number of rows and columns must be at least 10")
  }

  # check if mat_type is either dense or sparse
  if (mat_type != 'dense' & mat_type != 'sparse') {
    stop("mat_type must be either 'dense' or 'sparse'")
  }

  # for reproducibility
  set.seed(seed_num)

  # setup dummy matrix data
  if (mat_type == 'sparse') {
    # Simulate dgcMatrix
    mat <- sim_dgc(num_rows = num_rows, num_cols = num_cols)

    # Create sparse ijx representation
    ijx <- Matrix::summary(mat)
  } else {
    # dense matrix
    # Create dumby dataset
    mat <- matrix(rnorm(num_rows * num_cols), nrow = num_rows, ncol = num_cols)

    # hardcode dimnames for simulated data
    # row_names = as.factor(paste0("row", 1:num_rows))
    # col_names = as.factor(paste0("col", 1:num_cols))
    row_idx <- as.integer(1:num_rows)
    col_idx <- as.integer(1:num_cols)

    # set dimnames for matrix
    # rownames(mat) = row_names
    # colnames(mat) = col_names

    # Create dense ijx representation
    ijx <- dplyr::tibble(
      i = row_idx[row(mat)],
      j = col_idx[col(mat)],
      x = as.numeric(mat)
    )
  }

  return(ijx)
}

#' @describeIn simulate_objects Simulate a dbSparseMatrix in memory
#' @description  Simulate a dbSparseMatrix in memory
#' @export
sim_dbSparseMatrix <- function(
  num_rows = 50,
  num_cols = 50,
  seed_num = 42,
  name = 'sparse_test',
  memory = FALSE
) {
  # check input
  if (num_rows < 10 | num_cols < 10) {
    stop("Number of rows and columns must be at least 10.")
  }

  # simulate ijx matrix
  ijx <- sim_ijx_matrix(
    mat_type = 'sparse',
    num_rows = num_rows,
    num_cols = num_cols,
    seed_num = seed_num
  )

  ijx <- ijx |> as.data.frame()

  # add ijx as table to new duckdb connection
  data <- sim_duckdb(value = ijx, name = name, memory = memory)

  # save connection
  conn <- dbplyr::remote_con(data)

  # setup dimnames
  row_names <- as.factor(paste0("row", 1:num_rows))
  col_names <- as.factor(paste0("col", 1:num_cols))
  dim_names <- list(row_names, col_names)

  # setup dims
  dim <- c(as.integer(num_rows), as.integer(num_cols))

  # create dbSparseMatrix obj
  res <- dbMatrix(
    value = data,
    con = conn,
    name = name,
    dims = dim,
    dim_names = dim_names,
    overwrite = TRUE,
    class = "dbSparseMatrix",
    init = TRUE
  )

  # show
  res
}

#' @describeIn simulate_objects Simulate a dbDenseMatrix in memory
#' @description Simulate a dbDenseMatrix in memory.
#' @export
sim_dbDenseMatrix <- function(
  num_rows = 50,
  num_cols = 50,
  seed_num = 42,
  name = 'dense_test',
  memory = FALSE
) {
  # check input
  if (num_rows < 10 | num_cols < 10) {
    stop("Number of rows and columns must be at least 10.")
  }

  # simulate dense matrix
  data <- sim_ijx_matrix(
    mat_type = 'dense',
    num_rows = num_rows,
    num_cols = num_cols,
    seed_num = seed_num
  )

  # add dense matrix as table to duckdb database
  data <- sim_duckdb(value = data, name = name, memory = memory)

  # get connection
  conn <- dbplyr::remote_con(data)

  # setup dimnames
  row_names <- as.factor(paste0("row", 1:num_rows))
  col_names <- as.factor(paste0("col", 1:num_cols))
  # row_names = rownames(data)
  # col_names = colnames(data)
  dim_names <- list(row_names, col_names)

  # setup dims
  dim <- c(as.integer(num_rows), as.integer(num_cols))

  # create dbDenseMatrix object
  res <- dbMatrix(
    value = data,
    con = conn,
    name = name,
    dims = dim,
    dim_names = dim_names,
    overwrite = TRUE,
    class = "dbDenseMatrix",
    init = TRUE
  )

  # Show
  res
}

test_that("db_svd handles integer-typed x (Fast Path)", {
  testthat::skip_if_not_installed("duckdb")
  testthat::skip_if_not_installed("nanoarrow")

  ijx <- sim_ijx_matrix(mat_type = "sparse", num_rows = 30, num_cols = 40)
  ijx <- as.data.frame(ijx)
  ijx$x <- as.integer(round(abs(ijx$x) * 10))

  tbl <- sim_duckdb(value = ijx, name = "int64_sparse", memory = TRUE)
  con <- dbplyr::remote_con(tbl)

  dbm <- dbMatrix(
    value = tbl,
    con = con,
    name = "int64_sparse",
    dims = c(30L, 40L),
    dim_names = list(
      as.factor(paste0("row", seq_len(30))),
      as.factor(paste0("col", seq_len(40)))
    ),
    overwrite = TRUE,
    class = "dbSparseMatrix",
    init = TRUE
  )

  res <- db_svd(
    dbm = dbm,
    k = 5,
    center = FALSE,
    scale = FALSE,
    memory_limit = 10^12
  )

  testthat::expect_type(res$d, "double")
  testthat::expect_length(res$d, 5)
  testthat::expect_equal(dim(res$u), c(30L, 5L))
  testthat::expect_equal(dim(res$v), c(40L, 5L))
})

test_that("db_svd supports row-wise scaling+centering (GiottoDB PCA case)", {
  testthat::skip_if_not_installed("duckdb")
  testthat::skip_if_not_installed("nanoarrow")
  # This test uses the Fast Path to keep runtime and external dependencies low.

  dbm <- sim_dbSparseMatrix(num_rows = 30, num_cols = 40, memory = TRUE)

  res <- db_svd(
    dbm = dbm,
    k = 5,
    center = TRUE,
    scale = TRUE,
    center_rows = TRUE,
    memory_limit = 10^12
  )

  testthat::expect_type(res$d, "double")
  testthat::expect_length(res$d, 5)
  testthat::expect_equal(dim(res$u), c(30L, 5L))
  testthat::expect_equal(dim(res$v), c(40L, 5L))
})

test_that("db_svd errors on k < 1", {
  testthat::skip_if_not_installed("duckdb")
  testthat::skip_if_not_installed("nanoarrow")

  dbm <- sim_dbSparseMatrix(num_rows = 10, num_cols = 10, memory = TRUE)
  testthat::expect_error(db_svd(dbm, k = 0), "k must be at least 1")
})

test_that("db_svd warns and reduces k when exceeding max allowed", {
  testthat::skip_if_not_installed("duckdb")
  testthat::skip_if_not_installed("nanoarrow")

  dbm <- sim_dbSparseMatrix(num_rows = 10, num_cols = 20, memory = TRUE)
  # max_k = min(10-1, 20) = 9
  testthat::expect_warning(
    db_svd(dbm, k = 15, center = FALSE, scale = FALSE, memory_limit = 10^12),
    "k=15 exceeds max allowed"
  )

  res <- suppressWarnings(
    db_svd(dbm, k = 15, center = FALSE, scale = FALSE, memory_limit = 10^12)
  )
  testthat::expect_length(res$d, 9)
})

test_that("db_svd errors on center_rows=FALSE with centering", {
  testthat::skip_if_not_installed("duckdb")
  testthat::skip_if_not_installed("nanoarrow")

  dbm <- sim_dbSparseMatrix(num_rows = 10, num_cols = 10, memory = TRUE)
  testthat::expect_error(
    db_svd(dbm, k = 5, center = TRUE, center_rows = FALSE),
    "center_rows=FALSE.*not yet supported"
  )
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# Helper to compare results ignoring dimnames differences
expect_equal_values <- function(actual, expected) {
  dimnames(actual) <- NULL
  dimnames(expected) <- NULL
  expect_equal(actual, expected)
}

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
dgc <- dbMatrix:::sim_dgc(3, 5, 10)
mat <- as.matrix(dgc + 1)

dbdm <- as.dbMatrix(mat)

# ---------------------------------------------------------------------------- #
# Test dbMatrix-vector arithmetic
test_that("+ 1 equal", {
  res_mat <- mat + c(1, 2, 3)
  res_dbdm <- dbdm + c(1, 2, 3)
  res_dbdm <- as.matrix(res_dbdm)
  expect_equal_values(res_mat, res_dbdm)
})

test_that("-1 equal", {
  res_mat <- mat - c(1, 2, 3)
  res_dbdm <- dbdm - c(1, 2, 3)
  res_dbdm <- as.matrix(res_dbdm)
  expect_equal_values(res_mat, res_dbdm)
})

test_that("* 10 equal", {
  res_mat <- mat * c(1, 2, 3)
  res_dbdm <- dbdm * c(1, 2, 3)
  res_dbdm <- as.matrix(res_dbdm)
  expect_equal_values(res_mat, res_dbdm)
})

test_that("+0 equal", {
  res_mat <- mat + c(1, 2, 3)
  res_dbdm <- dbdm + c(1, 2, 3)
  res_dbdm <- as.matrix(res_dbdm)
  expect_equal_values(res_mat, res_dbdm)
})

test_that("/10 equal", {
  res_mat <- mat / c(1, 2, 3)
  res_dbdm <- dbdm / c(1, 2, 3)
  res_dbdm <- as.matrix(res_dbdm)
  expect_equal_values(res_mat, res_dbdm)
})

test_that("as.matrix works for dbDenseMatrix", {
  # Setup
  mat <- matrix(1:12, nrow = 4, ncol = 3)
  dimnames(mat) <- list(paste0("r", 1:4), paste0("c", 1:3))
  dbm <- as.dbMatrix(mat)

  # Test default (dense matrix)
  res <- as.matrix(dbm)
  expect_true(is.matrix(res))
  expect_equal(res, mat, ignore_attr = TRUE)

  # Test with names = FALSE
  res_no_names <- as.matrix(dbm, names = FALSE)
  expect_true(is.matrix(res_no_names))
  expect_null(dimnames(res_no_names))
  expect_equal(res_no_names, mat, ignore_attr = TRUE)
})

test_that("as.matrix works for dbSparseMatrix", {
  # Setup
  dgc <- Matrix::rsparsematrix(5, 5, 0.5)
  dimnames(dgc) <- list(paste0("r", 1:5), paste0("c", 1:5))
  dbsm <- as.dbMatrix(dgc)

  # Test default (dense matrix) - defaults to names=TRUE
  res <- as.matrix(dbsm)
  expect_true(is.matrix(res))
  expect_equal(res, as.matrix(dgc), ignore_attr = TRUE)

  # Test sparse = TRUE (dgCMatrix) - defaults to names=TRUE
  res_sparse <- as.matrix(dbsm, sparse = TRUE)
  expect_s4_class(res_sparse, "dgCMatrix")
  expect_equal(res_sparse, dgc, ignore_attr = TRUE)

  # Test with names = FALSE
  res_no_names <- as.matrix(dbsm, names = FALSE)
  expect_null(dimnames(res_no_names))
})

test_that("as.vector works for dbDenseMatrix", {
  # Setup 1D matrix (column vector)
  mat_col <- matrix(1:5, nrow = 5, ncol = 1)
  rownames(mat_col) <- paste0("r", 1:5)
  dbm_col <- as.dbMatrix(mat_col)

  res_col <- as.vector(dbm_col)
  expect_true(is.vector(res_col))
  expect_equal(res_col, as.vector(mat_col), ignore_attr = TRUE)
  expect_equal(names(res_col), rownames(mat_col))

  # Setup 1D matrix (row vector)
  mat_row <- matrix(1:5, nrow = 1, ncol = 5)
  colnames(mat_row) <- paste0("c", 1:5)
  dbm_row <- as.dbMatrix(mat_row)

  res_row <- as.vector(dbm_row)
  expect_true(is.vector(res_row))
  expect_equal(res_row, as.vector(mat_row), ignore_attr = TRUE)
  expect_equal(names(res_row), colnames(mat_row))

  # Fail for 2D matrix
  mat_2d <- matrix(1:4, nrow = 2)
  dbm_2d <- as.dbMatrix(mat_2d)
  expect_error(as.vector(dbm_2d), "Use `as.matrix\\(\\)`")
})

test_that("setAs coercion works", {
  mat <- matrix(1:6, nrow = 2)
  dbm <- as.dbMatrix(mat)

  # as(x, "matrix")
  res_mat <- as(dbm, "matrix")
  expect_true(is.matrix(res_mat))
  expect_equal(res_mat, mat, ignore_attr = TRUE)

  # as(x, "dgCMatrix")
  res_dgc <- as(dbm, "dgCMatrix")
  expect_s4_class(res_dgc, "dgCMatrix")
  expect_equal(as.matrix(res_dgc), mat, ignore_attr = TRUE)

})

test_that("setAs coercion works for simulated tbl-backed dbSparseMatrix", {
  dbsm <- sim_dbSparseMatrix(num_rows = 10, num_cols = 10, memory = TRUE)
  res_dgc <- as(dbsm, "dgCMatrix")
  expect_s4_class(res_dgc, "dgCMatrix")
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")
# Enable densification for tests
options(dbMatrix.allow_densify = TRUE)

# Helper to compare results ignoring dimnames differences
expect_equal_values <- function(actual, expected) {
  if (inherits(actual, "Matrix") || inherits(actual, "sparseMatrix") || 
      inherits(actual, "dgCMatrix")) {
    actual <- as.matrix(actual)
  }
  if (inherits(expected, "Matrix") || inherits(expected, "sparseMatrix") ||
      inherits(expected, "dgCMatrix")) {
    expected <- as.matrix(expected)
  }
  dimnames(actual) <- NULL
  dimnames(expected) <- NULL
  expect_equal(actual, expected)
}

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
dgc <- dbMatrix:::sim_dgc(3, 5, 10)

dbsm <- as.dbMatrix(dgc)
# ---------------------------------------------------------------------------- #
# Test scalar arithmetic
test_that("+ 1 equal", {
  res_dgc <- dgc + c(1, 2, 3)
  res_dgc <- res_dgc |> as.matrix() #dgeMatrix casting
  res_dbsm <- dbsm + c(1, 2, 3)
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("-1 equal", {
  res_dgc <- dgc - c(1, 2, 3)
  res_dgc <- as.matrix(res_dgc) #dgeMatrix casting
  res_dbsm <- dbsm - c(1, 2, 3)
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("* 10 equal", {
  res_dgc <- dgc * c(1, 2, 3)
  res_dbsm <- dbsm * c(1, 2, 3)
  res_dbsm <- as.matrix(res_dbsm, sparse = TRUE)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("+ vector equal", {
  res_dgc <- dgc + c(1, 2, 3)
  res_dgc <- as.matrix(res_dgc) #dgeMatrix casting
  res_dbsm <- dbsm + c(1, 2, 3)
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("/ vector equal", {
  res_dgc <- dgc / c(1, 2, 3)
  res_dbsm <- dbsm / c(1, 2, 3)
  res_dbsm <- as.matrix(res_dbsm, sparse = TRUE)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("writeMM works correctly", {
  skip_if_not_installed("duckdb")

  # Create a simulated dbSparseMatrix
  db_mat <- sim_dbSparseMatrix(10, 10)

  # Test writeMM
  temp_file <- tempfile(fileext = ".mtx")
  writeMM(db_mat, temp_file)

  expect_true(file.exists(temp_file))

  # Check if Matrix::readMM can read it
  m_read <- Matrix::readMM(temp_file)

  # Compare (ignoring dimnames for now as readMM doesn't preserve them fully in this context unless we check)
  # sim_dbSparseMatrix creates random data, so we check dimensions and class
  expect_equal(dim(m_read), dim(db_mat))

  unlink(temp_file)
})

test_that("as.matrix uses writeMM and cleans up", {
  skip_if_not_installed("duckdb")

  db_mat <- sim_dbSparseMatrix(20, 20)

  # Test as.matrix (sparse=TRUE)
  m_sparse <- as.matrix(db_mat, sparse = TRUE, names = TRUE)
  expect_s4_class(m_sparse, "dgCMatrix")
  expect_equal(dim(m_sparse), dim(db_mat))

  # Test as.matrix (sparse=FALSE)
  m_dense <- as.matrix(db_mat, sparse = FALSE, names = TRUE)
  expect_true(is.matrix(m_dense))
  expect_equal(dim(m_dense), dim(db_mat))
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# Enable densification for tests
options(dbMatrix.allow_densify = TRUE)

# ---------------------------------------------------------------------------- #
# Load the dgcMatrix
dbsm <- dbMatrix:::sim_dbSparseMatrix()
dgc <- as(dbsm, "dgCMatrix")
# ---------------------------------------------------------------------------- #

# For table of expected results see `?.eval_op_densify`

test_that(" + 0 equal", {
  res_dgc <- (dgc + 0)
  res_dbsm <- (dbsm + 0) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" + 1 equal", {
  res_dgc <- (dgc + 1) |> as.matrix()
  res_dbsm <- (dbsm + 1) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" + 100 equal", {
  res_dgc <- (dgc + 100) |> as.matrix()
  res_dbsm <- (dbsm + 100) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" + Inf equal", {
  res_dgc <- (dgc + Inf) |> as.matrix()
  res_dbsm <- (dbsm + Inf) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" + NaN equal", {
  res_dgc <- (dgc + NaN) |> as.matrix()
  res_dbsm <- (dbsm + NaN) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" - 0 equal", {
  res_dgc <- (dgc - 0)
  res_dbsm <- (dbsm - 0) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" - 1 equal", {
  res_dgc <- (dgc - 1) |> as.matrix()
  res_dbsm <- (dbsm - 1) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" - 100 equal", {
  res_dgc <- (dgc - 100) |> as.matrix()
  res_dbsm <- (dbsm - 100) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" - Inf equal", {
  res_dgc <- (dgc - Inf) |> as.matrix()
  res_dbsm <- (dbsm - Inf) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" - NaN equal", {
  res_dgc <- (dgc - NaN) |> as.matrix()
  res_dbsm <- (dbsm - NaN) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" * 0 equal", {
  res_dgc <- (dgc * 0)
  res_dbsm <- (dbsm * 0) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" * 1 equal", {
  res_dgc <- (dgc * 1)
  res_dbsm <- (dbsm * 1) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" * 100 equal", {
  res_dgc <- (dgc * 100)
  res_dbsm <- (dbsm * 100) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" * Inf equal", {
  res_dgc <- (dgc * Inf) |> as.matrix()
  res_dbsm <- (dbsm * Inf) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" * NaN equal", {
  res_dgc <- (dgc * NaN) |> as.matrix()
  res_dbsm <- (dbsm * NaN) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" ^ 0 equal", {
  res_dgc <- (dgc^0) |> as.matrix()
  res_dbsm <- (dbsm^0) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" ^ 1 equal", {
  res_dgc <- (dgc^1)
  res_dbsm <- (dbsm^1) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" ^ 100 equal", {
  res_dgc <- (dgc^100)
  res_dbsm <- (dbsm^100) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" ^ Inf equal", {
  res_dgc <- (dgc^Inf)
  res_dbsm <- (dbsm^Inf) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" ^ NaN equal", {
  res_dgc <- (dgc^NaN) |> as.matrix()
  res_dbsm <- (dbsm^NaN) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %% 0 equal", {
  res_dgc <- (dgc %% 0) |> as.matrix()
  res_dbsm <- (dbsm %% 0) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %% 1 equal", {
  res_dgc <- (dgc %% 1)
  res_dbsm <- (dbsm %% 1) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %% 100 equal", {
  res_dgc <- (dgc %% 100)
  res_dbsm <- (dbsm %% 100) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %% Inf equal", {
  res_dgc <- (dgc %% Inf)
  res_dbsm <- (dbsm %% Inf) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %% NaN equal", {
  res_dgc <- (dgc %% NaN) |> as.matrix()
  res_dbsm <- (dbsm %% NaN) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %/% 0 equal", {
  res_dgc <- (dgc %/% 0) |> as.matrix()
  res_dbsm <- (dbsm %/% 0) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %/% 1 equal", {
  res_dgc <- (dgc %/% 1)
  res_dbsm <- (dbsm %/% 1) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %/% 100 equal", {
  res_dgc <- (dgc %/% 100)
  res_dbsm <- (dbsm %/% 100) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %/% Inf equal", {
  res_dgc <- (dgc %/% Inf)
  res_dbsm <- (dbsm %/% Inf) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" %/% NaN equal", {
  res_dgc <- (dgc %/% NaN) |> as.matrix()
  res_dbsm <- (dbsm %/% NaN) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" / 0 equal", {
  res_dgc <- (dgc / 0) |> as.matrix()
  res_dbsm <- (dbsm / 0) |> as.matrix(sparse = FALSE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" / 1 equal", {
  res_dgc <- (dgc / 1)
  res_dbsm <- (dbsm / 1) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" / 100 equal", {
  res_dgc <- (dgc / 100)
  res_dbsm <- (dbsm / 100) |> as.matrix(sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

test_that(" / Inf equal", {
  res_dgc <- (dgc / Inf) |> as.matrix()
  res_dbsm <- (dbsm / Inf) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

test_that(" / NaN equal", {
  res_dgc <- (dgc / NaN) |> as.matrix()
  res_dbsm <- (dbsm / NaN) |> as.matrix()
  expect_equal(res_dgc, res_dbsm)
})

#' @importFrom Matrix writeMM
#' @export
#' @param obj dbMatrix object
#' @param file path to file
#' @param ... additional arguments
#' @rdname dbMatrix-methods
setMethod("writeMM", signature(obj = "dbMatrix"), function(obj, file, ...) {
  # 1. Get dimensions and nnz
  dims <- dim(obj)
  n_rows <- dims[1]
  n_cols <- dims[2]

  # Count non-zeros (this might be expensive if not cached, but necessary for header)
  # For dbSparseMatrix, it's the count of rows in the table.
  # For dbDenseMatrix, it's also the count (since we store triplets).
  # We can use tally() which is fast in DuckDB.
  nnz <- obj[] |>
    dplyr::tally() |>
    dplyr::pull(n)

  # 2. Prepare header
  # Standard Matrix Market header for sparse coordinate real matrix
  header <- c(
    "%%MatrixMarket matrix coordinate real general",
    "%",
    sprintf("%d %d %d", n_rows, n_cols, nnz)
  )

  # 3. Write data using DuckDB COPY
  # We write to a temporary file first because COPY overwrites
  temp_data_file <- tempfile(fileext = ".mtx_part")

  con <- get_con(obj)
  sql <- dbplyr::sql_render(obj[])

  # Ensure Matrix Market triplet column order and types are stable.
  # readMM expects: i (int), j (int), x (numeric) in this order.
  export_sql <- glue::glue(
    "SELECT CAST(i AS BIGINT) AS i, 
     CAST(j AS BIGINT) AS j, 
     CAST(x AS DOUBLE) AS x FROM ({sql}) t"
  )

  # Validate required columns exist before export.
  test_sql <- glue::glue("SELECT * FROM ({sql}) t LIMIT 0")
  test_df <- DBI::dbGetQuery(con, test_sql)
  if (!all(c("i", "j", "x") %in% colnames(test_df))) {
    stopf("writeMM requires triplet columns 'i', 'j', 'x'")
  }

  # DuckDB COPY command
  # We want space delimiter, no header, no quotes
  copy_sql <- glue::glue(
    "COPY ({export_sql}) TO '{temp_data_file}'
    (FORMAT CSV, DELIMITER ' ', HEADER FALSE, QUOTE '');"
  )

  DBI::dbExecute(con, copy_sql)

  # 4. Combine header and data
  # Write header to target file
  writeLines(header, file)

  # Append data
  # Efficiently append using file.append (binary copy)
  # Note: file.append works on file paths
  file.append(file, temp_data_file)

  # 5. Cleanup
  unlink(temp_data_file)

  return(invisible(TRUE))
})

# dbData ####
## Empty ####
### Extract [] ####
# Inherit from dbData class in {dbProject}

### Set [] ####
# Inherit from dbData class in {dbProject}

# dbMatrix ####
## vector indexing ####
### rows only ####
#' Extract or replace values in database-backed matrices
#' @description Methods for subsetting and replacing values in `dbMatrix`
#' objects.
#' @param x A `dbMatrix` object.
#' @param i Row, logical matrix, or matrix-style index.
#' @param j Column index.
#' @param ... Additional arguments.
#' @param drop Ignored; included for matrix API compatibility.
#' @param value Replacement value.
#' @return A subsetted or modified `dbMatrix`, or an extracted vector for
#' matrix-style indexing.
#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[',
  signature(x = 'dbMatrix', i = 'dbIndex', j = 'missing'),
  function(x, i, ..., drop = FALSE) {
    # get dbMatrix info
    con <- get_con(x)
    dim <- dim(x)

    # check inputs
    .check_extract(x = x, i = i, j = NULL, dim = dim)

    # Convert character indices to integers for faster extraction
    if (is.character(i)) {
      int_idx <- match(i, x@dim_names[[1]])
      if (anyNA(int_idx)) {
        bad_names <- i[is.na(int_idx)]
        stop(
          "Invalid row names: ",
          paste(head(bad_names, 5), collapse = ", "),
          if (length(bad_names) > 5) {
            paste0(", ... (", length(bad_names), " total)")
          }
        )
      }
      i <- int_idx
    }

    n_i <- NA_integer_

    if (is.numeric(i)) {
      # Handle negative indices: convert to positive by excluding
      if (any(i < 0)) {
        if (any(i > 0)) {
          stop("Cannot mix positive and negative indices")
        }
        i <- seq_len(dim[1])[i] # Convert negative indices to positive
      }

      if (is.logical(i)) {
        if (length(i) < dim[1]) {
          i <- rep_len(i, dim[1])
        }
        i <- which(i)
      }

      i_idx <- as.integer(i)
      n_i <- length(i_idx)

      # Empty selection: avoid generating invalid SQL like `VALUES )`
      if (n_i == 0L) {
        filter_i <- x@dim_names[[1]][i_idx] # Returns NULL if dim_names[[1]] is NULL
        x[] <- dplyr::filter(x[], FALSE)
        x@dim_names[[1L]] <- if (is.factor(filter_i)) {
          filter_i
        } else {
          as.factor(filter_i)
        }
        x@dims[1L] <- 0L
        x@name <- NA_character_
        return(x)
      }

      # Identity fast path: full set of rows in original order
      if (length(i_idx) == dim[1] && identical(i_idx, seq_len(dim[1]))) {
        return(x)
      }

      # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
      if (length(i_idx) < 2000) {
        values_list <- glue::glue_collapse(
          glue::glue("({seq_along(i_idx)}, {i_idx})"),
          sep = ", "
        )
        sql <- glue::glue(
          "SELECT * FROM (VALUES {values_list}) AS map(new_i, i)"
        )
        map_tbl <- dplyr::tbl(con, dplyr::sql(sql))
      } else {
        req_df <- data.frame(
          new_i = seq_along(i_idx),
          i = i_idx,
          stringsAsFactors = FALSE
        )
        req_tbl_name <- unique_table_name("__dbM_extract_req_i")
        dplyr::copy_to(
          dest = con,
          req_df,
          name = req_tbl_name,
          temporary = TRUE,
          overwrite = TRUE
        )
        map_tbl <- dplyr::tbl(con, req_tbl_name)
      }
      # Handle NULL dim_names: keep NULL or subset existing names
      filter_i <- x@dim_names[[1]][i_idx] # Returns NULL if dim_names[[1]] is NULL
    } else {
      filter_i <- get_dbM_sub_idx(
        index = i,
        dbM_dimnames = x@dim_names,
        dims = 1
      )

      n_i <- length(filter_i)

      # Empty selection: avoid generating invalid SQL like `VALUES )`
      if (n_i == 0L) {
        x[] <- dplyr::filter(x[], FALSE)
        x@dim_names[[1L]] <- if (is.factor(filter_i)) {
          filter_i
        } else {
          as.factor(filter_i)
        }
        x@dims[1L] <- 0L
        x@name <- NA_character_
        return(x)
      }

      # Identity fast path: if this selection is the full set of rows in the
      # original order, subsetting is a no-op and we can avoid building
      # mapping tables and joins.
      orig_i <- x@dim_names[[1]]
      if (!is.null(orig_i)) {
        if (
          length(filter_i) == length(orig_i) &&
            identical(as.character(filter_i), as.character(orig_i))
        ) {
          return(x)
        }
      }

      # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
      if (length(filter_i) < 2000) {
        safe_names <- gsub("'", "''", filter_i, fixed = TRUE)
        values_list <- glue::glue_collapse(
          glue::glue("({seq_along(filter_i)}, '{safe_names}')"),
          sep = ", "
        )
        sql <- glue::glue(
          "SELECT * FROM (VALUES {values_list}) AS map(new_i, rowname)"
        )
        req_tbl <- dplyr::tbl(con, dplyr::sql(sql))
      } else {
        req_df <- data.frame(
          new_i = seq_along(filter_i),
          rowname = as.character(filter_i),
          stringsAsFactors = FALSE
        )
        req_tbl_name <- unique_table_name("__dbM_extract_req_i")
        dplyr::copy_to(
          dest = con,
          req_df,
          name = req_tbl_name,
          temporary = TRUE,
          overwrite = TRUE
        )
        req_tbl <- dplyr::tbl(con, req_tbl_name)
      }

      # Use helper to create dimension mapping
      dim_map_tbl <- store_mapping(
        con = con,
        items = as.character(x@dim_names[[1]]),
        prefix = "__dbM_extract_dim_map_i",
        col_name_in_db = "rowname"
      )

      map_tbl <- req_tbl |>
        dplyr::inner_join(
          dim_map_tbl,
          by = "rowname"
        ) |>
        dplyr::rename(i = idx) |>
        dplyr::select(new_i, i)
    }

    x_tbl <- x[]
    if (exists("i_idx") && length(i_idx) > 0 && length(i_idx) < 2000) {
      x_tbl <- dplyr::filter(x_tbl, i %in% i_idx)
    }
    x[] <- x_tbl |>
      dplyr::inner_join(map_tbl, by = "i") |>
      dplyr::select(i = new_i, j, x)

    # Preserve factor status for dim_names
    x@dim_names[[1L]] <- if (is.factor(filter_i)) {
      filter_i
    } else {
      as.factor(filter_i)
    }
    x@dims[1L] <- n_i
    x@name <- NA_character_

    return(x)
  }
)

### cols only ####
#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[',
  signature(x = 'dbMatrix', i = 'missing', j = 'dbIndex'),
  function(x, j, ..., drop = FALSE) {
    con <- get_con(x)
    dim <- dim(x)

    # check for dims
    .check_extract(x = x, i = NULL, j = j, dim = dim)

    # Convert character indices to integers for faster extraction
    if (is.character(j)) {
      int_idx <- match(j, x@dim_names[[2]])
      if (anyNA(int_idx)) {
        bad_names <- j[is.na(int_idx)]
        stop(
          "Invalid column names: ",
          paste(head(bad_names, 5), collapse = ", "),
          if (length(bad_names) > 5) {
            paste0(", ... (", length(bad_names), " total)")
          }
        )
      }
      j <- int_idx
    }

    n_j <- NA_integer_

    if (is.numeric(j)) {
      # Handle negative indices: convert to positive by excluding
      if (any(j < 0)) {
        if (any(j > 0)) {
          stop("Cannot mix positive and negative indices")
        }
        j <- seq_len(dim[2])[j] # Convert negative indices to positive
      }

      if (is.logical(j)) {
        if (length(j) < dim[2]) {
          j <- rep_len(j, dim[2])
        }
        j <- which(j)
      }

      j_idx <- as.integer(j)
      n_j <- length(j_idx)

      # Empty selection: avoid generating invalid SQL like `VALUES )`
      if (n_j == 0L) {
        filter_j <- x@dim_names[[2]][j_idx] # Returns NULL if dim_names[[2]] is NULL
        x[] <- dplyr::filter(x[], FALSE)
        x@dim_names[[2L]] <- if (is.factor(filter_j)) {
          filter_j
        } else {
          as.factor(filter_j)
        }
        x@dims[2L] <- 0L
        x@name <- NA_character_
        return(x)
      }

      # Identity fast path: full set of columns in original order
      if (length(j_idx) == dim[2] && identical(j_idx, seq_len(dim[2]))) {
        return(x)
      }

      # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
      if (length(j_idx) < 2000) {
        values_list <- glue::glue_collapse(
          glue::glue("({seq_along(j_idx)}, {j_idx})"),
          sep = ", "
        )
        sql <- glue::glue(
          "SELECT * FROM (VALUES {values_list}) AS map(new_j, j)"
        )
        map_tbl <- dplyr::tbl(con, dplyr::sql(sql))
      } else {
        req_df <- data.frame(
          new_j = seq_along(j_idx),
          j = j_idx,
          stringsAsFactors = FALSE
        )
        req_tbl_name <- unique_table_name("__dbM_extract_req_j")
        dplyr::copy_to(
          dest = con,
          req_df,
          name = req_tbl_name,
          temporary = TRUE,
          overwrite = TRUE
        )
        map_tbl <- dplyr::tbl(con, req_tbl_name)
      }
      # Handle NULL dim_names: keep NULL or subset existing names
      filter_j <- x@dim_names[[2]][j_idx] # Returns NULL if dim_names[[2]] is NULL
    } else {
      filter_j <- get_dbM_sub_idx(
        index = j,
        dbM_dimnames = x@dim_names,
        dims = 2
      )

      n_j <- length(filter_j)

      # Empty selection: avoid generating invalid SQL like `VALUES )`
      if (n_j == 0L) {
        x[] <- dplyr::filter(x[], FALSE)
        x@dim_names[[2L]] <- if (is.factor(filter_j)) {
          filter_j
        } else {
          as.factor(filter_j)
        }
        x@dims[2L] <- 0L
        x@name <- NA_character_
        return(x)
      }

      # Identity fast path: if this selection is the full set of columns in the
      # original order, subsetting is a no-op and we can avoid building
      # mapping tables and joins.
      orig_j <- x@dim_names[[2]]
      if (!is.null(orig_j)) {
        if (
          length(filter_j) == length(orig_j) &&
            identical(as.character(filter_j), as.character(orig_j))
        ) {
          return(x)
        }
      }

      # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
      if (length(filter_j) < 2000) {
        safe_names <- gsub("'", "''", filter_j, fixed = TRUE)
        values_list <- glue::glue_collapse(
          glue::glue("({seq_along(filter_j)}, '{safe_names}')"),
          sep = ", "
        )
        sql <- glue::glue(
          "SELECT * FROM (VALUES {values_list}) AS map(new_j, colname)"
        )
        req_tbl <- dplyr::tbl(con, dplyr::sql(sql))
      } else {
        req_df <- data.frame(
          new_j = seq_along(filter_j),
          colname = as.character(filter_j),
          stringsAsFactors = FALSE
        )
        req_tbl_name <- unique_table_name("__dbM_extract_req_j")
        dplyr::copy_to(
          dest = con,
          req_df,
          name = req_tbl_name,
          temporary = TRUE,
          overwrite = TRUE
        )
        req_tbl <- dplyr::tbl(con, req_tbl_name)
      }

      # Use helper to create dimension mapping
      dim_map_tbl <- store_mapping(
        con = con,
        items = as.character(x@dim_names[[2]]),
        prefix = "__dbM_extract_dim_map_j",
        col_name_in_db = "colname"
      )

      map_tbl <- req_tbl |>
        dplyr::inner_join(
          dim_map_tbl,
          by = "colname"
        ) |>
        dplyr::rename(j = idx) |>
        dplyr::select(new_j, j)
    }

    x_tbl <- x[]
    if (exists("j_idx") && length(j_idx) > 0 && length(j_idx) < 2000) {
      x_tbl <- dplyr::filter(x_tbl, j %in% j_idx)
    }
    x[] <- x_tbl |>
      dplyr::inner_join(map_tbl, by = "j") |>
      dplyr::select(i, j = new_j, x)

    # Preserve factor status for dim_names
    x@dim_names[[2L]] <- if (is.factor(filter_j)) {
      filter_j
    } else {
      as.factor(filter_j)
    }
    x@dims[2L] <- n_j
    x@name <- NA_character_

    return(x)
  }
)

### rows and cols ####
#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[',
  signature(x = 'dbMatrix', i = 'dbIndex', j = 'dbIndex'),
  function(x, i, j, ..., drop = FALSE) {
    # get dbMatrix info
    con <- get_con(x)
    dim <- dim(x)

    # check for dims
    .check_extract(x = x, i = i, j = j, dim = dim)

    # Convert character indices to integers for faster extraction
    if (is.character(i)) {
      int_idx <- match(i, x@dim_names[[1]])
      if (anyNA(int_idx)) {
        bad_names <- i[is.na(int_idx)]
        stop(
          "Invalid row names: ",
          paste(head(bad_names, 5), collapse = ", "),
          if (length(bad_names) > 5) {
            paste0(", ... (", length(bad_names), " total)")
          }
        )
      }
      i <- int_idx
    }

    if (is.character(j)) {
      int_idx <- match(j, x@dim_names[[2]])
      if (anyNA(int_idx)) {
        bad_names <- j[is.na(int_idx)]
        stop(
          "Invalid column names: ",
          paste(head(bad_names, 5), collapse = ", "),
          if (length(bad_names) > 5) {
            paste0(", ... (", length(bad_names), " total)")
          }
        )
      }
      j <- int_idx
    }

    # Process i index (same logic as row-only subsetting)
    if (is.numeric(i)) {
      # Handle negative indices: convert to positive by excluding
      if (any(i < 0)) {
        if (any(i > 0)) {
          stop("Cannot mix positive and negative indices")
        }
        i <- seq_len(dim[1])[i] # Convert negative indices to positive
      }

      if (is.logical(i)) {
        if (length(i) < dim[1]) {
          i <- rep_len(i, dim[1])
        }
        i <- which(i)
      }

      filter_i <- x@dim_names[[1]][i] # Returns NULL if dim_names[[1]] is NULL

      # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
      if (length(i) == 0L) {
        map_tbl_i <- dplyr::tbl(
          con,
          dplyr::sql("SELECT 1 AS new_i, 1 AS i WHERE 0=1")
        )
      } else if (length(i) < 2000) {
        values_list <- glue::glue_collapse(
          glue::glue("({seq_along(i)}, {as.integer(i)})"),
          sep = ", "
        )
        sql <- glue::glue(
          "SELECT * FROM (VALUES {values_list}) AS map(new_i, i)"
        )
        map_tbl_i <- dplyr::tbl(con, dplyr::sql(sql))
      } else {
        req_df <- data.frame(
          new_i = seq_along(i),
          i = as.integer(i),
          stringsAsFactors = FALSE
        )
        req_tbl_name <- unique_table_name("__dbM_extract_req_i")
        dplyr::copy_to(
          dest = con,
          req_df,
          name = req_tbl_name,
          temporary = TRUE,
          overwrite = TRUE
        )
        map_tbl_i <- dplyr::tbl(con, req_tbl_name)
      }
    } else {
      filter_i <- get_dbM_sub_idx(
        index = i,
        dbM_dimnames = x@dim_names,
        dims = 1
      )

      if (length(filter_i) == 0L) {
        map_tbl_i <- dplyr::tbl(
          con,
          dplyr::sql("SELECT 1 AS new_i, 1 AS i WHERE 0=1")
        )
      } else {
        # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
        if (length(filter_i) < 2000) {
          safe_names <- gsub("'", "''", filter_i, fixed = TRUE)
          values_list <- glue::glue_collapse(
            glue::glue("({seq_along(filter_i)}, '{safe_names}')"),
            sep = ", "
          )
          sql <- glue::glue(
            "SELECT * FROM (VALUES {values_list}) AS map(new_i, rowname)"
          )
          req_tbl <- dplyr::tbl(con, dplyr::sql(sql))
        } else {
          req_df <- data.frame(
            new_i = seq_along(filter_i),
            rowname = as.character(filter_i),
            stringsAsFactors = FALSE
          )
          req_tbl_name <- unique_table_name("__dbM_extract_req_i")
          dplyr::copy_to(
            dest = con,
            req_df,
            name = req_tbl_name,
            temporary = TRUE,
            overwrite = TRUE
          )
          req_tbl <- dplyr::tbl(con, req_tbl_name)
        }

        # Use helper to create dimension mapping
        dim_map_tbl <- store_mapping(
          con = con,
          items = as.character(x@dim_names[[1]]),
          prefix = "__dbM_extract_dim_map_i",
          col_name_in_db = "rowname"
        )

        map_tbl_i <- req_tbl |>
          dplyr::inner_join(
            dim_map_tbl,
            by = "rowname"
          ) |>
          dplyr::rename(i = idx) |>
          dplyr::select(new_i, i)
      }
    }

    # Process j index (same logic as column-only subsetting)
    if (is.numeric(j)) {
      # Handle negative indices: convert to positive by excluding
      if (any(j < 0)) {
        if (any(j > 0)) {
          stop("Cannot mix positive and negative indices")
        }
        j <- seq_len(dim[2])[j] # Convert negative indices to positive
      }

      if (is.logical(j)) {
        if (length(j) < dim[2]) {
          j <- rep_len(j, dim[2])
        }
        j <- which(j)
      }

      filter_j <- x@dim_names[[2]][j] # Returns NULL if dim_names[[2]] is NULL

      # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
      if (length(j) == 0L) {
        map_tbl_j <- dplyr::tbl(
          con,
          dplyr::sql("SELECT 1 AS new_j, 1 AS j WHERE 0=1")
        )
      } else if (length(j) < 2000) {
        values_list <- glue::glue_collapse(
          glue::glue("({seq_along(j)}, {as.integer(j)})"),
          sep = ", "
        )
        sql <- glue::glue(
          "SELECT * FROM (VALUES {values_list}) AS map(new_j, j)"
        )
        map_tbl_j <- dplyr::tbl(con, dplyr::sql(sql))
      } else {
        req_df <- data.frame(
          new_j = seq_along(j),
          j = as.integer(j),
          stringsAsFactors = FALSE
        )
        req_tbl_name <- unique_table_name("__dbM_extract_req_j")
        dplyr::copy_to(
          dest = con,
          req_df,
          name = req_tbl_name,
          temporary = TRUE,
          overwrite = TRUE
        )
        map_tbl_j <- dplyr::tbl(con, req_tbl_name)
      }
    } else {
      filter_j <- get_dbM_sub_idx(
        index = j,
        dbM_dimnames = x@dim_names,
        dims = 2
      )

      if (length(filter_j) == 0L) {
        map_tbl_j <- dplyr::tbl(
          con,
          dplyr::sql("SELECT 1 AS new_j, 1 AS j WHERE 0=1")
        )
      } else {
        # <2000 use inline SQL, >=2000 use register (avoids massive SQL strings)
        if (length(filter_j) < 2000) {
          safe_names <- gsub("'", "''", filter_j, fixed = TRUE)
          values_list <- glue::glue_collapse(
            glue::glue("({seq_along(filter_j)}, '{safe_names}')"),
            sep = ", "
          )
          sql <- glue::glue(
            "SELECT * FROM (VALUES {values_list}) AS map(new_j, colname)"
          )
          req_tbl <- dplyr::tbl(con, dplyr::sql(sql))
        } else {
          req_df <- data.frame(
            new_j = seq_along(filter_j),
            colname = as.character(filter_j),
            stringsAsFactors = FALSE
          )
          req_tbl_name <- unique_table_name("__dbM_extract_req_j")
          dplyr::copy_to(
            dest = con,
            req_df,
            name = req_tbl_name,
            temporary = TRUE,
            overwrite = TRUE
          )
          req_tbl <- dplyr::tbl(con, req_tbl_name)
        }

        # Use helper to create dimension mapping
        dim_map_tbl <- store_mapping(
          con = con,
          items = as.character(x@dim_names[[2]]),
          prefix = "__dbM_extract_dim_map_j",
          col_name_in_db = "colname"
        )

        map_tbl_j <- req_tbl |>
          dplyr::inner_join(
            dim_map_tbl,
            by = "colname"
          ) |>
          dplyr::rename(j = idx) |>
          dplyr::select(new_j, j)
      }
    }

    x[] <- x[] |>
      dplyr::inner_join(map_tbl_i, by = "i") |>
      dplyr::inner_join(map_tbl_j, by = "j") |>
      dplyr::select(i = new_i, j = new_j, x)

    # update dbMatrix attributes - preserve factor status
    x@dim_names[[1L]] <- if (is.factor(filter_i)) {
      filter_i
    } else {
      as.factor(filter_i)
    }
    x@dim_names[[2L]] <- if (is.factor(filter_j)) {
      filter_j
    } else {
      as.factor(filter_j)
    }
    x@dims[1L] <- if (!is.null(filter_i)) length(filter_i) else length(i)
    x@dims[2L] <- if (!is.null(filter_j)) length(filter_j) else length(j)
    x@name <- NA_character_

    return(x)
  }
)

#' @description
#' Internal function to index `dbMatrix` objects by `dbIndex` superclass.
#' Can apply to both rows (dims = 1) and columns (dims = 2)
#' @keywords internal
#' @noRd
get_dbM_sub_idx <- function(index, dbM_dimnames, dims) {
  # check that idx is 1 or 2
  if (dims != 1 && dims != 2) {
    stop("dims must be 1 (rows) or 2 (columns)")
  }
  dims <- as.integer(dims)

  if (is.character(index)) {
    return(index)
  }

  if (is.logical(index)) {
    index <- recycle_boolean_index(index, length(dbM_dimnames[[dims]]))
  }

  if (is(index, 'dbDenseMatrix')) {
    dimensions <- dim(index)

    # check that dimensions has 1 in the [1] or [2] position
    if (sum(dimensions == 1) != 1) {
      stop("dbDenseMatrix is not a 1D dbMatrix")
    }

    # FIXME: If dbmatrix@x is logical support recycle_boolean_index
    # below only supports character indexing, no recycling
    is_logical <- index@value |> head(1) |> dplyr::pull(x) |> is.logical()
    sub_names <- dbM_dimnames[[dims]]
    if (is_logical) {
      filtered_index <- index[] |>
        dplyr::filter(x)
      if (dims == 1L || is(index, "dbDenseMatrix")) {
        #FIXME 1D dbMatrix
        # a_rownames |>
        #    dplyr::semi_join(filtered_index, by = "i")
        index <- filtered_index |>
          dplyr::arrange(i) |>
          dplyr::distinct(i) |>
          dplyr::pull(i)
        return(sub_names[index])
      } else {
        index <- filtered_index |>
          dplyr::arrange(j) |>
          dplyr::distinct(j) |>
          dplyr::pull(j)
        return(sub_names[index])
      }
    }

    if (dimensions[1] <= length(dbM_dimnames[[dims]])) {
      index <- rownames(index)
    } else if (dimensions[2] <= length(dbM_dimnames[[dims]])) {
      index <- colnames(index)
    }

    if (all(index %in% dbM_dimnames[[dims]])) {
      return(index)
    } else {
      stop("1D dbMatrix dimensions do not match dbMatrix dimensions")
    }
  }

  sub_names <- dbM_dimnames[[dims]]
  return(sub_names[index])
}

#' @noRd
#' @keywords internal
recycle_boolean_index <- function(index, length) {
  #FIXME: 1D dbMatrix
  if (is.logical(index) && length(index) < length) {
    recycled <- rep_len(index, length)
    return(which(recycled))
  }
  return(index)
}

#' @noRd
#' @keywords internal
.check_extract <- function(x = x, i = NULL, j = NULL, dim) {
  if (!is.null(j)) {
    if (
      (is.numeric(j) || is.logical(j)) &&
        length(j) > 0L &&
        !all(is.na(j)) &&
        max(j, na.rm = TRUE) > dim[2]
    ) {
      stopf("Index exceeds column dimension of", dim[2])
    } else if (is.character(j) && !all(j %in% colnames(x))) {
      missing_cols <- j[!j %in% colnames(x)]
      stopf("Column(s) not found in dbMatrix: \n", missing_cols)
    }
  }

  if (!is.null(i)) {
    if (
      (is.numeric(i) || is.logical(i)) &&
        length(i) > 0L &&
        !all(is.na(i)) &&
        max(i, na.rm = TRUE) > dim[1]
    ) {
      stopf("Index exceeds row dimension of", dim[1])
    } else if (is.character(i) && !all(i %in% rownames(x))) {
      missing_rows <- i[!i %in% rownames(x)]
      stopf("Row(s) not found in dbMatrix: \n", missing_rows)
    }
  }
}

## matrix index ####
#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[',
  signature(x = 'dbMatrix', i = 'dbMatrix', j = 'missing'),
  function(x, i, ..., drop = FALSE) {
    # Check dimensions match
    if (!identical(dim(x), dim(i))) {
      stop("Matrix index 'i' must have the same dimensions as 'x'")
    }

    # Perform join and filter
    res_tbl <- dplyr::inner_join(
      x[],
      i[],
      by = c("i", "j"),
      suffix = c(".x", ".i")
    ) |>
      dplyr::filter(x.i) |>
      dplyr::arrange(j, i)

    # Pull the result into memory as a vector
    result_vector <- res_tbl |>
      dplyr::pull(x.x)

    # Note: Base R doesn't preserve names in this type of subsetting
    return(result_vector)
  }
)


# *** matrix index assignment ####
#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[<-',
  signature(x = 'dbMatrix', i = 'dbMatrix', j = 'missing', value = 'ANY'),
  function(x, i, j, value) {
    if (!identical(dim(x), dim(i))) {
      stop("Logical index matrix 'i' must have the same dimensions as 'x'")
    }

    # Ensure value is scalar for this implementation
    if (length(value) != 1) {
      stop(
        "Replacement value must be a single scalar for matrix indexing assignment"
      )
    }

    con <- get_con(x)
    tbl_x <- x[]
    tbl_i <- i[]

    # Identify rows in tbl_x to update based on tbl_i
    rows_to_update <- dplyr::inner_join(
      tbl_x |> dplyr::select(i, j),
      tbl_i,
      by = c("i", "j")
    ) |>
      dplyr::filter(x) |>
      dplyr::select(i, j)

    update_data <- rows_to_update |>
      dplyr::mutate(new_value = value)

    # Use rows_update for efficient update based on keys (i, j)
    update_data <- update_data |> dplyr::rename(x = new_value)

    # Perform the update
    updated_tbl <- dplyr::rows_update(
      tbl_x,
      update_data,
      by = c("i", "j"),
      unmatched = "ignore" # Keep rows from x that weren't updated
    )

    # Update the dbMatrix object's value slot
    x@value <- updated_tbl
    x@name <- NA_character_

    return(x)
  }
)

# dbDenseMatrix indexing methods ####
# These methods enable 1D dbMatrix logical indexing for dbMatrix objects

#' Internal helper to store mapping table for subsetting
#' Uses inline SQL for small N, registered virtual table (view) for large N
#' Fallback to registered view ensures efficiency and clean database state
#' @keywords internal
#' @noRd
store_mapping <- function(con, items, prefix, col_name_in_db) {
  # Inline SQL
  if (length(items) < 2000) {
    safe_items <- gsub("'", "''", items, fixed = TRUE)
    items_sql <- glue::glue_collapse(glue::glue("'{safe_items}'"), sep = ", ")

    mapping_sql <- glue::glue(
      "SELECT ROW_NUMBER() OVER () as idx, unnest as {col_name_in_db} 
       FROM (SELECT UNNEST([{items_sql}]) as unnest)"
    )
    return(dplyr::tbl(con, dplyr::sql(glue::glue("({mapping_sql})"))))
  }

  # Registered virtual table
  df <- data.frame(
    idx = seq_along(items),
    name = items,
    stringsAsFactors = FALSE
  )
  colnames(df)[2] <- col_name_in_db

  tbl_name <- unique_table_name(prefix)
  dplyr::copy_to(
    dest = con,
    df,
    name = tbl_name,
    temporary = TRUE,
    overwrite = TRUE
  )
  return(dplyr::tbl(con, tbl_name))
}

#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[',
  signature(x = 'dbMatrix', i = 'dbDenseMatrix', j = 'missing'),
  function(x, i, ..., drop = FALSE) {
    # Check if i is a valid 1D dbMatrix (1D matrix)
    if (!1 %in% dim(i)) {
      stopf("dbDenseMatrix index must be a 1D dbMatrix (have 1 in dimensions)")
    }

    # Check dimensional compatibility - only row indexing supported
    if (dim(i)[1] != nrow(x)) {
      stopf(
        "Row indexing: dbDenseMatrix must have same number of rows as matrix"
      )
    }

    # Convert dbDenseMatrix to logical vector for indexing
    logical_index <- i[] |>
      dplyr::arrange(i) |>
      dplyr::pull(x) |>
      as.logical()

    # Use the existing dbIndex row indexing method
    return(x[logical_index, , drop = drop])
  }
)

#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[',
  signature(x = 'dbMatrix', i = 'missing', j = 'dbDenseMatrix'),
  function(x, i, j, ..., drop = FALSE) {
    # Column indexing with dbDenseMatrix not supported due to dimensional constraints
    stopf(
      "Column indexing with dbDenseMatrix not supported. Use as.vector(j) to convert to logical vector."
    )
  }
)

#' @rdname extract-dbMatrix
#' @concept dbMatrix
#' @export
setMethod(
  '[',
  signature(x = 'dbMatrix', i = 'dbDenseMatrix', j = 'dbDenseMatrix'),
  function(x, i, j, ..., drop = FALSE) {
    # Dual indexing with dbDenseMatrix not supported due to dimensional constraints
    stopf(
      "Dual indexing with dbDenseMatrix not supported. Convert indices using as.vector() first."
    )
  }
)

#' Input validation for data arg
#' @param value A `Matrix`, `matrix`, or `tbl_duckdb_connection` object
#' @return No return value. Called for input validation and throws an error if
#'   `value` is invalid.
#' @keywords internal
.check_value <- function(value) {
  if (is.character(value)) {
    if (!file.exists(value)) {
      stopf('File does not exist. Please provide a valid file path.')
    }
    return(invisible(NULL))
  }
  is_valid <- inherits(value, c('Matrix', 'matrix', 'tbl_duckdb_connection')) ||
    is.null(value)

  if (!is_valid) {
    stopf('Invalid "value" input passed.')
  }
}

#' @keywords internal
#' @noRd
.get_dbMatrix_info <- function(db_sparse) {
  dims <- dim(db_sparse)
  list(
    con = dbplyr::remote_con(db_sparse[]),
    dims = dims,
    dim_names = dimnames(db_sparse),
    remote_name = db_sparse@name,
    n_rows = bit64::as.integer64(dims[1]),
    n_cols = bit64::as.integer64(dims[2])
  )
}

#' @keywords internal
#' @noRd
.attach_precomp_db <- function(con) {
  db_path <- getOption("dbMatrix.precomp_db")
  if (is.null(db_path)) {
    return(FALSE)
  }

  if (!file.exists(db_path)) {
    cli::cli_alert_warning(
      "Option 'dbMatrix.precomp_db' is set but file not found: {db_path}"
    )
    return(FALSE)
  }

  # Check if already attached
  dbs <- DBI::dbGetQuery(con, "PRAGMA database_list")
  if ("dbmatrix_precomp" %in% dbs$name) {
    return(TRUE)
  }

  tryCatch(
    {
      DBI::dbExecute(
        con,
        glue::glue("ATTACH '{db_path}' AS dbmatrix_precomp (READ_ONLY)")
      )
      return(TRUE)
    },
    error = function(e) {
      cli::cli_alert_warning("Failed to attach precomputed DB: {e$message}")
      return(FALSE)
    }
  )
}

#' @keywords internal
#' @noRd
.find_precompute_table <- function(con, n_rows, n_cols) {
  # Convert to numeric to avoid integer64 overflow in comparisons
  n_rows <- as.numeric(n_rows)
  n_cols <- as.numeric(n_cols)

  .attach_precomp_db(con)

  query <- "SELECT table_catalog, table_schema, table_name FROM information_schema.tables WHERE table_name LIKE 'precomp_%'"

  tables <- tryCatch(
    DBI::dbGetQuery(con, query),
    error = function(e) NULL
  )

  if (is.null(tables) || nrow(tables) == 0) {
    return(NULL)
  }

  tables$full_name <- paste(
    tables$table_catalog,
    tables$table_schema,
    tables$table_name,
    sep = "."
  )

  dims_str <- gsub("precomp_", "", tables$table_name, fixed = TRUE)
  dims_list <- strsplit(dims_str, "x")

  valid_fmt <- lengths(dims_list) == 2
  if (!any(valid_fmt)) {
    return(NULL)
  }

  candidates <- tables[valid_fmt, ]
  dims_list <- dims_list[valid_fmt]
  dim_matrix <- do.call(rbind, lapply(dims_list, function(x) as.numeric(x)))

  # Check normal orientation
  valid_normal <- dim_matrix[, 1] >= n_rows & dim_matrix[, 2] >= n_cols
  # Check transposed orientation
  valid_transposed <- dim_matrix[, 1] >= n_cols & dim_matrix[, 2] >= n_rows

  if (!any(valid_normal) && !any(valid_transposed)) {
    return(NULL)
  }

  # Find best fit across both orientations
  best_table <- NULL
  best_dist <- Inf
  transposed <- FALSE

  if (any(valid_normal)) {
    m <- dim_matrix[valid_normal, , drop = FALSE]
    c <- candidates[valid_normal, ]
    d <- abs(m[, 1] - n_rows) + abs(m[, 2] - n_cols)
    idx <- which.min(d)
    if (d[idx] < best_dist) {
      best_dist <- d[idx]
      best_table <- c$full_name[idx]
      transposed <- FALSE
    }
  }

  if (any(valid_transposed)) {
    m <- dim_matrix[valid_transposed, , drop = FALSE]
    c <- candidates[valid_transposed, ]
    d <- abs(m[, 1] - n_cols) + abs(m[, 2] - n_rows)
    idx <- which.min(d)
    if (d[idx] < best_dist) {
      best_dist <- d[idx]
      best_table <- c$full_name[idx]
      transposed <- TRUE
    }
  }

  if (is.null(best_table)) {
    return(NULL)
  }

  # Validate structure
  tryCatch(
    {
      cols <- DBI::dbGetQuery(con, glue::glue("DESCRIBE {best_table}"))
      if (!all(c("i", "j") %in% cols$column_name)) {
        cli::cli_alert_warning(
          "Found candidate table '{best_table}' but it lacks 'i' or 'j' columns. Skipping."
        )
        return(NULL)
      }
    },
    error = function(e) NULL
  )

  list(name = best_table, transposed = transposed)
}

#' @description
#' This function will create a dense COO table if one does not already exist
#' or if the existing table is not large enough.
#' @keywords internal
#' @noRd
.initialize_precompute_matrix <- function(con, n_rows, n_cols) {
  precomp_result <- .find_precompute_table(con, n_rows, n_cols)

  if (is.null(precomp_result)) {
    cli::cli_alert_info(
      "Computing new dense COO table with {n_rows} rows and {n_cols} columns..."
    )
    return(precompute(conn = con, m = n_rows, n = n_cols))
  }

  precompute_name <- precomp_result$name
  use_transposed <- precomp_result$transposed

  precomp <- dplyr::tbl(con, precompute_name)

  # Convert to numeric for dplyr filter compatibility
  nr <- as.numeric(n_rows)
  nc <- as.numeric(n_cols)

  if (use_transposed) {
    precomp <- precomp |>
      dplyr::rename(i_orig = i, j_orig = j) |>
      dplyr::rename(i = j_orig, j = i_orig) |>
      dplyr::filter(i <= !!nr, j <= !!nc)
  } else {
    precomp <- precomp |>
      dplyr::filter(i <= !!nr, j <= !!nc)
  }

  return(precomp)
}

#' Compute a dense COO table in a database connection
#'
#' @param conn duckdb database connection
#' @param m number of rows of precomputed dbMatrix table
#' @param n number of columns of precomputed dbMatrix table
#' @param verbose logical, print progress messages. default: FALSE.
#' @description
#' Precomputes a COO list table in a specificied database connection in column-
#' major order.
#' This can speed up operations that involve breaking
#' sparsity of a \code{dbSparseMatrix},
#' such as in cases when performing + or - arithmetic operations.
#'
#' @details
#' The \code{m} and \code{n} parameters must exceed the
#' maximum row and column indices of the \code{dbMatrix} in order to be used for
#' densifying any \code{dbMatrix}. If these params are less than the maximum
#' row and column indices, a new precomputed table will be automatically
#' generated with the name 'precomp_mXn'.
#'
#' In such cases, run this function again with a larger
#' \code{n_rows} and \code{num_cols}, or to manually remove the precomputed
#' table set \code{options(dbMatrix.precomp = NULL)} in the R console.
#'
#' @return A `tbl_dbi` object referencing the newly created precomputed lookup
#'   table in DuckDB.
#' @keywords internal
#' @concept dbMatrix
precompute <- function(conn, m, n, verbose = FALSE) {
  .check_con(conn = conn)

  if (!(is.numeric(m)) || !(is.numeric(n))) {
    stop("m and n must be integers or numerics")
  }

  n_rows <- bit64::as.integer64(m)
  n_cols <- bit64::as.integer64(n)
  total <- n_rows * n_cols
  name <- paste0("precomp_", n_rows, "x", n_cols)

  # Use BIGINT if needed
  int32_limit <- bit64::as.integer64(2147483647)
  index_type <- if (n_rows > int32_limit | n_cols > int32_limit) {
    "BIGINT"
  } else {
    "INT"
  }

  # Generate grid using parquet for efficient storage/compression
  # Note: implicit order by j,i for downstream operations
  sql <- glue::glue(
    "
    COPY (
      SELECT
        CAST(((row_id.generate_series - 1) % {n_rows} + 1) AS {index_type}) AS i,
        CAST(FLOOR((row_id.generate_series - 1) / {n_rows}) + 1 AS {index_type}) AS j,
        row_id.generate_series - 1 AS idx
      FROM generate_series(1, {total}) AS row_id
    )
    TO '{name}.parquet' (FORMAT PARQUET, ROW_GROUP_SIZE 1000000, COMPRESSION LZ4_RAW);
  "
  )
  invisible(DBI::dbExecute(conn, sql))

  # Load back as table
  invisible(DBI::dbExecute(
    conn,
    glue::glue(
      "CREATE OR REPLACE TABLE {name} AS SELECT * FROM read_parquet('{name}.parquet');"
    )
  ))
  file.remove(paste0(name, ".parquet"))

  if (verbose) {
    cat(
      glue::glue(
        "Precomputed tbl '{name}' with {n_rows} rows and {n_cols} columns"
      ),
      "\n"
    )
  }

  return(dplyr::tbl(conn, name))
}

# initialize dbMatrix####

# Method for initializing dbMatrix. Concerns only the processing that is related
# to elements internal to the object.

# For pre-object construction data operations/massaging, see the constructor
# function dbMatrix()
#' @keywords internal
#' @noRd
setMethod(
  f = 'initialize',
  signature(.Object = 'dbMatrix'),
  function(.Object, dim_names, dims, ...) {
    # call dbMatrix initialize
    .Object <- methods::callNextMethod(.Object, ...)

    # matrix specific data input #
    # -------------------------- #
    if (!missing(dim_names)) {
      .Object@dim_names <- dim_names
    }
    if (!missing(dims)) {
      .Object@dims <- dims
    }

    # default values if no input provided #
    # ----------------------------------- #
    if (is.null(.Object@value)) {
      .Object@dims <- c(0L, 0L)
      .Object@dim_names <- list(NULL, NULL)
    }
    # tbl_name = dbplyr::remote_name(.Object[])
    # .Object@name = ifelse(is.null(tbl_name), NA_character_, tbl_name)

    # check and return #
    # ---------------- #
    validObject(.Object)
    return(.Object)
  }
)

# show ####

## dbDenseMatrix ####
setMethod("show", signature("dbDenseMatrix"), function(object) {
  row_names <- rownames(object)
  col_names <- colnames(object)
  dims <- dim(object)
  dim_row <- dims[1]
  dim_col <- dims[2]

  # print class and dims #
  # -------------------- #

  if (identical(dims, c(0L, 0L))) {
    cat('0 x 0 matrix of class "dbDenseMatrix"\n')
    return()
  } else if (ncol(object) == 1) {
    cat(dim_row, 'x 1 dbMatrix of class "dbDenseMatrix"\n')
  } else {
    cat(dim_row, "x", dim_col, ' dbMatrix of class "dbDenseMatrix"\n')
  }

  # preview print #
  # ------------- #

  # print colnames
  colname_show_n <- dim_col - 6L
  if (ncol(object) == 1) {
    colname_show_n <- 0L
  } else if (colname_show_n < 0L) {
    message("[[ Colnames: ", vector_to_string(col_names), " ]]")
  } else if (colname_show_n >= 1L) {
    message(
      "[[ Colnames ",
      vector_to_string(head(col_names, 3L)),
      " ... suppressing ",
      colname_show_n,
      " ...",
      vector_to_string(tail(col_names, 3L)),
      " ]]"
    )
  }

  # get matrix i and j to print
  suppress_rows <- FALSE # flag for whether rows are being suppressed

  # Determine column indices to show (first 10)
  if (dim_col > 10L) {
    filter_j <- seq_len(10L)
    p_coln <- col_names[filter_j]
  } else {
    filter_j <- seq_len(dim_col)
    p_coln <- col_names
  }

  # Determine row indices to show (head 3, tail 3 if large)
  if (dim_row > 6L) {
    filter_i <- c(seq_len(3L), seq(from = dim_row - 2L, to = dim_row))
    p_rown <- row_names[filter_i]
    suppress_rows <- TRUE
  } else {
    filter_i <- seq_len(dim_row)
    p_rown <- row_names
  }

  # prepare subset to print
  if (suppress_rows) {
    # Head query
    df_head <- object@value |>
      dplyr::filter(j <= 10L, i <= 3L) |>
      dplyr::arrange(i, j) |>
      dplyr::collect()

    # Tail query
    df_tail <- object@value |>
      dplyr::filter(j <= 10L, i >= (dim_row - 2L)) |>
      dplyr::arrange(i, j) |>
      dplyr::collect()

    preview_tbl <- dplyr::bind_rows(df_head, df_tail)
  } else {
    preview_tbl <- object@value |>
      dplyr::filter(j <= 10L) |>
      dplyr::arrange(i, j) |>
      head(100L) |>
      dplyr::collect()
  }

  # ij indices for printing
  a_i <- match(preview_tbl$i, filter_i)
  a_j <- match(preview_tbl$j, filter_j)

  if (length(a_i) == 0L) {
    a_i <- NULL
  }
  if (length(a_j) == 0L) {
    a_j <- NULL
  }
  a_x <- NULL

  if (length(preview_tbl$x) != 0L) {
    # catch sparse case where if/else: null
    n_digits <- getOption("dbMatrix.digits", default = 7)
    a_x <- format(round(preview_tbl$x, n_digits), nsmall = n_digits)
  }

  # print matrix values
  if (suppress_rows) {
    # suppressed lines: capture, split, then print individually
    # when suppressed, currently hardcoded to show 3 from head and 3 from tail
    a_out <- capture.output(print_array(
      i = a_i,
      j = a_j,
      x = a_x,
      dims = c(length(p_rown), length(p_coln)),
      class = "dense",
      rownames = p_rown
    ))
    writeLines(a_out[1:4])

    dim_col_out <- dim_col - 10L
    dim_row_out <- dim_row - 6L

    if (ncol(object) == 1) {
      sprintf("\n...suppressing %d elements\n\n", dim_row_out) |> cat()
    } else if (dim_col_out < 0) {
      sprintf("\n......suppressing %d rows\n\n", dim_row_out) |> cat()
    } else if (dim_col_out == 0) {
      sprintf("\n...... suppressing %d rows ......\n\n", dim_row_out) |>
        cat()
    } else {
      sprintf(
        "\n......suppressing %d columns and %d rows\n\n",
        dim_col_out,
        dim_row_out
      ) |>
        cat()
    }

    writeLines(a_out[5:7])
  } else {
    # no suppressed lines: Directly print
    print_array(
      i = a_i,
      j = a_j,
      x = a_x,
      dims = c(length(p_rown), length(p_coln)),
      class = "dense",
      rownames = p_rown
    )
  }
})

##  dbSparseMatrix ####
setMethod("show", signature("dbSparseMatrix"), function(object) {
  row_names <- rownames(object)
  col_names <- colnames(object)
  dims <- dim(object)
  dim_row <- dims[1]
  dim_col <- dims[2]

  # print class and dims #
  # -------------------- #

  if (identical(dims, c(0L, 0L))) {
    cat('0 x 0 dbMatrix of class "dbSparseMatrix"\n')
    return()
  } else {
    cat(dim_row, "x", dim_col, ' dbMatrix of class "dbSparseMatrix"\n')
  }

  # preview print #
  # ------------- #

  # print colnames
  colname_show_n <- dim_col - 6L
  if (colname_show_n < 0L) {
    message("[[ Colnames: ", vector_to_string(col_names), " ]]")
  } else if (colname_show_n >= 1L) {
    message(
      "[[ Colnames ",
      vector_to_string(head(col_names, 3L)),
      " ... suppressing ",
      colname_show_n,
      " ...",
      vector_to_string(tail(col_names, 3L)),
      " ]]"
    )
  }

  # get matrix i and j to print
  suppress_rows <- FALSE # flag for whether rows are being suppressed

  # Determine column indices to show (first 10)
  if (dim_col > 10L) {
    filter_j <- seq_len(10L)
    p_coln <- col_names[filter_j]
  } else {
    filter_j <- seq_len(dim_col)
    p_coln <- col_names
  }

  # Determine row indices to show (head 3, tail 3 if large)
  if (dim_row > 6L) {
    filter_i <- c(seq_len(3L), seq(from = dim_row - 2L, to = dim_row))
    p_rown <- row_names[filter_i]
    suppress_rows <- TRUE
  } else {
    filter_i <- seq_len(dim_row)
    p_rown <- row_names
  }

  # prepare subset to print
  if (suppress_rows) {
    # Head query
    df_head <- object@value |>
      dplyr::filter(j <= 10L, i <= 3L) |>
      dplyr::arrange(i, j) |>
      dplyr::collect()

    # Tail query
    df_tail <- object@value |>
      dplyr::filter(j <= 10L, i >= (dim_row - 2L)) |>
      dplyr::arrange(i, j) |>
      dplyr::collect()

    preview_tbl <- dplyr::bind_rows(df_head, df_tail)
  } else {
    preview_tbl <- object@value |>
      dplyr::filter(j <= 10L) |>
      dplyr::arrange(i, j) |>
      head(100L) |>
      dplyr::collect()
  }

  # ij indices for printing
  a_i <- match(preview_tbl$i, filter_i)
  a_j <- match(preview_tbl$j, filter_j)

  if (length(a_i) == 0L) {
    a_i <- NULL
  }
  if (length(a_j) == 0L) {
    a_j <- NULL
  }
  a_x <- NULL

  if (length(preview_tbl$x) != 0L) {
    # catch sparse case where if/else: null
    n_digits <- getOption("dbMatrix.digits", default = 7)
    a_x <- format(round(preview_tbl$x, n_digits), nsmall = n_digits)
  }

  # print matrix values
  if (suppress_rows) {
    # suppressed lines: capture, split, then print individually
    # when suppressed, currently hardcoded to show 3 from head and 3 from tail
    a_out <- capture.output(print_array(
      i = a_i,
      j = a_j,
      x = a_x,
      dims = c(length(p_rown), length(p_coln)),
      class = "sparse",
      rownames = p_rown
    ))
    writeLines(a_out[1:4])

    dim_col_out <- dim_col - 10L
    dim_row_out <- dim_row - 6L

    if (dim_col_out < 0) {
      sprintf("\n......suppressing %d rows\n\n", dim_row_out) |>
        cat()
    } else if (dim_col_out == 0) {
      sprintf("\n.......... suppressing %d rows ..........\n\n", dim_row_out) |>
        cat()
    } else {
      sprintf(
        "\n......suppressing %d columns and %d rows\n\n",
        dim_col_out,
        dim_row_out
      ) |>
        cat()
    }

    writeLines(a_out[5:7])
  } else {
    # no suppressed lines: Directly print
    print_array(
      i = a_i,
      j = a_j,
      x = a_x,
      dims = c(length(p_rown), length(p_coln)),
      class = "sparse",
      rownames = p_rown
    )
  }
})

# constructors ####

#' @title Create a \code{dbSparseMatrix} or \code{dbDenseMatrix} object
#' @description
#' Create an S4 \code{dbMatrix} object in sparse or dense triplet vector format.
#' @param value data to be added to the database. See details for supported data types \code{(required)}
#' @param name table name to assign within database \code{(required, default: "dbMatrix")}
#' @param con DBI or duckdb connection object \code{(required)}
#' @param overwrite whether to overwrite if table already exists in database \code{(required)}
#' @param class class of the dbMatrix: \code{dbDenseMatrix} or \code{dbSparseMatrix} \code{(required)}
#' @param dims dimensions of the matrix \code{(optional: [int, int])}
#' @param dim_names dimension names of the matrix \code{(optional: list(enum, enum))}
#' @param mtx_rowname_file_path path to .mtx rowname file to be read into \code{(optional)}
#' database. by default, no header is assumed.
#' @param mtx_rowname_col_idx column index of row name file \code{(optional)}
#' @param mtx_colname_file_path path to .mtx colname file to be read into
#' database. by default, no header is assumed. \code{(optional)}
#' @param mtx_colname_col_idx column index of column name file \code{(optional)}
#' @param ... additional params to pass to \code{dplyr::copy_to}
#' @details This function reads in data into a pre-existing DuckDB database.
#' Supported \code{value} data types:
#' \itemize{
#'  \item [`Matrix::dgCMatrix-class`] In-memory sparse matrix from the [`Matrix`] package
#'  \item [`Matrix::dgTMatrix-class`] In-memory triplet vector or COO matrix
#'  \item [`matrix`] In-memory dense matrix from base R
#'  \item \code{.mtx} Path to [.mtx](https://math.nist.gov/MatrixMarket/formats.html) file
#'  \item \code{.csv} Path to .csv file
#'  \item `tbl_duckdb_connection` Table in [`duckdb`] database in ijx format from
#'  existing [`dbMatrix`] object. \code{dims} and \code{dim_names} must be
#'  specified if \code{value} is \code{tbl_duckdb_connection}.
#' }
#' @return A [`dbDenseMatrix`] or [`dbSparseMatrix`] object, depending on
#'   `class`, pointing to matrix data stored in DuckDB. The object records the
#'   matrix dimensions and dimension names.
#' @concept dbMatrix
#' @export
#' @examples
#' dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))
#' con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
#' dbSparse <- dbMatrix(
#'   value = dgc,
#'   con = con,
#'   name = "sparse_matrix",
#'   class = "dbSparseMatrix",
#'   overwrite = TRUE
#' )
#' dbSparse
dbMatrix <- function(
  value,
  class = NULL,
  con = NULL,
  overwrite = FALSE,
  name = "dbMatrix",
  dims = NULL,
  dim_names = NULL,
  mtx_rowname_file_path,
  mtx_rowname_col_idx = 1,
  mtx_colname_file_path,
  mtx_colname_col_idx = 1,
  ...
) {
  # check inputs
  .check_value(value)
  .check_con(con)
  .check_name(name)

  if (!inherits(value, "tbl_duckdb_connection")) {
    .check_overwrite(
      conn = con,
      overwrite = overwrite,
      name = name,
      skip_value_check = TRUE
    )
  }

  # check class
  if (is.null(class)) {
    stopf("Invalid class: choose 'dbDenseMatrix' or 'dbSparseMatrix'")
  }
  if (
    !is.character(class) || !(class %in% c("dbDenseMatrix", "dbSparseMatrix"))
  ) {
    stopf("Invalid class: choose 'dbDenseMatrix' or 'dbSparseMatrix'")
  }

  # check value and class mismatch
  if (
    (inherits(value, "matrix") || inherits(value, "denseMatrix")) &&
      class == "dbSparseMatrix"
  ) {
    stopf("Class mismatch: set class to 'dbDenseMatrix' for dense matrices")
  }
  if (
    (inherits(value, "dgCMatrix") || inherits(value, "sparseMatrix")) &&
      class == "dbDenseMatrix"
  ) {
    stopf("Class mismatch: set class to 'dbSparseMatrix' for sparse matrices")
  }

  # check dims, dim_names
  if (inherits(value, "tbl_duckdb_connection")) {
    if (is.null(dims) || is.null(dim_names)) {
      stop(
        "Invalid dims or dim_names: must be provided for tbl_duckdb_connection objects"
      )
    }
  }

  # initialize data (value)
  data <- NULL

  if (inherits(value, "tbl_duckdb_connection")) {
    # data is already in DB
    data <- value
    dims <- dims
    dim_names <- dim_names
  } else {
    # data must be read in
    if (is.character(value)) {
      # read in from file
      if (grepl("\\.csv|\\.tsv|\\.txt", value)) {
        stop("File type not yet supported. Please use .mtx file format.")
        # TODO: implement dense to sparse conversion. Long to wide pivot not yet
        #       supported for out of memory in duckdb.
      } else if (grepl("\\.mtx", value)) {
        data <- readMM(
          con = con,
          value = value,
          name = name,
          overwrite = overwrite
        )

        dims <- get_MM_dim(value)

        dim_names <- get_MM_dimnames(
          mtx_file_path = value,
          mtx_rowname_file_path = mtx_rowname_file_path,
          mtx_rowname_col_idx = mtx_rowname_col_idx,
          mtx_colname_file_path = mtx_colname_file_path,
          mtx_colname_col_idx = mtx_colname_col_idx
        )
      } else {
        stop(
          "Invalid file type. Please provide a .mtx, .csv, .txt, or .tsv file."
        )
      }
    } else if (inherits(value, "matrix") || inherits(value, "Matrix")) {
      return(as.dbMatrix(
        value,
        con = con,
        name = name,
        overwrite = overwrite,
        ...
      ))
    } else {
      stopf('Invalid "value" provided. See ?dbMatrix for help.')
    }
  }

  # Set dimnames if not provided
  if (is.null(unlist(dim_names))) {
    row_names <- as.factor(paste0("row", 1:dims[1]))
    col_names <- as.factor(paste0("col", 1:dims[2]))
    dim_names <- list(row_names, col_names)
  }

  if (class == "dbSparseMatrix") {
    set_class <- "dbSparseMatrix"
  } else if (class == "dbDenseMatrix") {
    set_class <- "dbDenseMatrix"
  } else {
    ## redundant check from above
    stopf("Please specify dbMatrix class: 'dbDenseMatrix' or 'dbSparseMatrix'")
  }

  res <- new(
    Class = set_class,
    value = data,
    name = name,
    init = TRUE,
    dim_names = dim_names,
    dims = dims
  )

  return(res)
}

# converters ####

#' Convert a dbSparseMatrix to dbDenseMatrix
#' @description Internal function to convert a [`dbSparseMatrix`] to
#' [`dbDenseMatrix`].
#' @param x A [`dbSparseMatrix`] object
#' @param chunk_size integer. Number of columns to process per chunk during densification.
#' If NULL (default), the function first checks the global option `dbMatrix.chunk_size`.
#' If that is also NULL, it calculates a chunk size such that the estimated memory usage
#' of each chunk does not exceed `dbMatrix.max_mem_convert` (default 8GB).
#' If the total size is within the limit, a single chunk is used.
#' @return A [`dbDenseMatrix`] object
#' @keywords internal
.to_db_dense <- function(x, chunk_size = NULL) {
  if (!inherits(x, "dbSparseMatrix")) {
    stopf("Input must be a dbSparseMatrix object")
  }

  # Guard: Check if densification is allowed
  if (!getOption("dbMatrix.allow_densify", default = FALSE)) {
    stop(
      "Automatic sparse-to-dense conversion is disabled.\n",
      "To enable, review the documentation: ?dbMatrix_options\n",
      "Then set: options(dbMatrix.allow_densify = TRUE)",
      call. = FALSE
    )
  }

  info <- .get_dbMatrix_info(x)
  con <- info$con
  n_cols <- as.integer(info$n_cols)[1]
  n_rows <- as.integer(info$n_rows)[1]
  verbose <- getOption("dbMatrix.verbose", default = TRUE)

  # 1. Hot Path: Use Precomputed Table
  precomp_result <- .find_precompute_table(con, n_rows, n_cols)
  if (!is.null(precomp_result)) {
    precomp_name <- precomp_result$name
    use_transposed <- precomp_result$transposed

    if (verbose) {
      msg <- if (use_transposed) {
        "Using precomputed table '{precomp_name}' for densification (transposed)."
      } else {
        "Using precomputed table '{precomp_name}' for densification."
      }
      cli::cli_alert_info(msg)
    }

    precomp <- dplyr::tbl(con, precomp_name)

    if (use_transposed) {
      precomp <- precomp |>
        dplyr::rename(i_orig = i, j_orig = j) |>
        dplyr::rename(i = j_orig, j = i_orig) |>
        dplyr::filter(i <= !!n_rows, j <= !!n_cols)
    } else {
      precomp <- precomp |>
        dplyr::filter(i <= !!n_rows, j <= !!n_cols)
    }

    x_tbl <- x[]

    if ("idx" %in% colnames(precomp) && !use_transposed) {
      x_tbl <- x_tbl |> dplyr::mutate(idx = (j - 1) * !!n_rows + (i - 1))
      data <- precomp |>
        dplyr::left_join(x_tbl, by = "idx", suffix = c("", ".dbsm"))
    } else {
      data <- precomp |>
        dplyr::left_join(x_tbl, by = c("i", "j"), suffix = c("", ".dbsm"))
    }

    val_col <- if ("x.dbsm" %in% colnames(data)) "x.dbsm" else "x"

    res <- new(
      "dbDenseMatrix",
      value = data |>
        dplyr::transmute(i, j, x = dplyr::coalesce(!!dplyr::sym(val_col), 0)),
      name = unique_table_name(prefix = "tmp_dbDenseMatrix_hot"),
      dims = info$dims,
      dim_names = info$dim_names,
      init = TRUE
    )
    return(res)
  }

  # 2. Cold Path: JIT Densification
  if (verbose) {
    cli::cli_alert_info(
      "Performing on-the-fly densification (cold path). See ?dbMatrix_options for details."
    )
  }

  if (is.null(chunk_size)) {
    chunk_size <- getOption("dbMatrix.chunk_size")
    if (is.null(chunk_size)) {
      limit <- getOption("dbMatrix.max_mem_convert", default = 8 * 1024^3)
      chunk_size <- max(
        1L,
        min(floor(limit / (as.numeric(n_rows) * 8)), n_cols)
      )
    }
  }

  # Cap chunks to avoid parser limits
  MAX_CHUNKS <- getOption("dbMatrix.max_chunks", default = 10000L)
  if (ceiling(n_cols / chunk_size) > MAX_CHUNKS) {
    chunk_size <- ceiling(n_cols / MAX_CHUNKS)
    cli::cli_alert_warning(
      "Adjusted chunk_size to {chunk_size} to limit query complexity."
    )
  }

  col_starts <- seq(1, as.integer(n_cols), by = chunk_size)
  base_sql <- dbplyr::sql_render(x[])

  queries <- lapply(col_starts, function(start) {
    end <- min(start + chunk_size - 1, n_cols)
    grid_sql <- glue::glue(
      "SELECT i, j FROM range(1, {n_rows} + 1) t1(i) CROSS JOIN range({start}, {end} + 1) t2(j)"
    )
    glue::glue(
      "SELECT grid.i, grid.j, COALESCE(CAST(data.x AS DOUBLE), 0.0) as x FROM ({grid_sql}) grid LEFT JOIN ({base_sql}) data ON grid.i = data.i AND grid.j = data.j"
    )
  })

  .combine_tree <- function(qs) {
    if (length(qs) == 0) {
      return(NULL)
    }
    if (length(qs) == 1) {
      return(qs[[1]])
    }
    mid <- floor(length(qs) / 2)
    glue::glue(
      "({.combine_tree(qs[1:mid])}) UNION ALL ({.combine_tree(qs[(mid + 1):length(qs)])})"
    )
  }

  new(
    "dbDenseMatrix",
    value = dplyr::tbl(con, dplyr::sql(.combine_tree(queries))),
    name = NA_character_,
    dims = info$dims,
    dim_names = info$dim_names,
    init = TRUE
  )
}

#' @description
#' Convert a dbDenseMatrix to a dbSparseMatrix on disk using SQL.
#' @param db_dense dbDenseMatrix object to convert to dbSparseMatrix
#' @noRd
#' @keywords internal
.to_db_sparse <- function(db_dense) {
  stopf("Not yet supported")
}

# Create ijx vector representation of sparse matrix, keeping zeros
# Updates dgcmatrix by reference
# Copied from below:
# https://stackoverflow.com/questions/64473488/melting-a-sparse-matrix-dgcmatrix-and-keeping-its-zeros
#' @noRd
get_dense_ijx_dt <- function(x) {
  dplyr::tibble(
    i = rownames(x)[row(x)],
    j = colnames(x)[col(x)],
    x = as.numeric(x)
  )
}

#' to_ijx_disk
#'
#' @param con duckdb connection
#' @param name name of table to convert to ijx on disk
#'
#' @return remote table in long format unpivoted from wide format matrix
#' @keywords internal
to_ijx_disk <- function(con, name) {
  stopf("Not yet supported")
}

## as.matrix ####
#' Convert [`dbMatrix`] to in-memory matrix
#'
#' @param x A [`dbMatrix`] object (dbSparseMatrix or dbDenseMatrix)
#' @param ... Additional arguments (not used)
#' @param sparse Logical indicating if the output should be a sparse matrix \code{default:FALSE}
#' @param names Logical indicating if the output should have dimnames. \code{default:FALSE}
#' @description
#' Converts a [`dbMatrix`] object into an in-memory matrix or sparse matrix.
#'
#' @details
#' This method converts a [`dbMatrix`] object into an in-memory
#' [`Matrix::dgCMatrix-class`] (sparse = TRUE) or `matrix()` (default, sparse = FALSE).
#'
#' **Warning: This function can cause memory issues for large [`dbMatrix`] objects.**
#'
#' Set `sparse = TRUE` to convert to a sparse matrix.
#' Set `names = TRUE` to keep dimnames.
#'
#' @return A [`Matrix::dgCMatrix-class`] or [`matrix`]
#' @export
#' @concept dbMatrix
#' @method as.matrix dbMatrix
#' @export
as.matrix.dbMatrix <- function(x, ..., sparse = FALSE, names = TRUE) {
  dims <- dim(x)
  n_rows <- dims[1]
  n_cols <- dims[2]
  dim_names <- dimnames(x)

  # checks - only block 1x1 (scalar), allow 1xN and Nx1 matrices
  if (all(dims == 1)) {
    stopf("Use `as.vector()` for scalar (1x1) dbMatrix objects")
  }

  if (dims[1] > 1e5 || dims[2] > 1e5) {
    cli::cli_alert_warning(
      "Warning: Converting large dbMatrix to in-memory Matrix."
    )
  }

  if (is(x, "dbDenseMatrix") && sparse) {
    stopf("Cannot convert dbDensematrix into sparse matrix. Set sparse=FALSE")
  }

  if (is(x, "dbSparseMatrix") && !sparse) {
    cli::cli_alert_info(
      "Converting dbSparseMatrix into dense matrix. Set 'sparse=TRUE' to construct 'dgCMatrix'."
    )
  }

  # Chunked dense matrix conversion to avoid memory inflation
  if (!sparse) {
    # Limit max memory usage
    .check_mem_limit(x)

    # Calculate memory limits
    limit <- getOption("dbMatrix.max_mem_convert", default = 8 * 1024^3)
    est_final_size <- as.numeric(n_rows) * as.numeric(n_cols) * 8

    # Intermediate df is ~16 bytes per element (4+4+8)
    # Use 24 to be safe and account for vectors
    est_intermediate_size <- as.numeric(n_rows) * as.numeric(n_cols) * 24
    est_peak_memory <- est_final_size + est_intermediate_size

    # Pre-allocate dense matrix with zeros
    mat <- matrix(0, nrow = n_rows, ncol = n_cols)

    if (est_peak_memory < limit) {
      if (getOption("dbMatrix.verbose", default = TRUE)) {
        cli::cli_alert_info("Using fast in-memory conversion.")
      }

      # Get all sparse data first
      con <- dbplyr::remote_con(x[])
      sql <- dbplyr::sql_render(x[])
      dat <- DBI::dbGetQuery(con, sql)

      # Fill sparse values into dense matrix
      if (nrow(dat) > 0) {
        idx <- (as.integer(dat$j) - 1L) *
          as.numeric(n_rows) +
          as.integer(dat$i)
        mat[idx] <- dat$x
      }
    } else {
      # Single-fetch using collect() is faster than LIMIT/OFFSET chunking
      dat <- dplyr::collect(x[])

      if (nrow(dat) > 0) {
        idx <- (as.integer(dat$j) - 1L) *
          as.numeric(n_rows) +
          as.integer(dat$i)
        mat[idx] <- dat$x
      }
    }

    if (names) {
      dimnames(mat) <- dim_names
    } else {
      dimnames(mat) <- NULL
    }
    return(mat)
  }

  # Build sparse matrix directly from i/j/x triplets.
  # This avoids Matrix Market scan/parsing issues and keeps conversion sparse.
  dat <- x[] |>
    dplyr::select(i, j, x) |>
    dplyr::collect()

  i_vals <- suppressWarnings(as.integer(dat$i))
  j_vals <- suppressWarnings(as.integer(dat$j))

  bad_i <- !is.na(dat$i) & is.na(i_vals)
  bad_j <- !is.na(dat$j) & is.na(j_vals)
  if (any(bad_i) || any(bad_j)) {
    stopf("Invalid sparse indices: non-integer i/j values encountered during sparse coercion")
  }

  mat <- Matrix::sparseMatrix(
    i = i_vals,
    j = j_vals,
    x = as.numeric(dat$x),
    dims = c(n_rows, n_cols)
  )

  if (names) {
    dimnames(mat) <- dim_names
  }

  return(as(mat, "CsparseMatrix"))
}

#' @method as.matrix dbSparseMatrix
#' @export
as.matrix.dbSparseMatrix <- function(x, ...) {
  as.matrix.dbMatrix(x, ...)
}

#' @method as.matrix dbDenseMatrix
#' @export
as.matrix.dbDenseMatrix <- function(x, ...) {
  as.matrix.dbMatrix(x, ...)
}

#' @noRd
#' @keywords internal
#' @param x dbDenseMatrix containing 1 in dim
# NOTE: 1D dbMatrix objects (dim has 1) can be converted to vector
setMethod(
  "as.vector",
  signature(x = "dbDenseMatrix"),
  function(x, mode = "any") {
    if (1 %in% dim(x)) {
      if (dim(x)[1] == 1) {
        out <- x[] |>
          dplyr::select(j, x) |>
          dplyr::arrange(j) |>
          dplyr::collect() |>
          dplyr::pull(x)
        names(out) <- colnames(x)
      } else {
        out <- x[] |>
          dplyr::select(i, x) |>
          dplyr::arrange(i) |>
          dplyr::collect() |>
          dplyr::pull(x)
        names(out) <- rownames(x)
      }
      return(out)
    }

    stopf("Use `as.matrix()` for dbMatrix objects")
  }
)

#' Convert [`Matrix`] to [`dbMatrix`]
#' @description
#' Converts in-memory [`matrix`], [`Matrix::dgeMatrix-class`], or
#' [`Matrix::dgCMatrix-class`] into a [`dbMatrix`] object.
#'
#' @details
#' If no `con` is provided, a temporary in-memory database connection is created.
#' If no `name` is provided, a unique table name is generated.
#'
#' @param x [`matrix`], [`Matrix::dgeMatrix-class`], or [`Matrix::dgCMatrix-class`] \code{required}
#' @param con `tbl_duckdb_connection` \code{default:"memory"} Connection to
#' DuckDB database connection. If not provided, a temporary in-memory
#' DuckDB database is created. \code{':temp:'}  will create a DuckDB database
#' in the temporary directory. \code{':memory:'} will create a DuckDB database
#' in memory.
#' @param name \code{default:"memory"} table name in the database. If not
#' provided, a unique table name is generated.
#' @param ... Additional arguments passed to [`dbMatrix`]
#' @export
#' @concept dbMatrix
as.dbMatrix <- function(x, con, name, ...) {
  # checks
  stopifnot(is(x, "matrix") | is(x, "dgeMatrix") | is(x, "dgCMatrix"))
  if (missing(con) || (is.character(con) && con == ":memory:")) {
    con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
    cli::cli_alert_info("Creating in-memory database connection...")
  } else if (is.character(con) && con == ":temp:") {
    con <- DBI::dbConnect(
      duckdb::duckdb(),
      tempfile(
        tmpdir = tempdir(),
        fileext = ".db"
      )
    )
    cli::cli_alert_info("Creating temporary database connection...")
  } else {
    stopifnot(inherits(con, "duckdb_connection"))
  }

  if (missing(name)) {
    name <- unique_table_name("dbMatrix")
  } else {
    stopifnot(is.character(name))
  }

  # Convert matrix into dbMatrix
  if (inherits(x, "matrix") || inherits(x, "dgeMatrix")) {
    class <- "dbDenseMatrix"
  } else if (inherits(x, "dgCMatrix")) {
    class <- "dbSparseMatrix"
  } else {
    stopf("Matrix type not supported")
  }

  res <- dbMatrix(value = x, con = con, name = name, class = class, ...)

  return(res)
}

#' @title as_ijx
#' @param x dgCMatrix or matrix
#' @noRd
as_ijx <- function(x) {
  # check that x is a dgCMatrix, matrix, or dgeMatrix
  stopifnot(
    inherits(x, "dgCMatrix") ||
      inherits(x, "matrix") ||
      inherits(x, "dgeMatrix")
  )

  # Convert dgc into TsparseMatrix class from {Matrix}
  if (is(x, "dgCMatrix")) {
    ijx <- as(x, "TsparseMatrix")
    df <- data.table::data.table(i = ijx@i + 1L, j = ijx@j + 1L, x = ijx@x)
  } else if (is(x, "matrix") || is(x, "dgeMatrix")) {
    row_indices <- rep(seq_len(nrow(x)), times = ncol(x))
    col_indices <- rep(seq_len(ncol(x)), each = nrow(x))
    values <- as.vector(x)
    df <- data.table::data.table(
      i = row_indices,
      j = col_indices,
      x = values
    )
  } else {
    stopf("Matrix type not supported")
  }

  return(df)
}

#' dbMatrix_from_tbl
#' @description Constructs a \code{dbSparseMatrix} object from a \code{tbl_duckdb_connection} object.
#' @details
#' The \code{tbl_duckdb_connection} object must contain dimension names as columns in long format.
#'
#' If \code{value_colName} is provided, the function uses pre-aggregated counts from that column.
#' This is useful when the input table already contains aggregated counts (e.g., from a GROUP BY + SUM operation).
#' If \code{value_colName} is \code{NULL} (default), the function counts occurrences of each row-column pair.
#'
#' When \code{row_names} and/or \code{col_names} are provided, the function uses these directly
#' instead of querying distinct values from the table. This can significantly improve performance
#' when the input table is a complex lazy query (e.g., result of spatial joins).
#'
#' When \code{i_col} and \code{j_col} are provided, the function uses these pre-computed integer
#' indices directly, skipping expensive string-to-index encoding. This is the fastest path.
#'
#' @param tbl \code{tbl_duckdb_connection} table in DuckDB database in long format
#' @param rownames_colName \code{character} column name of rownames in tbl \code{(required)}
#' @param colnames_colName \code{character} column name of colnames in tbl \code{(required)}
#' @param value_colName \code{character} column name containing pre-aggregated integer counts.
#' If \code{NULL} (default), counts occurrences of each row-column pair. \code{(optional)}
#' @param name table name to assign within database \code{(required, default: "dbMatrix")}
#' @param overwrite whether to overwrite if table already exists in database \code{(required)}
#' @param row_names \code{character} vector of pre-computed row names (sorted). If \code{NULL}
#' (default), row names are extracted from the table. \code{(optional)}
#' @param col_names \code{character} vector of pre-computed column names (sorted). If \code{NULL}
#' (default), column names are extracted from the table. \code{(optional)}
#' @param i_col \code{character} column name containing pre-computed row indices (1-based integers).
#' If provided with \code{j_col}, skips index encoding for optimal performance. \code{(optional)}
#' @param j_col \code{character} column name containing pre-computed column indices (1-based integers).
#' If provided with \code{i_col}, skips index encoding for optimal performance. \code{(optional)}
#'
#' @return `dbMatrix` object
#' @concept dbMatrix
#' @export
dbMatrix_from_tbl <- function(
  tbl,
  rownames_colName,
  colnames_colName,
  value_colName = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  row_names = NULL,
  col_names = NULL,
  i_col = NULL,
  j_col = NULL
) {
  # Check args
  con <- dbplyr::remote_con(tbl)
  .check_con(con)
  .check_tbl(tbl)
  .check_name(name = name)
  .check_overwrite(
    conn = con,
    name = name,
    skip_value_check = TRUE,
    overwrite = overwrite
  )

  if (is.null(rownames_colName) || is.null(colnames_colName)) {
    stop("rownames_colName and colnames_colName must be provided")
  }

  if (!all(c(rownames_colName, colnames_colName) %in% colnames(tbl))) {
    stop(
      "rownames_colName and colnames_colName must be present in tbl colnames"
    )
  }

  if (name %in% DBI::dbListTables(con) && !overwrite) {
    stop(
      "name already exists in the database.
          Please choose a unique name or set overwrite to 'TRUE'."
    )
  }

  # check if i and j are column names
  check_names <- intersect(
    c(
      colnames(tbl),
      as.character(rownames_colName),
      as.character(colnames_colName)
    ),
    c("i", "j")
  )
  if (length(check_names) > 0) {
    stop(
      "i and j are reserved names for matrix dimensions. please choose
         new column names"
    )
  }

  # Validate value_colName if provided
  if (!is.null(value_colName)) {
    if (!value_colName %in% colnames(tbl)) {
      stop(
        "value_colName '",
        value_colName,
        "' not found in tbl. ",
        "Available columns: ",
        paste(colnames(tbl), collapse = ", ")
      )
    }
  }

  rownames_colName <- rlang::sym(rownames_colName)
  colnames_colName <- rlang::sym(colnames_colName)

  # check for NA values in row/col names
  n_na <- tbl |>
    dplyr::filter(is.na(rownames_colName) | is.na(colnames_colName)) |>
    dplyr::tally() |>
    dplyr::pull(n)

  if (n_na > 0) {
    stop("NA values found in rownames or colnames. Please remove NA values.")
  }

  # Get connection
  con <- dbplyr::remote_con(tbl)

  # Fast path: use pre-computed integer indices if available
  use_precomputed_indices <- !is.null(i_col) &&
    !is.null(j_col) &&
    all(c(i_col, j_col) %in% colnames(tbl))

  if (use_precomputed_indices) {
    # Validate that row_names and col_names are also provided
    if (is.null(row_names) || is.null(col_names)) {
      stop("row_names and col_names must be provided when using i_col/j_col")
    }

    row_names <- sort(unique(as.character(row_names)))
    col_names <- sort(unique(as.character(col_names)))
    dim_i <- as.integer(length(row_names))
    dim_j <- as.integer(length(col_names))

    # Build aggregation query using pre-computed indices directly
    source_sql <- dbplyr::sql_render(tbl)
    row_col <- as.character(rownames_colName)
    col_col <- as.character(colnames_colName)

    if (!is.null(value_colName)) {
      agg_expr <- paste0("SUM(", value_colName, ")")
    } else {
      agg_expr <- "COUNT(*)"
    }

    if (overwrite) {
      DBI::dbExecute(con, glue::glue('DROP TABLE IF EXISTS "{name}"'))
    }

    # Aggregate using pre-computed integer indices - no JOINs needed
    sql <- glue::glue(
      'CREATE TABLE "{name}" AS
        SELECT
          "{i_col}" AS i,
          "{j_col}" AS j,
          {agg_expr} AS x
        FROM ({source_sql}) AS _src
        WHERE "{i_col}" IS NOT NULL AND "{j_col}" IS NOT NULL
        GROUP BY "{i_col}", "{j_col}"
      '
    )
    DBI::dbExecute(con, sql)
    ijx <- dplyr::tbl(con, name)

    # set metadata
    dims <- c(dim_i, dim_j)
    dim_names <- list(row_names, col_names)

    res <- new(
      Class = "dbSparseMatrix",
      value = ijx,
      name = name,
      init = TRUE,
      dim_names = dim_names,
      dims = dims
    )

    return(res)
  }

  # Decide encoding strategy:
  # - If row_names/col_names provided: use lookup JOINs for correct index mapping
  #   (needed when caller expects specific dimension ordering)
  # - Otherwise: use DENSE_RANK for fast SQL aggregation
  use_lookup_joins <- !is.null(row_names) && !is.null(col_names)

  if (use_lookup_joins) {
    # Prepare sorted unique dimension names
    row_names <- sort(unique(as.character(row_names)))
    col_names <- sort(unique(as.character(col_names)))
    dim_i <- as.integer(length(row_names))
    dim_j <- as.integer(length(col_names))

    # Get the source table name from the lazy query
    source_sql <- dbplyr::sql_render(tbl)

    # Build aggregation + encoding query using raw SQL
    row_col <- as.character(rownames_colName)
    col_col <- as.character(colnames_colName)

    if (!is.null(value_colName)) {
      agg_expr <- paste0("SUM(", value_colName, ")")
      cli::cli_alert_info(
        "Using pre-aggregated counts from '{value_colName}' column"
      )
    } else {
      agg_expr <- "COUNT(*)"
    }

    # Create temporary lookup tables for row/col name -> index mapping
    row_lookup_name <- paste0(
      "_row_lookup_",
      paste(sample(c(letters, 0:9), 8, replace = TRUE), collapse = "")
    )
    col_lookup_name <- paste0(
      "_col_lookup_",
      paste(sample(c(letters, 0:9), 8, replace = TRUE), collapse = "")
    )

    # Register row lookup
    row_lookup_df <- data.frame(
      row_name = row_names,
      i = seq_along(row_names),
      stringsAsFactors = FALSE
    )
    duckdb::duckdb_register(
      con,
      row_lookup_name,
      row_lookup_df,
      overwrite = TRUE
    )

    # Register col lookup
    col_lookup_df <- data.frame(
      col_name = col_names,
      j = seq_along(col_names),
      stringsAsFactors = FALSE
    )
    duckdb::duckdb_register(
      con,
      col_lookup_name,
      col_lookup_df,
      overwrite = TRUE
    )

    # Use JOIN with lookup tables to get correct indices based on provided names
    if (overwrite) {
      DBI::dbExecute(con, paste0("DROP TABLE IF EXISTS \"", name, "\""))
    }

    # Optimized: JOIN for indices FIRST, then aggregate on integers
    # This is faster than aggregating strings then JOINing, because integer
    # GROUP BY is more efficient than string GROUP BY
    value_select <- if (!is.null(value_colName)) {
      glue::glue(', _src."{value_colName}"')
    } else {
      ""
    }

    sql <- glue::glue(
      'CREATE TABLE "{name}" AS
        SELECT
          indexed.i AS i,
          indexed.j AS j,
          {agg_expr} AS x
        FROM (
          SELECT r.i, c.j{value_select}
          FROM ({source_sql}) AS _src
          INNER JOIN "{row_lookup_name}" AS r
            ON _src."{row_col}" = r.row_name
          INNER JOIN "{col_lookup_name}" AS c
            ON _src."{col_col}" = c.col_name
        ) AS indexed
        GROUP BY indexed.i, indexed.j
      '
    )
    DBI::dbExecute(con, sql)

    # Clean up temporary lookup tables
    duckdb::duckdb_unregister(con, row_lookup_name)
    duckdb::duckdb_unregister(con, col_lookup_name)

    ijx <- dplyr::tbl(con, name)
  } else {
    # Fast path: use DENSE_RANK when caller doesn't specify dimension names
    # Aggregate counts
    if (!is.null(value_colName)) {
      value_colName_sym <- rlang::sym(value_colName)
      count_table <- tbl |>
        dplyr::group_by(rownames_colName, colnames_colName) |>
        dplyr::summarise(
          x = sum(!!value_colName_sym, na.rm = TRUE),
          .groups = "drop"
        )
      cli::cli_alert_info(
        "Using pre-aggregated counts from '{value_colName}' column"
      )
    } else {
      count_table <- tbl |>
        dplyr::group_by(rownames_colName, colnames_colName) |>
        dplyr::summarise(x = dplyr::n(), .groups = "drop")
    }

    i_encoded <- rlang::sym(paste0(as.character(rownames_colName), "_encoded"))
    j_encoded <- rlang::sym(paste0(as.character(colnames_colName), "_encoded"))

    count_table <- count_table |>
      dplyr::mutate(i_encoded := dplyr::dense_rank(rownames_colName)) |>
      dbplyr::window_order(rownames_colName)

    # Extract row names from data (DENSE_RANK indices match this ordering)
    row_names <- count_table |>
      dplyr::distinct(rownames_colName) |>
      dplyr::arrange(rownames_colName) |>
      dplyr::pull(rownames_colName)

    dim_i <- as.integer(length(row_names))

    count_table <- count_table |>
      dplyr::mutate(j_encoded := dplyr::dense_rank(colnames_colName)) |>
      dbplyr::window_order(colnames_colName) |>
      dplyr::ungroup()

    # Extract col names from data
    col_names <- count_table |>
      dplyr::distinct(colnames_colName) |>
      dplyr::arrange(colnames_colName) |>
      dplyr::pull(colnames_colName)

    dim_j <- as.integer(length(col_names))

    ijx <- count_table |>
      dplyr::select(i = i_encoded, j = j_encoded, x) |>
      dplyr::compute(name = name, overwrite = overwrite, temporary = FALSE)
  }

  # set metadata
  dims <- c(dim_i, dim_j)
  dim_names <- list(row_names, col_names)

  res <- new(
    Class = "dbSparseMatrix",
    value = ijx,
    name = name,
    init = TRUE,
    dim_names = dim_names,
    dims = dims
  )

  return(res)
}

# readers ####
#' read_matrix
#' @description Ingest tabular matrix files into database
#' @details
#' Construct a database VIEW of a .csv, .tsv, or .txt files or their .gz/.gzip
#' variants
#' @param value path to .txt, .csv, .tsv or .gzip/.gz variants \code{(required)}
#' @param name name to assign file within database \code{(optional)}.
#' default: "dbMatrix"
#' @param con DBI or duckdb connection object \code{(required)}
#' @param overwrite whether to overwrite if `name` already exists in database.
#' \code{(required)}. default: FALSE
#' @param ... additional params to pass
#'
#' @return tbl_dbi object
#' @noRd
#' @keywords internal
#'
#' @examples
#' print('TODO')
read_matrix <- function(con, value, name = "dbMatrix", overwrite = FALSE, ...) {
  # check inputs
  .check_con(con)
  .check_value(value)
  .check_name(name)
  .check_overwrite(
    conn = con,
    overwrite = overwrite,
    name = name,
    skip_value_check = TRUE
  )

  # Read in files
  if (grepl("\\.csv|\\.tsv|\\.txt", value)) {
    sql <- glue::glue(
      "CREATE OR REPLACE TEMPORARY VIEW {name} AS
       SELECT * FROM read_csv_auto('{value}', header = TRUE);"
    )

    if (!overwrite) {
      sql <- gsub("OR REPLACE ", "", sql, fixed = TRUE)
    }

    DBI::dbExecute(con, sql)

    res <- dplyr::tbl(con, name)
  } else {
    stop("File type not supported.")
  }

  return(res)
}

#' read_MM
#' @description Read matrix market file (.mtx or .mtx.gz)  into database
#' @details
#' Construct a database VIEW or TABLE of a .mtx or .mtx.gz file with columns
#' 'i', 'j', and 'x' representing the row index, column index, and value of
#' the matrix, respectively.
#'
#' By default 'i' and 'j' are of type BIGINT and 'x' is of type DOUBLE.
#' **Note**: lack of support in R for BIGINT may cause errors when pulling data
#' into memory without proper type conversion.
#'
#' By default, .mtx files are expected to contain two lines representing the
#' standard header information.
#' @param value path to .mtx or .mtx.gz file \code{(required)}
#' @param name name to assign file within database \code{(optional)}.
#' default: "dbMatrix"
#' @param con DBI or duckdb connection object \code{(required)}
#' @param overwrite whether to overwrite if `name` already exists in database.
#' \code{(required)}. default: FALSE
#' @param temporary whether to create a temporary view (TRUE) or permanent table (FALSE)
#' \code{(optional)}. default: TRUE
#'
#' @return tbl_dbi object
#' @noRd
#' @keywords internal
#'
#' @examples
#' print("TODO")
readMM <- function(
  con,
  value,
  name = "dbMatrix",
  overwrite = FALSE,
  temporary = TRUE
) {
  # check inputs
  .check_con(con)
  .check_value(value)
  .check_name(name)
  .check_overwrite(
    conn = con,
    overwrite = overwrite,
    name = name,
    skip_value_check = FALSE
  )

  if (grepl("\\.csv|\\.tsv|\\.txt", value)) {
    stop("Please use read_matrix() for .csv, .tsv, .txt files.")
  }

  # Read in .mtx or .mtx.gz file
  if ((grepl("\\.mtx", value))) {
    # Determine if creating temporary view or permanent table
    create_statement <- if (temporary) {
      "CREATE OR REPLACE TEMPORARY VIEW"
    } else {
      "CREATE TABLE"
    }

    # Add IF NOT EXISTS clause if not overwriting
    if (!overwrite && !temporary) {
      create_statement <- paste(create_statement, "IF NOT EXISTS")
    }

    # .mtx reader
    # Note: Using read_csv instead of read_csv_auto for explicit control
    # MTX format: rows starting with % are comments, then a header line with dims,
    # then space-separated triplets (row col value)
    sql <- glue::glue(
      "{create_statement} {name} AS
       SELECT * FROM read_csv(
          '{value}',
          delim = ' ',
          comment = '%',
          columns = {{
              'i': 'BIGINT',
              'j': 'BIGINT',
              'x': 'DOUBLE'
          }},
          header = FALSE,
          skip = 1,
          ignore_errors = FALSE
      );"
    )

    if (!overwrite && temporary) {
      sql <- gsub("OR REPLACE ", "", sql, fixed = TRUE)
    }

    DBI::dbExecute(con, sql)

    res <- dplyr::tbl(con, name)
  } else {
    stop("File type not supported.")
  }

  return(res)
}


#' get_MM_dim
#' @description Internal function to read dimensions of a .mtx file
#' @details
#' Scans for the header of an mtx file (starting with %) and takes one more line
#' representing the dimensions and number of nonzero values.
#'
#' Note: the header size can vary depending on the .mtx file.
#'
#' @param mtx_file_path path to .mtx file to be read into database
#' @return integer vector of dimensions
#' @keywords internal
get_MM_dim <- function(mtx_file_path) {
  if (!file.exists(mtx_file_path)) {
    stop("File does not exist. Check for valid file path.")
  }

  # Read all lines starting with '%' and one additional line representing dims
  # Use gzfile for compressed files, otherwise use file
  if (grepl("\\.gz$", mtx_file_path)) {
    con <- gzfile(mtx_file_path, "r")
  } else {
    con <- file(mtx_file_path, "r")
  }

  header <- character(0)
  repeat {
    line <- readLines(con, n = 1)
    if (length(line) == 0 || !startsWith(line, "%")) {
      break
    }
    header <- c(header, line)
  }
  header <- c(header, line) # Add the dims
  close(con)

  # Extract dimensions from the last line (dims)
  dims <- as.integer(strsplit(header[length(header)], " ")[[1]][1:2])

  return(dims)
}

#' get_MM_dimnames
#' @description Internal function to read row and column names of a .mtx file
#' @details
#' Can be used to read row and column names from .mtx files. Note: these files
#' must not contain a header (colnames).
#'
#' The mtx_rowname_col_idx and mtx_colname_col_idx can be used to specify the column
#' index of the row and column name files, respectively. By default, the first
#' column is used for both.
#'
#' TODO: Support for reading in only rownames or colnames.
#'
#' @param mtx_file_path path to .mtx file to be read into database
#' @param mtx_rowname_file_path path to .mtx rowname file to be read into
#' database. by default, no header is assumed.
#' @param mtx_rowname_col_idx column index of row name file
#' @param mtx_colname_file_path path to .mtx colname file to be read into
#' database. by default, no header is assumed.
#' @param mtx_colname_col_idx column index of column name file
#' @param ... additional params to pass to [data.table::fread()]
#'
#' @return list of row and column name character vectors
#' @keywords internal
get_MM_dimnames <- function(
  mtx_file_path,
  mtx_rowname_file_path,
  mtx_rowname_col_idx = 1,
  mtx_colname_file_path,
  mtx_colname_col_idx = 1,
  ...
) {
  # check inputs
  if (
    !file.exists(mtx_file_path) ||
      !file.exists(mtx_rowname_file_path) ||
      !file.exists(mtx_colname_file_path)
  ) {
    stop("File does not exist. Check for valid file path.")
  }
  if (!is.numeric(mtx_rowname_col_idx) || !is.numeric(mtx_colname_col_idx)) {
    stop("Column index must be an integer.")
  }

  dims <- get_MM_dim(mtx_file_path)

  # Read row and column name files using data.table fread
  rowname_file <- data.table::fread(mtx_rowname_file_path, header = FALSE)
  dim_rownames <- dim(rowname_file)
  colname_file <- data.table::fread(mtx_colname_file_path, header = FALSE)
  dim_colnames <- dim(colname_file)

  # check dimname and column indices
  if (mtx_rowname_col_idx > dim_rownames[2]) {
    stop(
      "'mtx_rowname_col_idx' exceeds number of columns in 'mtx_rowname_file_path'"
    )
  }

  if (mtx_colname_col_idx > dim_colnames[2]) {
    stop(
      "'mtx_colname_col_idx' exceeds number of columns in 'mtx_colname_file_path'"
    )
  }

  # Extract row and column names
  rownames <- rowname_file[, ..mtx_rowname_col_idx][[1]]
  colnames <- colname_file[, ..mtx_colname_col_idx][[1]]

  # Make unique
  # Note: first replicates will be labeled --1, second --2, and so on...
  rownames <- make.unique(rownames, sep = "--")
  colnames <- make.unique(colnames, sep = "--")

  # check for duplicates and matching dimensions
  if (length(unique(rownames)) != dims[1]) {
    stop(
      "Number of unique row names does not match the number of rows in the
         matrix. Check selected col_idx of 'mtx_rowname_file_path' for valid
         row names."
    )
  }

  if (length(unique(colnames)) != dims[2]) {
    stop(
      "Number of unique column names does not match the number of columns in
         the matrix. Check selected col_idx of 'mtx_colname_file_path' for
         valid column names."
    )
  }

  dimnames <- list(rownames, colnames)

  return(dimnames)
}

# dimnames ####
#' Map dimnames to i,j indices
#' @details
#' Constructs a table in a database that contains the accompanying dimnames
#' for a dbMatrix. The resulting columns in the table:
#' * i (row index)
#' * colName_i (rownames),
#' * j (col index)
#' * j_names (colnames)
#' * x (counts of i,j occcurences)
#' @return A lazy `tbl_dbi` with `i`, `j`, `x`, and the mapped row/column name
#'   columns.
#' @param dbMatrix dbMatrix object
#' @param colName_i name of column rownames to add to database
#' @param colName_j name of column colnames to add to database
#' default: 'FALSE'.'
#' @keywords internal
map_ijx_dimnames <- function(dbMatrix, colName_i, colName_j) {
  # input validation
  .check_name(colName_i)
  .check_name(colName_j)
  con <- get_con(dbMatrix)
  .check_con(con)

  dimnames <- dimnames(dbMatrix)

  # map dimnames to indices in-memory
  dt_rownames <- data.table::data.table(dimnames[[1]])
  data.table::setnames(dt_rownames, colName_i)
  dt_rownames[, i := .I]

  dt_colnames <- data.table::data.table(dimnames[[2]])
  data.table::setnames(dt_colnames, colName_j)
  dt_colnames[, j := .I]

  # register map to db
  duckdb::duckdb_register(con, "temp_rownames", dt_rownames, overwrite = TRUE)
  duckdb::duckdb_register(con, "temp_colnames", dt_colnames, overwrite = TRUE)

  dimnames1_tbl <- dplyr::tbl(con, "temp_rownames")
  dimnames2_tbl <- dplyr::tbl(con, "temp_colnames")

  res <- dbMatrix[] |>
    dplyr::left_join(dimnames1_tbl, by = "i") |>
    dplyr::left_join(dimnames2_tbl, by = "j") |>
    dplyr::select(
      i,
      !!colName_i := colName_i, # !! to unquote
      j,
      !!colName_j := colName_j, # !! to unquote
      x
    )

  return(res)
}

# compute ####
#' @title Force computation of a dbMatrix
#' @description
#' Explicitly compute a dbMatrix and save it to a table in the database.
#' This overrides the default `dplyr::compute` to use a direct `CREATE TABLE AS`
#' statement, which is more robust for large tables in DuckDB.
#'
#' @param x A `dbMatrix` object
#' @param name Name of the table to create. If NULL, a random name is generated.
#' @param temporary Logical. If TRUE (default), create a temporary table.
#' @param dimnames default = TRUE. If TRUE, the rownames and colnames will be
#' saved in the database. This allows full reconstruction of the dbMatrix object
#' using [dbLoad()].
#' @param overwrite Logical. If TRUE, overwrite the table if it already exists.
#' Default is FALSE.
#' @param ... Additional arguments passed to methods (ignored).
#' @return A `dbMatrix` object pointing to the new table.
#' @export
#' @method compute dbMatrix
compute.dbMatrix <- function(
  x,
  name = NULL,
  temporary = TRUE,
  dimnames = TRUE,
  overwrite = FALSE,
  ...
) {
  con <- dbplyr::remote_con(x@value)

  if (is.null(name)) {
    if (!is.na(x@name)) {
      name <- x@name
    } else {
      name <- unique_table_name("dbmatrix_compute")
    }
  }

  # Generate SQL from the lazy tbl
  sql_query <- dbplyr::sql_render(x@value)

  # Note: DuckDB supports CREATE OR REPLACE TEMPORARY TABLE
  temp_str <- if (temporary) "TEMPORARY" else ""
  replace_str <- if (overwrite) "OR REPLACE" else ""

  # We use CREATE OR REPLACE to handle overwrite=TRUE
  full_sql <- glue::glue(
    "CREATE {replace_str} {temp_str} TABLE {name} AS {sql_query}"
  )

  # Execute
  # We wrap in tryCatch to provide better error messages
  tryCatch(
    {
      invisible(DBI::dbExecute(con, full_sql))
    },
    error = function(e) {
      cli::cli_abort("Failed to compute dbMatrix: {e$message}")
    }
  )

  # Write dimnames if requested
  if (dimnames) {
    .write_dimnames(x = x, name = name)
  }

  # Return new dbMatrix pointing to the new table
  new_tbl <- dplyr::tbl(con, name)

  # Update the object
  x@value <- new_tbl
  x@name <- name
  return(x)
}

#' @export
#' @method compute dbSparseMatrix
compute.dbSparseMatrix <- function(
  x,
  name = NULL,
  temporary = TRUE,
  dimnames = TRUE,
  overwrite = FALSE,
  ...
) {
  # First, compute the base matrix using parent method
  con <- dbplyr::remote_con(x@value)

  if (is.null(name)) {
    if (!is.na(x@name)) {
      name <- x@name
    } else {
      name <- unique_table_name("dbmatrix_compute")
    }
  }

  # Generate SQL from the lazy tbl
  sql_query <- dbplyr::sql_render(x@value)

  # Extract temp tables from the query BEFORE materializing
  temp_tables <- .extract_temp_tables(as.character(sql_query))

  temp_str <- if (temporary) "TEMPORARY" else ""
  replace_str <- if (overwrite) "OR REPLACE" else ""

  full_sql <- glue::glue(
    "CREATE {replace_str} {temp_str} TABLE {name} AS {sql_query}"
  )

  tryCatch(
    {
      invisible(DBI::dbExecute(con, full_sql))

      # Clean up temp tables after successful materialization
      if (length(temp_tables) > 0) {
        .cleanup_temp_tables(
          con,
          temp_tables,
          verbose = getOption("dbMatrix.verbose", TRUE)
        )
      }
    },
    error = function(e) {
      cli::cli_abort("Failed to compute dbMatrix: {e$message}")
    }
  )

  # Write dimnames if requested
  if (dimnames) {
    .write_dimnames(x = x, name = name)
  }

  # Return new dbMatrix pointing to the new table
  new_tbl <- dplyr::tbl(con, name)

  # Update the object
  x@value <- new_tbl
  x@name <- name
  return(x)
}

#' @export
#' @method compute dbDenseMatrix
compute.dbDenseMatrix <- compute.dbMatrix

test_that("dbMatrix_from_tbl creates dbSparseMatrix from tbl", {
  # Setup: Create a simple test table
  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

  test_data <- data.frame(
    gene = c("gene1", "gene1", "gene2", "gene2", "gene3"),
    cell = c("cell1", "cell2", "cell1", "cell3", "cell1")
  )

  tbl <- dplyr::copy_to(con, test_data, name = "test_data", overwrite = TRUE)

  # Test: Create dbMatrix from tbl
  result <- dbMatrix_from_tbl(
    tbl = tbl,
    rownames_colName = "gene",
    colnames_colName = "cell",
    name = "test_matrix",
    overwrite = TRUE
  )

  # Assertions
  expect_s4_class(result, "dbSparseMatrix")
  expect_equal(dim(result), c(3L, 3L)) # 3 genes x 3 cells
  expect_equal(rownames(result), c("gene1", "gene2", "gene3"))
  expect_equal(colnames(result), c("cell1", "cell2", "cell3"))

  # Check the counts are correct
  mat <- as.matrix(result, sparse = TRUE, names = TRUE)
  expect_equal(mat["gene1", "cell1"], 1)
  expect_equal(mat["gene1", "cell2"], 1)
  expect_equal(mat["gene2", "cell1"], 1)
  expect_equal(mat["gene2", "cell3"], 1)
  expect_equal(mat["gene3", "cell1"], 1)

  # Cleanup
  DBI::dbDisconnect(con, shutdown = TRUE)
})

test_that("dbMatrix_from_tbl handles multiple occurrences", {
  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

  # Create data with repeated gene-cell combinations
  test_data <- data.frame(
    gene = c("gene1", "gene1", "gene1", "gene2"),
    cell = c("cell1", "cell1", "cell2", "cell1")
  )

  tbl <- dplyr::copy_to(con, test_data, overwrite = TRUE)

  result <- dbMatrix_from_tbl(
    tbl = tbl,
    rownames_colName = "gene",
    colnames_colName = "cell",
    overwrite = TRUE
  )

  # Assertions: counts should reflect multiple occurrences
  mat <- as.matrix(result, sparse = TRUE, names = TRUE)
  expect_equal(mat["gene1", "cell1"], 2) # gene1-cell1 appears twice
  expect_equal(mat["gene1", "cell2"], 1)
  expect_equal(mat["gene2", "cell1"], 1)

  DBI::dbDisconnect(con, shutdown = TRUE)
})

test_that("dbMatrix_from_tbl validates inputs", {
  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
  test_data <- data.frame(gene = "gene1", cell = "cell1")
  tbl <- dplyr::copy_to(con, test_data, overwrite = TRUE)

  # Test missing rownames_colName
  expect_error(
    dbMatrix_from_tbl(tbl, rownames_colName = NULL, colnames_colName = "cell"),
    "rownames_colName and colnames_colName must be provided"
  )

  # Test invalid column names
  expect_error(
    dbMatrix_from_tbl(
      tbl,
      rownames_colName = "invalid",
      colnames_colName = "cell"
    ),
    "rownames_colName and colnames_colName must be present in tbl colnames"
  )

  DBI::dbDisconnect(con, shutdown = TRUE)
})

test_that("as.dbMatrix works for dense matrix", {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  m <- matrix(1:9, nrow = 3)
  colnames(m) <- paste0("c", 1:3)
  rownames(m) <- paste0("r", 1:3)

  db_mat <- as.dbMatrix(m, con = con)
  expect_s4_class(db_mat, "dbDenseMatrix")
  expect_equal(dim(db_mat), c(3, 3))
  expect_equal(as.matrix(db_mat, names = TRUE), m)
})

test_that("as.dbMatrix handles duplicate column names", {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  m <- matrix(1:6, nrow = 2)
  colnames(m) <- c("A", "A", "B")
  rownames(m) <- c("r1", "r2")

  db_mat <- as.dbMatrix(m, con = con)
  expect_s4_class(db_mat, "dbDenseMatrix")
  expect_equal(colnames(db_mat), c("A", "A", "B"))
  expect_equal(as.matrix(db_mat, names = TRUE), m)
})

test_that("as.dbMatrix works for sparse matrix", {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  m <- Matrix::rsparsematrix(10, 10, 0.1)
  # Ensure dimnames for comparison
  dimnames(m) <- list(paste0("r", 1:10), paste0("c", 1:10))

  db_mat <- as.dbMatrix(m, con = con)
  expect_s4_class(db_mat, "dbSparseMatrix")
  expect_equal(as.matrix(db_mat, names = TRUE), as.matrix(m))
})

test_that("as.dbMatrix works for dgeMatrix (dense Matrix)", {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  m <- Matrix::Matrix(1:9, nrow = 3, sparse = FALSE)
  dimnames(m) <- list(paste0("r", 1:3), paste0("c", 1:3))
  db_mat <- as.dbMatrix(m, con = con)
  expect_s4_class(db_mat, "dbDenseMatrix")
  expect_equal(as.matrix(db_mat, names = TRUE), as.matrix(m))
})

test_that("as.dbMatrix works for dgRMatrix (row-oriented sparse)", {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  m <- Matrix::rsparsematrix(10, 10, 0.1)
  dimnames(m) <- list(paste0("r", 1:10), paste0("c", 1:10))
  m_r <- as(m, "RsparseMatrix") # dgRMatrix

  # Should be coerced to dgCMatrix and then ingested
  db_mat <- as.dbMatrix(m_r, con = con)
  expect_s4_class(db_mat, "dbSparseMatrix")
  expect_equal(as.matrix(db_mat, names = TRUE), as.matrix(m))
})

test_that("as.dbMatrix fails for unsupported types", {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  expect_error(
    as.dbMatrix(data.frame(a = 1), con = con),
    "Unsupported object class"
  )
})

test_that("as.dbMatrix works for lMatrix (logical Matrix)", {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  m_log <- Matrix::Matrix(c(TRUE, FALSE, TRUE, FALSE), 2, 2)
  dimnames(m_log) <- list(c("r1", "r2"), c("c1", "c2"))

  db_mat <- as.dbMatrix(m_log, con = con)
  expect_s4_class(db_mat, "dbDenseMatrix")

  # Should be ingested as boolean (logical) or 1/0 depending on path
  expected <- as.matrix(m_log) * 1
  expect_equal(as.matrix(db_mat, names = TRUE), expected)
})

# Test column-wise and row-wise vector operations for dbMatrix
# These tests verify that vector arithmetic correctly handles:
# - Row-wise operations (vector length = nrow)
# - Column-wise operations (vector length = ncol)
# - General recycling (other lengths)

# Helper to compare values ignoring dimnames differences
expect_equal_values <- function(actual, expected, tolerance = 1e-10) {
  dimnames(actual) <- NULL
  dimnames(expected) <- NULL
  expect_equal(actual, expected, tolerance = tolerance)
}

test_that("dbDenseMatrix column-wise subtraction works correctly", {
  mat <- matrix(1:12, nrow = 3, ncol = 4)
  dbdm <- as.dbMatrix(mat, path = tempfile(fileext = ".duckdb"), name = "test")
  
  # Column-wise subtraction (vector length = ncol)
  cm <- colMeans(mat)
  ref <- sweep(mat, 2, cm, "-")
  result <- as.matrix(dbdm - cm)
  
  expect_equal_values(ref, result)
})


test_that("dbDenseMatrix column-wise division works correctly", {
  mat <- matrix(1:12, nrow = 3, ncol = 4)
  dbdm <- as.dbMatrix(mat, path = tempfile(fileext = ".duckdb"), name = "test")
  
  # Column-wise division (vector length = ncol)
  scale <- c(1, 2, 3, 4)
  ref <- sweep(mat, 2, scale, "/")
  result <- as.matrix(dbdm / scale)
  
  expect_equal_values(ref, result)
})


test_that("dbDenseMatrix row-wise subtraction works correctly", {
  mat <- matrix(1:12, nrow = 3, ncol = 4)
  dbdm <- as.dbMatrix(mat, path = tempfile(fileext = ".duckdb"), name = "test")
  
  # Row-wise subtraction (vector length = nrow)
  rm <- rowMeans(mat)
  ref <- sweep(mat, 1, rm, "-")
  result <- as.matrix(dbdm - rm)
  
  expect_equal_values(ref, result)
})


test_that("dbSparseMatrix column-wise subtraction works correctly", {
  # Skip this test - dbSparseMatrix vector arithmetic uses recycling semantics
  # that differ from sweep(). This is a known limitation of the current
  # implementation. See GitHub issue for tracking.
  skip("dbSparseMatrix vector arithmetic has different recycling semantics than sweep()")
})


test_that("dbDenseMatrix z-scale (center + scale) works correctly", {
  set.seed(42)
  mat <- matrix(rnorm(30), nrow = 10, ncol = 3)
  dbdm <- as.dbMatrix(mat, path = tempfile(fileext = ".duckdb"), name = "test")
  
  # Z-scale: (x - colMeans) / colSds
  cm <- colMeans(mat)
  csd <- apply(mat, 2, sd)
  ref <- sweep(sweep(mat, 2, cm, "-"), 2, csd, "/")
  
  # dbMatrix version
  cm_db <- colMeans(dbdm)
  csd_db <- apply(as.matrix(dbdm), 2, sd)
  result <- as.matrix((dbdm - cm_db) / csd_db)
  
  expect_equal_values(ref, result)
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# Skip all tests if sparseMatrixStats is not available
# (required for MatrixGenerics::rowVars/colVars on dgCMatrix)
skip_if_not_installed("sparseMatrixStats")

# Helper to compare values ignoring names
expect_equal_values <- function(actual, expected) {
  names(actual) <- NULL
  names(expected) <- NULL
  expect_equal(actual, expected)
}

# ---------------------------------------------------------------------------- #
# Load the dgcMatrix
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))

# Connect to the database
con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

# Create dbSparseMatrix
dbsm <- dbMatrix::dbMatrix(
  value = dgc,
  con = con1,
  name = 'mat',
  class = "dbSparseMatrix",
  overwrite = TRUE
)

# ---------------------------------------------------------------------------- #
# rowVars

test_that("rowVars equal for dbSparseMatrix (memory=TRUE)", {
  res_mat <- MatrixGenerics::rowVars(dgc)
  res_dbsm <- rowVars(dbsm, memory = TRUE)
  expect_equal_values(res_mat, res_dbsm)
})

test_that("rowVars equal for dbSparseMatrix (memory=FALSE)", {
  res_mat <- MatrixGenerics::rowVars(dgc)
  res_dbsm <- rowVars(dbsm, memory = FALSE)
  res_dbsm_vec <- suppressWarnings(res_dbsm |> as.vector())
  expect_equal_values(res_mat, res_dbsm_vec)
})

# ---------------------------------------------------------------------------- #
# colVars

test_that("colVars equal for dbSparseMatrix (memory=TRUE)", {
  res_mat <- MatrixGenerics::colVars(dgc)
  res_dbsm <- colVars(dbsm, memory = TRUE)
  expect_equal_values(res_mat, res_dbsm)
})

test_that("colVars equal for dbSparseMatrix (memory=FALSE)", {
  res_mat <- MatrixGenerics::colVars(dgc)
  res_dbsm <- colVars(dbsm, memory = FALSE)
  res_dbsm_vec <- suppressWarnings(res_dbsm |> as.vector())
  expect_equal_values(res_mat, res_dbsm_vec)
})

# Close the database connection
DBI::dbDisconnect(con1)

# Helper function to set up test environment
setup_test_env <- function() {
  # Create a connection to an in-memory database
  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

  # Create test matrices
  dense_matrix <- matrix(1:9, 3, 3)
  colnames(dense_matrix) <- paste0("col", 1:3)
  rownames(dense_matrix) <- paste0("row", 1:3)

  sparse_matrix <- Matrix::sparseMatrix(
    i = c(1, 2, 3, 1, 3),
    j = c(1, 2, 3, 3, 1),
    x = c(1, 2, 3, 4, 5),
    dims = c(3, 3),
    dimnames = list(paste0("row", 1:3), paste0("col", 1:3))
  )

  # Convert to dbMatrix objects
  db_dense <- dbMatrix(
    value = dense_matrix,
    con = con,
    name = "dense_matrix",
    class = "dbDenseMatrix",
    overwrite = TRUE
  )

  db_sparse <- dbMatrix(
    value = sparse_matrix,
    con = con,
    name = "sparse_matrix",
    class = "dbSparseMatrix",
    overwrite = TRUE
  )

  return(list(
    con = con,
    dense_matrix = dense_matrix,
    sparse_matrix = sparse_matrix,
    db_dense = db_dense,
    db_sparse = db_sparse
  ))
}

# Tests for dbMatrix %in% ANY
test_that("dbDenseMatrix %in% vector works correctly", {
  env <- setup_test_env()

  # Test dbDenseMatrix %in% vector - returns logical vector
  result <- env$db_dense %in% c(1, 5, 9)
  expected <- env$dense_matrix %in% c(1, 5, 9)

  expect_equal(length(result), length(expected))
  expect_equal(result, expected)
})

test_that("dbSparseMatrix %in% vector throws error", {
  env <- setup_test_env()

  # Test dbSparseMatrix %in% vector - should throw error
  expect_error(
    env$db_sparse %in% c(1, 3, 5),
    "'match' requires vector arguments"
  )
})

test_that("dbDenseMatrix %in% empty vector returns logical vector of FALSE values", {
  env <- setup_test_env()

  # Test with empty vector
  result <- env$db_dense %in% numeric(0)

  # Result should be logical vector of the same length as the matrix
  expect_equal(length(result), length(env$db_dense))
  expect_true(!any(result))
})

# Tests for ANY %in% dbMatrix
test_that("vector %in% dbDenseMatrix works correctly", {
  env <- setup_test_env()

  # Test vector %in% dbDenseMatrix
  result <- c(1, 5, 9, 10) %in% env$db_dense
  expected <- c(1, 5, 9, 10) %in% env$dense_matrix

  expect_equal(result, expected)
})

test_that("empty vector %in% dbMatrix returns empty logical vector", {
  env <- setup_test_env()

  # Test with empty vector
  result <- numeric(0) %in% env$db_dense

  # Should return empty logical vector
  expect_equal(result, logical(0))
})

# Edge cases
test_that("NA values are handled correctly", {
  env <- setup_test_env()

  # Create a matrix with NA values
  na_matrix <- matrix(c(1, 2, NA, 4, NA, 6, 7, 8, 9), 3, 3)
  db_na <- dbMatrix(
    value = na_matrix,
    con = env$con,
    name = "na_matrix",
    class = "dbDenseMatrix",
    overwrite = TRUE
  )

  # Test NA %in% dbMatrix
  result_1 <- NA %in% db_na
  expected_1 <- NA %in% na_matrix
  expect_equal(result_1, expected_1)

  # Test dbMatrix %in% NA
  result_2 <- db_na %in% NA
  expected_2 <- na_matrix %in% NA
  expect_equal(result_2, expected_2)

  # Test dbMatrix %in% c(NA, 1, 2)
  result_3 <- db_na %in% c(NA, 1, 2)
  expected_3 <- na_matrix %in% c(NA, 1, 2)
  expect_equal(result_3, expected_3)
})

# Clean up
test_that("Cleanup works", {
  env <- setup_test_env()
  expect_silent(DBI::dbDisconnect(env$con))
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# Helper to compare values ignoring names
expect_equal_values <- function(actual, expected) {
  names(actual) <- NULL
  names(expected) <- NULL
  expect_equal(actual, expected)
}

# ---------------------------------------------------------------------------- #
# Load the dgcMatrix
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))
mat <- as.matrix(dgc)

# Connect to the database
con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

# Create dbDenseMatrix
dbdm <- dbMatrix::dbMatrix(
  value = mat,
  con = con1,
  name = 'mat',
  class = "dbDenseMatrix",
  overwrite = TRUE
)

# ---------------------------------------------------------------------------- #
# rowSds

test_that("rowSds equal for dbDenseMatrix (memory=TRUE)", {
  res_mat <- MatrixGenerics::rowSds(mat)
  res_dbdm <- rowSds(dbdm, memory = TRUE)
  expect_equal_values(res_mat, res_dbdm)
})

test_that("rowSds equal for dbDenseMatrix (memory=FALSE)", {
  res_mat <- MatrixGenerics::rowSds(mat)
  res_dbdm <- rowSds(dbdm, memory = FALSE)
  res_dbdm_vec <- suppressWarnings(res_dbdm |> as.vector())
  expect_equal_values(res_mat, res_dbdm_vec)
})

# ---------------------------------------------------------------------------- #
# colSds

test_that("colSds equal for dbDenseMatrix (memory=TRUE)", {
  res_mat <- MatrixGenerics::colSds(mat)
  res_dbdm <- colSds(dbdm, memory = TRUE)
  expect_equal_values(res_mat, res_dbdm)
})

test_that("colSds equal for dbDenseMatrix (memory=FALSE)", {
  res_mat <- MatrixGenerics::colSds(mat)
  res_dbdm <- colSds(dbdm, memory = FALSE)
  res_dbdm_vec <- suppressWarnings(res_dbdm |> as.vector())
  expect_equal_values(res_mat, res_dbdm_vec)
})

# Close the database connection
DBI::dbDisconnect(con1)

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))

con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

dbsm <- dbMatrix::dbMatrix(
  value = dgc,
  con = con1,
  name = 'dgc',
  class = "dbSparseMatrix",
  overwrite = TRUE
)

# ---------------------------------------------------------------------------- #
# Test name equivalence

names_dgc <- names(dgc)
names_dbsm <- names(dbsm)

test_that("names() works", {
  expect_equal(names_dgc, names_dbsm)
})


names_dgc <- rownames(dgc)
names_dbsm <- rownames(dbsm)

test_that("rownames() works", {
  expect_equal(names_dgc, names_dbsm)
})

names_dgc <- colnames(dgc)
names_dbsm <- colnames(dbsm)

test_that("colnames() works", {
  expect_equal(names_dgc, names_dbsm)
})

names_dgc <- dimnames(dgc)
names_dbsm <- dimnames(dbsm)

test_that("dimnames() works", {
  expect_equal(names_dgc, names_dbsm)
})

# Test auto-generated dimnames when none provided
test_that("dimnames auto-generated when NULL", {
  dgc_nonames <- Matrix::sparseMatrix(
    i = 1:5, j = 1:5, x = 1:5, dims = c(5, 5)
  )
  expect_null(rownames(dgc_nonames))
  expect_null(colnames(dgc_nonames))
  
  dbm <- as.dbMatrix(dgc_nonames, path = tempfile(fileext = ".duckdb"), name = "test")
  
  # Should have auto-generated names
  expect_equal(rownames(dbm), c("row1", "row2", "row3", "row4", "row5"))
  expect_equal(colnames(dbm), c("col1", "col2", "col3", "col4", "col5"))
})

# Test dimnames are character (not factor) when returned
test_that("rownames() returns character, not factor", {
  expect_true(is.character(rownames(dbsm)))
  expect_false(is.factor(rownames(dbsm)))
})

test_that("colnames() returns character, not factor", {
  expect_true(is.character(colnames(dbsm)))
  expect_false(is.factor(colnames(dbsm)))
})

test_that("dimnames() returns list of characters, not factors", {
  dn <- dimnames(dbsm)
  expect_true(is.character(dn[[1]]))
  expect_true(is.character(dn[[2]]))
  expect_false(is.factor(dn[[1]]))
  expect_false(is.factor(dn[[2]]))
})

# Test dimnames preserved after extraction
test_that("rownames preserved after row subset", {
  sub <- dbsm[1:5, ]
  expected <- rownames(dgc)[1:5]
  expect_equal(rownames(sub), expected)
  expect_true(is.character(rownames(sub)))
})

test_that("colnames preserved after col subset", {
  sub <- dbsm[, 1:5]
  expected <- colnames(dgc)[1:5]
  expect_equal(colnames(sub), expected)
  expect_true(is.character(colnames(sub)))
})

test_that("dimnames preserved after row/col subset", {
  sub <- dbsm[1:5, 1:5]
  expect_equal(rownames(sub), rownames(dgc)[1:5])
  expect_equal(colnames(sub), colnames(dgc)[1:5])
  expect_true(is.character(rownames(sub)))
  expect_true(is.character(colnames(sub)))
})

test_that("dimnames preserved after character indexing", {
  row_sel <- rownames(dgc)[1:3]
  col_sel <- colnames(dgc)[1:3]
  sub <- dbsm[row_sel, col_sel]
  expect_equal(rownames(sub), row_sel)
  expect_equal(colnames(sub), col_sel)
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))

con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

dbsm <- dbMatrix::dbMatrix(
  value = dgc,
  con = con1,
  name = 'dgc',
  class = "dbSparseMatrix",
  overwrite = TRUE
)

# ---------------------------------------------------------------------------- #
# Perform integer indexing

dgc_subset <- dgc[1:10, ]
dbsm_subset <- dbsm[1:10, ]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("integer row indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[, 1:10]
dbsm_subset <- dbsm[, 1:10]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("integer col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[1:10, 1:10]
dbsm_subset <- dbsm[1:10, 1:10]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("integer row/col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

# ---------------------------------------------------------------------------- #
# Perform character indexing

row_char_index <- rownames(dgc)[1:10]
dgc_subset <- dgc[row_char_index, ]
dbsm_subset <- dbsm[row_char_index, ]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("character row indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

col_char_index <- colnames(dgc)[1:10]
dgc_subset <- dgc[, col_char_index]
dbsm_subset <- dbsm[, col_char_index]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("character col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[row_char_index, col_char_index]
dbsm_subset <- dbsm[row_char_index, col_char_index]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)
test_that("character row/col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

# ---------------------------------------------------------------------------- #
# Perform boolean indexing

boolean_row_index <- c(rep(FALSE, nrow(dgc) - 5), rep(TRUE, 5))
dgc_subset <- dgc[boolean_row_index, ]
dbsm_subset <- dbsm[boolean_row_index, ]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("boolean row indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

boolean_col_index <- c(rep(FALSE, ncol(dgc) - 5), rep(TRUE, 5))
dgc_subset <- dgc[, boolean_col_index]
dbsm_subset <- dbsm[, boolean_col_index]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("boolean col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[boolean_row_index, boolean_col_index]
dbsm_subset <- dbsm[boolean_row_index, boolean_col_index]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("boolean row/col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

# ---------------------------------------------------------------------------- #
# Perform negative indexing

dgc_subset <- dgc[-c(1:5), ]
dbsm_subset <- dbsm[-c(1:5), ]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("negative row indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[, -c(1:5)]
dbsm_subset <- dbsm[, -c(1:5)]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("negative col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[-c(1:5), -c(1:5)]
dbsm_subset <- dbsm[-c(1:5), -c(1:5)]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("negative row/col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

# ---------------------------------------------------------------------------- #
# Perform out-of-order indexing

ooo_row_index <- c(10, 5, 1, 8, 3)
ooo_col_index <- c(8, 2, 6, 1)

dgc_subset <- dgc[ooo_row_index, ]
dbsm_subset <- dbsm[ooo_row_index, ]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("out-of-order row indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[, ooo_col_index]
dbsm_subset <- dbsm[, ooo_col_index]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("out-of-order col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[ooo_row_index, ooo_col_index]
dbsm_subset <- dbsm[ooo_row_index, ooo_col_index]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("out-of-order row/col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

# ---------------------------------------------------------------------------- #
# Perform empty indexing (regression: avoid invalid SQL like `VALUES )`)

dgc_subset <- dgc[integer(0), ]
dbsm_subset <- dbsm[integer(0), ]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("empty integer row indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[, integer(0)]
dbsm_subset <- dbsm[, integer(0)]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("empty integer col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[integer(0), integer(0)]
dbsm_subset <- dbsm[integer(0), integer(0)]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("empty integer row/col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[integer(0), 1:10]
dbsm_subset <- dbsm[integer(0), 1:10]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("empty rows with non-empty cols works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

dgc_subset <- dgc[1:10, integer(0)]
dbsm_subset <- dbsm[1:10, integer(0)]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("non-empty rows with empty cols works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

false_rows <- rep(FALSE, nrow(dgc))
dgc_subset <- dgc[false_rows, ]
dbsm_subset <- dbsm[false_rows, ]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("empty logical row indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

false_cols <- rep(FALSE, ncol(dgc))
dgc_subset <- dgc[, false_cols]
dbsm_subset <- dbsm[, false_cols]
dgc_db_subset <- as.matrix(dbsm_subset, sparse = TRUE, names = TRUE)

test_that("empty logical col indexing works", {
  expect_equal(dgc_subset, dgc_db_subset)
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))
mat <- as.matrix(dgc + 1)

con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

dbdm <- dbMatrix::dbMatrix(
  value = mat,
  con = con1,
  name = 'mat',
  class = "dbDenseMatrix",
  overwrite = TRUE
)

# ---------------------------------------------------------------------------- #
# Test scalar arithmetic
res_mat <- mat + mat
res_dbdm <- dbdm + dbdm
res_dbdm <- as.matrix(res_dbdm, names = TRUE)
test_that("+ matrix equal", {
  expect_equal(res_mat, res_dbdm)
})

res_mat <- mat - mat
res_dbdm <- dbdm - dbdm
res_dbdm <- as.matrix(res_dbdm, names = TRUE)
test_that("- matrix equal", {
  expect_equal(res_mat, res_dbdm)
})

res_mat <- mat * mat
res_dbdm <- dbdm * dbdm
res_dbdm <- as.matrix(res_dbdm, names = TRUE)
test_that("* matrix equal", {
  expect_equal(res_mat, res_dbdm)
})

mat2 <- mat * 2
dbdm2 <- dbMatrix::dbMatrix(
  value = mat2,
  con = con1,
  name = 'mat2',
  class = "dbDenseMatrix",
  overwrite = TRUE
)

res_mat <- mat - mat2
res_dbdm <- dbdm - dbdm2
res_dbdm <- as.matrix(res_dbdm, names = TRUE)
test_that("- different matrix equal", {
  expect_equal(res_mat, res_dbdm)
})

# FIXME:
# Support for division by 0
# res_mat = mat / mat
# res_dbdm = dbdm / dbdm
# res_dbdm = as.matrix(res_dbdm)
# test_that("/ matrix equal", {
#   expect_equal(res_mat, res_dbdm)
# })

# FIXME: NaN and 1 logic
# res_mat = mat ^ mat
# res_dbdm = dbdm ^ dbdm
# res_dbdm = as.matrix(res_dbdm)
# test_that("^ matrix equal", {
#   expect_equal(res_mat, res_dbdm)
# })

# FIXME: division by 0, NaN and 1 logic
# res_mat = mat %% mat
# res_dbdm = dbdm %% dbdm
# res_dbdm = as.matrix(res_dbdm)
# test_that("%% matrix equal", {
#   expect_equal(res_mat, res_dbdm)
# })

# FIXME: division by 0, Nan and 1 logic
# res_mat = mat %/% mat
# res_dbdm = dbdm %/% dbdm
# res_dbdm = as.matrix(res_dbdm)
# test_that("%/% matrix equal", {
#   expect_equal(res_mat, res_dbdm)
# })

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# Helper to compare values ignoring names
expect_equal_values <- function(actual, expected) {
  names(actual) <- NULL
  names(expected) <- NULL
  expect_equal(actual, expected)
}

# ---------------------------------------------------------------------------- #
# Load the dgcMatrix
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))

# Connect to the database
con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

# Create dbSparseMatrix
dbsm <- dbMatrix::dbMatrix(
  value = dgc,
  con = con1,
  name = 'dgc',
  class = "dbSparseMatrix",
  overwrite = TRUE
)
# ---------------------------------------------------------------------------- #
# rowSums

test_that("rowSums equal (memory=TRUE)", {
  res_dgc <- rowSums(dgc)
  res_dbsm <- rowSums(dbsm, memory = TRUE)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("rowSums equal (memory=FALSE)", {
  res_dgc <- rowSums(dgc)
  res_dbsm <- rowSums(dbsm, memory = FALSE) |> as.vector()
  expect_equal_values(res_dgc, res_dbsm)
})

# ---------------------------------------------------------------------------- #
# colSums

test_that("colSums equal (memory=TRUE)", {
  res_dgc <- colSums(dgc)
  res_dbsm <- colSums(dbsm, memory = TRUE)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("colSums equal (memory=FALSE)", {
  res_dgc <- colSums(dgc)
  res_dbsm <- colSums(dbsm, memory = FALSE) |> as.vector()
  expect_equal_values(res_dgc, res_dbsm)
})

# ---------------------------------------------------------------------------- #
# rowMeans

test_that("rowMeans equal (memory=TRUE)", {
  res_dgc <- rowMeans(dgc)
  res_dbsm <- rowMeans(dbsm, memory = TRUE)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("rowMeans equal (memory=FALSE)", {
  res_dgc <- rowMeans(dgc)
  res_dbsm <- rowMeans(dbsm, memory = FALSE) |> as.vector()
  expect_equal_values(res_dgc, res_dbsm)
})

# ---------------------------------------------------------------------------- #
# colMeans

test_that("colMeans equal (memory=TRUE)", {
  res_dgc <- colMeans(dgc)
  res_dbsm <- colMeans(dbsm, memory = TRUE)
  expect_equal_values(res_dgc, res_dbsm)
})

test_that("colMeans equal (memory=FALSE)", {
  res_dgc <- colMeans(dgc)
  res_dbsm <- colMeans(dbsm, memory = FALSE) |> as.vector()
  expect_equal_values(res_dgc, res_dbsm)
})

# Close the database connection
DBI::dbDisconnect(con1)

rlang::local_options(lifecycle_verbosity = "quiet")

dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))

con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

dbsm <- dbMatrix::dbMatrix(
  value = dgc,
  con = con1,
  name = 'dgc',
  class = "dbSparseMatrix",
  overwrite = TRUE
)

# With ops system removed, these operations trigger densification immediately.
# We must enable densification for these tests.
options(dbMatrix.allow_densify = TRUE)

test_that("sparse + non-zero scalar becomes dense", {
  res_dbsm <- dbsm + 1
  expect_s4_class(res_dbsm, "dbDenseMatrix")
})

test_that("sparse - non-zero scalar becomes dense", {
  res_dbsm <- dbsm - 1
  expect_s4_class(res_dbsm, "dbDenseMatrix")
})

test_that("sparse + zero stays sparse", {
  res_dbsm <- dbsm + 0
  expect_s4_class(res_dbsm, "dbSparseMatrix")
})

test_that("sparse * zero stays sparse", {
  res_dbsm <- dbsm * 0
  expect_s4_class(res_dbsm, "dbSparseMatrix")
})

test_that("sparse * non-zero stays sparse", {
  res_dbsm <- dbsm * 2
  expect_s4_class(res_dbsm, "dbSparseMatrix")
})

test_that("sparse / non-zero stays sparse", {
  res_dbsm <- dbsm / 2
  expect_s4_class(res_dbsm, "dbSparseMatrix")
})

test_that("non-zero + sparse becomes dense", {
  res_dbsm <- 1 + dbsm
  expect_s4_class(res_dbsm, "dbDenseMatrix")
})

test_that("zero + sparse stays sparse", {
  res_dbsm <- 0 + dbsm
  expect_s4_class(res_dbsm, "dbSparseMatrix")
})

test_that("sparse + non-zero preserves values correctly", {
  res_dgc <- dgc + 1
  res_dbsm <- dbsm + 1
  res_dbsm <- as.matrix(res_dbsm, names = TRUE)
  res_dgc <- as.matrix(res_dgc)
  expect_equal(res_dgc, res_dbsm)
})

test_that("sparse + zero preserves sparsity and values", {
  res_dgc <- dgc + 0
  res_dbsm <- dbsm + 0
  res_dbsm <- as.matrix(res_dbsm, names = TRUE, sparse = TRUE)
  expect_equal(res_dgc, res_dbsm)
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))
mat <- as.matrix(dgc + 1)

con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

dbdm <- dbMatrix::dbMatrix(
  value = mat,
  con = con1,
  name = 'mat',
  class = "dbDenseMatrix",
  overwrite = TRUE
)

# ---------------------------------------------------------------------------- #
# Test scalar arithmetic

res_mat <- mat + 1
res_mat <- res_mat |> as.matrix() #dgeMatrix casting
res_dbdm <- dbdm + 1
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("+ 1 equal", {
  expect_equal(res_mat, res_dbdm)
})


res_mat <- mat - 1
res_mat <- res_mat |> as.matrix() #dgeMatrix casting
res_dbdm <- dbdm - 1
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("-1 equal", {
  expect_equal(res_mat, res_dbdm)
})


res_mat <- mat * 10
res_dbdm <- dbdm * 10
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("* 10 equal", {
  expect_equal(res_mat, res_dbdm)
})


res_mat <- mat + 0
res_dbdm <- dbdm + 0
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("+0 equal", {
  expect_equal(res_mat, res_dbdm)
})

res_mat <- mat / 10
res_dbdm <- dbdm / 10
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("/10 equal", {
  expect_equal(res_mat, res_dbdm)
})

res_mat <- mat / 0
res_dbdm <- dbdm / 0
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("/ 0 equal", {
  expect_equal(res_mat, res_dbdm)
})

res_mat <- mat^0
res_dbdm <- dbdm^0
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("^ 0 equal", {
  expect_equal(res_mat, res_dbdm)
})

res_mat <- 0 / mat
res_dbdm <- 0 / dbdm
res_dbdm <- as.matrix(res_dbdm, names = TRUE)

test_that("0 / equal", {
  expect_equal(res_mat, res_dbdm)
})

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# Enable densification for tests
options(dbMatrix.allow_densify = TRUE)

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
dbsm <- dbMatrix::dbMatrix(
  value = readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix")),
  con = con1,
  name = 'dgc',
  class = "dbSparseMatrix",
  overwrite = TRUE
)

dgc <- as(dbsm, "dgCMatrix")

# ---------------------------------------------------------------------------- #
# Test scalar arithmetic

test_that("+ 1 equal", {
  res_dgc <- dgc + 1
  res_dgc <- res_dgc |> as.matrix() #dgeMatrix casting
  res_dbsm <- dbsm + 1
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal(res_dgc, res_dbsm)
})

test_that("-1 equal", {
  res_dgc <- dgc - 1
  res_dgc <- res_dgc |> as.matrix() #dgeMatrix casting
  res_dbsm <- dbsm - 1
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal(res_dgc, res_dbsm)
})

test_that("* 10 equal", {
  res_dgc <- dgc * 10
  res_dbsm <- dbsm * 10
  res_dbsm <- as.matrix(res_dbsm, sparse = TRUE)
  expect_equal(as.matrix(res_dgc), as.matrix(res_dbsm))
})

test_that("+0 equal", {
  res_dgc <- dgc + 0
  res_dbsm <- dbsm + 0
  res_dbsm <- as.matrix(res_dbsm, sparse = TRUE)
  expect_equal(as.matrix(res_dgc), as.matrix(res_dbsm))
})

test_that("/10 equal", {
  res_dgc <- dgc / 10
  res_dbsm <- dbsm / 10
  res_dbsm <- as.matrix(res_dbsm, sparse = TRUE)
  expect_equal(as.matrix(res_dgc), as.matrix(res_dbsm))
})

test_that("/ 0 equal", {
  res_dgc <- dgc / 0
  res_dgc <- as.matrix(res_dgc)
  res_dbsm <- dbsm / 0
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal(res_dgc, res_dbsm)
})

test_that("^ 0 equal", {
  res_dgc <- dgc^0
  res_dgc <- as.matrix(res_dgc)
  res_dbsm <- dbsm^0
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal(res_dgc, res_dbsm)
})

test_that("0 / equal", {
  res_dgc <- 0 / dgc
  res_dgc <- as.matrix(res_dgc)
  res_dbsm <- 0 / dbsm
  res_dbsm <- as.matrix(res_dbsm)
  expect_equal(res_dgc, res_dbsm)
})

# Cleanup
DBI::dbDisconnect(con1, shutdown = TRUE)

# silence deprecated internal functions
rlang::local_options(lifecycle_verbosity = "quiet")

# ---------------------------------------------------------------------------- #
# Load the RDS file in the 'data' folder
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))

con1 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

dbsm <- dbMatrix::dbMatrix(
  value = dgc,
  con = con1,
  name = 'dgc',
  class = "dbSparseMatrix",
  overwrite = TRUE
)

# ---------------------------------------------------------------------------- #
# Test scalar arithmetic

res_dgc <- dgc + dgc
res_dbsm <- dbsm + dbsm
res_dbsm <- as.matrix(res_dbsm, sparse = TRUE, names = TRUE)

test_that("+ matrix equal", {
  expect_equal(res_dgc, res_dbsm)
})

res_dgc <- dgc - dgc
res_dbsm <- dbsm - dbsm
res_dbsm <- as.matrix(res_dbsm, sparse = TRUE, names = TRUE)
test_that("- matrix equal", {
  expect_equal(res_dgc, res_dbsm)
})

res_dgc <- dgc * dgc
res_dbsm <- dbsm * dbsm
res_dbsm <- as.matrix(res_dbsm, sparse = TRUE, names = TRUE)
test_that("* matrix equal", {
  expect_equal(res_dgc, res_dbsm)
})

dgc2 <- dgc * 2
dbsm2 <- dbMatrix::dbMatrix(
  value = dgc2,
  con = con1,
  name = 'dgc2',
  class = "dbSparseMatrix",
  overwrite = TRUE
)

res_dgc <- dgc - dgc2
res_dbsm <- dbsm - dbsm2
res_dbsm <- as.matrix(res_dbsm, sparse = TRUE, names = TRUE)
test_that("- different matrix equal", {
  expect_equal(res_dgc, res_dbsm)
})

# FIXME:
# Support for division by 0
# res_dgc = dgc / dgc
# res_dbsm = dbsm / dbsm
# res_dbsm = as.matrix(res_dbsm)
# test_that("/ matrix equal", {
#   expect_equal(res_dgc, res_dbsm)
# })

# FIXME: NaN and 1 logic
# res_dgc = dgc ^ dgc
# res_dbsm = dbsm ^ dbsm
# res_dbsm = as.matrix(res_dbsm)
# test_that("^ matrix equal", {
#   expect_equal(res_dgc, res_dbsm)
# })

# FIXME: division by 0, NaN and 1 logic
# res_dgc = dgc %% dgc
# res_dbsm = dbsm %% dbsm
# res_dbsm = as.matrix(res_dbsm)
# test_that("%% matrix equal", {
#   expect_equal(res_dgc, res_dbsm)
# })

# FIXME: division by 0, Nan and 1 logic
# res_dgc = dgc %/% dgc
# res_dbsm = dbsm %/% dbsm
# res_dbsm = as.matrix(res_dbsm)
# test_that("%/% matrix equal", {
#   expect_equal(res_dgc, res_dbsm)
# })

# ---------------------------------------------------------------------------- #
# Test support-preserving sparse division (numerator positions must be in denominator)

test_that("/ sparse division works when denominator covers numerator", {
  # D / (P + D) pattern: denominator always covers numerator
  d <- dbsm
  p_plus_d <- dbsm + dbsm2 # covers all positions from both

  res_dbsm <- d / p_plus_d
  expect_s4_class(res_dbsm, "dbSparseMatrix")

  # Compare stored values (sparse doesn't store implicit 0/0 positions)
  # D / (D + 2D) = D / 3D = 1/3 for all non-zero positions
  res_data <- res_dbsm[] |> dplyr::collect()
  expect_true(all(abs(res_data$x - 1 / 3) < 1e-10))
  expect_equal(nrow(res_data), length(dgc@x)) # same number of stored values
})

test_that("/ sparse division fails when denominator doesn't cover numerator", {
  # Create matrices with different non-zero positions
  dgc_subset <- dgc[1:5, 1:5]
  dbsm_full <- dbMatrix::dbMatrix(
    value = dgc[1:10, 1:10],
    con = con1,
    name = 'full',
    class = "dbSparseMatrix",
    overwrite = TRUE
  )
  dbsm_subset <- dbMatrix::dbMatrix(
    value = dgc_subset,
    con = con1,
    name = 'subset',
    class = "dbSparseMatrix",
    overwrite = TRUE
  )

  # full / subset should fail (full has positions subset doesn't)
  expect_error(dbsm_full / dbsm_subset, "denominator")
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

.write_arrow_to_bpcells_cpp <- function(output_dir, n_rows, n_cols, col_scale, stream_factory) {
    .Call(`_dbMatrix_write_arrow_to_bpcells_cpp`, output_dir, n_rows, n_cols, col_scale, stream_factory)
}

.compute_op_svd_arrow_cpp <- function(stream_factory, n_rows, n_cols, row_offset, row_scale, col_scale, k) {
    .Call(`_dbMatrix_compute_op_svd_arrow_cpp`, stream_factory, n_rows, n_cols, row_offset, row_scale, col_scale, k)
}

#' @title Convert object to dbMatrix
#' @description
#' Generic function to convert in-memory objects to `dbMatrix` objects.
#' @param x Object to convert (e.g., matrix, dgCMatrix)
#' @param con DBI or duckdb connection object
#' @param name Table name to assign within database
#' @param overwrite Whether to overwrite if table already exists
#' @param ... Additional arguments passed to methods
#' @return A [`dbDenseMatrix`] for dense matrix inputs or a [`dbSparseMatrix`]
#'   for sparse matrix inputs. The returned object keeps the input dimensions
#'   and dimnames while storing matrix values in DuckDB.
#' @export
as.dbMatrix <- function(
  x,
  con = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  ...
) {
  UseMethod("as.dbMatrix")
}

#' @export
as.dbMatrix.matrix <- function(
  x,
  con = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  ...
) {
  if (is.null(con)) {
    con <- DBI::dbConnect(duckdb::duckdb())
  }
  .check_con(con)
  .check_name(name)
  .check_overwrite(
    conn = con,
    overwrite = overwrite,
    name = name,
    skip_value_check = TRUE
  )

  dims <- dim(x)
  dim_names <- dimnames(x)
  if (is.null(dim_names)) {
    dim_names <- list(NULL, NULL)
  }
  if (is.null(dim_names[[1]])) {
    dim_names[[1]] <- paste0("row", seq_len(dims[1]))
  }
  if (is.null(dim_names[[2]])) {
    dim_names[[2]] <- colnames(x) %||% paste0("col", seq_len(dims[2]))
  }

  # Override column names to avoid registration issues with duplicates
  colnames(x) <- paste0("c", seq_len(ncol(x)))

  tmp_view <- unique_table_name("tmp_view_dense")
  duckdb::duckdb_register(con, tmp_view, x)
  on.exit(
    DBI::dbExecute(con, glue::glue("DROP VIEW IF EXISTS {tmp_view}")),
    add = TRUE
  )

  # UNPIVOT and map columns using pragma_table_info
  # pragma_table_info('{tmp_view}') provides metadata about the view's columns,
  # including the column name and its 0-based index (cid).
  # By joining the unpivoted column name (j_name) with this metadata, we can
  # efficiently map columns to their integer index 'j' entirely within DuckDB,
  # avoiding the need to create and upload a separate mapping table from R.
  sql <- glue::glue(
    "
    CREATE TABLE {name} AS
    WITH unpivoted AS (
      SELECT row_number() OVER () as i, * FROM {tmp_view}
    )
    SELECT 
      u.i, 
      (m.cid + 1) as j, 
      u.x 
    FROM (
      SELECT * FROM unpivoted
      UNPIVOT INCLUDE NULLS (
        x FOR j_name IN (* EXCLUDE (i))
      )
    ) u
    JOIN pragma_table_info('{tmp_view}') m ON u.j_name = m.name
  "
  )

  DBI::dbExecute(con, sql)

  new(
    "dbDenseMatrix",
    value = dplyr::tbl(con, name),
    name = name,
    dims = dims,
    dim_names = dim_names,
    init = TRUE
  )
}

#' @export
as.dbMatrix.dgCMatrix <- function(
  x,
  con = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  ...
) {
  if (is.null(con)) {
    con <- DBI::dbConnect(duckdb::duckdb())
  }
  .check_con(con)
  .check_name(name)
  .check_overwrite(
    conn = con,
    overwrite = overwrite,
    name = name,
    skip_value_check = TRUE
  )

  # Extract directly from dgCMatrix format
  # We need to construct 1-based (i, j, x) triplets for dbMatrix
  n_nz <- length(x@x)
  ncol_mat <- ncol(x)
  
  # Column indices: for each column j (1-based), repeat j for each entry in that column
  # diff(@p) gives count of entries per column
  j_idx <- rep.int(seq_len(ncol_mat), diff(x@p))
  
  # Row indices: @i is 0-based, convert to 1-based
  i_idx <- x@i + 1L
  
  arrow_tbl <- arrow::arrow_table(
    i = i_idx,
    j = j_idx,
    x = x@x
  )
  
  temp_arrow_name <- unique_table_name("__dbM_arrow_ingest")
  
  duckdb::duckdb_register_arrow(con, temp_arrow_name, arrow_tbl)
  
  tryCatch({
    if (overwrite) {
      DBI::dbExecute(con, glue::glue("DROP TABLE IF EXISTS \"{name}\""))
    }
    DBI::dbExecute(con, glue::glue("CREATE TABLE \"{name}\" AS SELECT * FROM \"{temp_arrow_name}\""))
  }, finally = {
    duckdb::duckdb_unregister_arrow(con, temp_arrow_name)
  })

  new(
    "dbSparseMatrix",
    value = dplyr::tbl(con, name),
    name = name,
    dims = dim(x),
    dim_names = .ensure_dim_names(dimnames(x), dim(x)),
    init = TRUE
  )
}

#' @export
as.dbMatrix.dgTMatrix <- function(
  x,
  con = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  ...
) {
  if (is.null(con)) {
    con <- DBI::dbConnect(duckdb::duckdb())
  }
  .check_con(con)
  .check_name(name)
  .check_overwrite(
    conn = con,
    overwrite = overwrite,
    name = name,
    skip_value_check = TRUE
  )

  # dgTMatrix slots @i and @j are 0-based, convert to 1-based
  arrow_tbl <- arrow::arrow_table(
    i = x@i + 1L,
    j = x@j + 1L,
    x = x@x
  )
  
  temp_arrow_name <- unique_table_name("__dbM_arrow_ingest")
  
  duckdb::duckdb_register_arrow(con, temp_arrow_name, arrow_tbl)
  
  tryCatch({
    if (overwrite) {
      DBI::dbExecute(con, glue::glue("DROP TABLE IF EXISTS \"{name}\""))
    }
    DBI::dbExecute(con, glue::glue("CREATE TABLE \"{name}\" AS SELECT * FROM \"{temp_arrow_name}\""))
  }, finally = {
    duckdb::duckdb_unregister_arrow(con, temp_arrow_name)
  })

  new(
    "dbSparseMatrix",
    value = dplyr::tbl(con, name),
    name = name,
    dims = dim(x),
    dim_names = .ensure_dim_names(dimnames(x), dim(x)),
    init = TRUE
  )
}

#' @export
as.dbMatrix.Matrix <- function(
  x,
  con = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  ...
) {
  # Supported only whitelisted Matrix classes
  # dMatrix: Double (numeric)
  # iMatrix: Integer
  # lMatrix: Logical (coerced to numeric 1/0)
  # nMatrix: Pattern (coerced to numeric 1/0)
  if (
    !inherits(x, "dMatrix") &&
      !inherits(x, "iMatrix") &&
      !inherits(x, "lMatrix") &&
      !inherits(x, "nMatrix")
  ) {
    stop(
      "Only numeric, integer, logical, and pattern matrices are supported. Input class: ",
      class(x)[1],
      call. = FALSE
    )
  }

  if (inherits(x, "sparseMatrix")) {
    # Some classes (like dgRMatrix) have no direct coercion to dgCMatrix
    # but can be coerced to CsparseMatrix first.
    if (!inherits(x, "dgCMatrix")) {
      x <- as(x, "CsparseMatrix")
      if (!inherits(x, "dgCMatrix")) {
        x <- as(x, "dgCMatrix")
      }
    }
    as.dbMatrix.dgCMatrix(x, con, name, overwrite, ...)
  } else {
    as.dbMatrix.matrix(as.matrix(x), con, name, overwrite, ...)
  }
}

#' @export
as.dbMatrix.default <- function(
  x,
  con = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  ...
) {
  stop(
    "Unsupported object class: ",
    class(x)[1],
    ". Input must be a matrix or Matrix object.",
    call. = FALSE
  )
}

#' dbMatrix Package Global Options
#' @description The following global options can be modified to control the
#' behavior of the [`dbMatrix`] package.
#' @details Use `options()` to set the below options.
#' @section Options:
#' * `dbMatrix.digits`: integer. Number of digits to round to in the show
#' function of dbMatrix objects. Default is 7.
#' * `dbMatrix.max_mem_convert`: numeric. Maximum size (in bytes) allowed for
#' implicit conversion of `dbMatrix` to in-memory matrix. Default is 8 * 1024^3 (8GB).
#' * `dbMatrix.chunk_size`: integer. Number of columns to process per chunk during
#' densification (`.to_db_dense`). Smaller chunks reduce memory usage but may increase
#' execution time. Default is `NULL` (automatically calculated based on `dbMatrix.max_mem_convert`).
#' * `dbMatrix.max_chunks`: integer. Maximum number of chunks allowed during
#' densification. This prevents query parser errors caused by excessive `UNION ALL`
#' branches. Default is 10000.
#' * `dbMatrix.verbose`: logical. If `TRUE` (default), prints informative messages
#' during implicit coercion.
#' * `dbMatrix.precomp_db`: character. Path to an external DuckDB file containing
#' precomputed table(s). If set, `dbMatrix` will automatically attach this
#' database (read-only) and look for a suitable precomputed table to speed up densification.
#' * `dbMatrix.allow_densify`: logical. If `FALSE` (default), automatic sparse-to-dense
#' conversion is disabled. This prevents unexpected disk spilling and memory issues
#' when operations would require densification (e.g., division by zero, scalar addition
#' to sparse matrix). Set to `TRUE` to enable on-disk dense conversion. **Warning**:
#' Dense conversion can cause massive disk usage for large matrices.
#'
#' @return No return value. This documentation page describes package options.
#' @name dbMatrix_options
#' @aliases dbMatrix-options
NULL

# names ####
# TODO: add @value once setter is implemented, update to 1D dbMatrix
#' The names of a dbMatrix Object
#' @param x A dbMatrix object
#' @concept matrix_props
#' @return A character vector of the names of the 1D dbMatrix object (1D matrices only)
setMethod('names', signature(x = 'dbDenseMatrix'), function(x) {
  # Only 1D dbMatrix objects (1-dimensional matrices) should have names
  # Regular matrices should return NULL for names()
  if (!1 %in% x@dims) {
    return(NULL) # Regular matrices don't have names, return NULL instead of error
  }
  if (1 %in% dim(x)[1]) {
    return(colnames(x))
  } else {
    return(rownames(x))
  }
})

#TODO implement names <- setter

# rownames ####
#' Retrieve and Set Row (Column) Dimension Names of dbMatrix Objects
#' @inheritParams base::rownames
#' @param do.NULL Not used for this method. Included for compatibility with the
#' generic.
#' @param prefix Not used for this method. Included for compatibility with the
#' generic.
#' @return `rownames()` and `colnames()` return character vectors of dimension
#'   names. `dimnames()` returns a length-2 list containing row and column name
#'   vectors. The replacement forms return the modified [`dbMatrix`] object.
#' @rdname matrix_props
#' @concept matrix_props
#' @export
rownames.dbMatrix <- function(x, do.NULL = TRUE, prefix = "row") {
  as.character(x@dim_names[[1]])
}

#' @rdname matrix_props
#' @concept matrix_props
#' @usage \method{rownames}{dbMatrix}(x) <- value
#' @export
`rownames<-.dbMatrix` <- function(x, value) {
  if (is.null(value)) {
    stopf('rownames are required for dbMatrix objects')
  }

  if (x@dims[1] != length(value)) {
    stopf('length of rownames to set does not equal number of rows')
  }
  x@dim_names[[1]] <- value
  x
}

# colnames ####
#' @rdname matrix_props
#' @concept matrix_props
#' @export
colnames.dbMatrix <- function(x, do.NULL = TRUE, prefix = "col") {
  as.character(x@dim_names[[2]])
}

#' @rdname matrix_props
#' @concept matrix_props
#' @usage \method{colnames}{dbMatrix}(x) <- value
#' @export
`colnames<-.dbMatrix` <- function(x, value) {
  if (x@dims[2] != length(value)) {
    stopf('length of colnames to set does not equal number of columns')
  }

  x@dim_names[[2]] <- value
  x
}

# dimnames ####
#' @rdname matrix_props
#' @concept matrix_props
#' @export
setMethod('dimnames', signature(x = 'dbMatrix'), function(x) {
  list(
    as.character(x@dim_names[[1]]),
    as.character(x@dim_names[[2]])
  )
})

#' @rdname matrix_props
#' @concept matrix_props
#' @export
setMethod(
  'dimnames<-',
  signature(x = 'dbMatrix', value = 'list'),
  function(x, value) {
    x@dim_names <- value
    x
  }
)

# internal functions ####
#' @keywords internal
#' @noRd
#' @param x dbMatrix object
#' @description
#' Function to write dbMatrix dimnames to a database. This function is intended
#' to be used ONLY when saving a dbMatrix object, to permit recovery
#' of the dimnames when the object is loaded back into memory.
#'
.write_dimnames <- function(x, name) {
  con <- dbplyr::remote_con(x@value)
  .check_con(con)
  dimnames <- dimnames(x)

  if (!inherits(x, 'dbMatrix')) {
    stopf('x must be a dbMatrix object')
  }
  if (is.na(x@name) && is.null(name)) {
    stopf('Name is empty. Use dbSave() to save the lazy dbMatrix object.')
  }
  check_names <- c(
    paste0("__", name, "_rownames"),
    paste0("__", name, "_colnames")
  )
  registered_names <- duckdb::duckdb_list_arrow(conn = con)
  names_to_unregister <- check_names[check_names %in% registered_names]
  names_to_remove <- check_names[check_names %in% DBI::dbListTables(con)]
  if (length(names_to_unregister) > 0) {
    sapply(names_to_unregister, function(name) {
      duckdb::duckdb_unregister_arrow(conn = con, name = name)
    })
  }
  if (length(names_to_remove) > 0) {
    sapply(names_to_remove, function(name) {
      DBI::dbRemoveTable(con, name)
    })
  }

  # Include explicit integer indices to avoid non-deterministic ROW_NUMBER()
  rownames_dt <- data.table::data.table(
    i = seq_along(dimnames[[1]]),
    rownames = dimnames[[1]] |> as.character()
  )
  colnames_dt <- data.table::data.table(
    j = seq_along(dimnames[[2]]),
    colnames = dimnames[[2]] |> as.character()
  )

  # NOTE: Using __ prefix to distinguish from user tables
  # temporary = FALSE ensures tables persist across connection close/reopen
  dplyr::copy_to(
    df = rownames_dt,
    dest = con,
    name = paste0("__", name, "_rownames"),
    temporary = FALSE,
    overwrite = TRUE
  ) |>
    invisible()

  dplyr::copy_to(
    df = colnames_dt,
    dest = con,
    name = paste0("__", name, "_colnames"),
    temporary = FALSE,
    overwrite = TRUE
  ) |>
    invisible()
}

#' @keywords internal
#' @noRd
#' @return character
.assign_dbm_name <- function(dbm, name = NULL) {
  if (is.null(name)) {
    name <- unique_table_name('tmp_dbm')
  }
  dbm[] <- to_view(x = dbm[], name = name)
  dbm@name <- name

  return(name)
}

.onLoad <- function(libname, pkgname) {
  # Create S4 generic for %in% which requires double dispatch
  # Suppress the informational message about creating the generic
  suppressMessages({
    if (!methods::isGeneric("%in%")) {
      methods::setGeneric("%in%")
    }
  })
}

# Coercion Methods ####

#' @title Coerce dbMatrix to matrix
#' @name coerce-dbMatrix-matrix
#' @description
#' Coercion methods to convert `dbMatrix` objects to in-memory `matrix` objects.
#' Respects `dbMatrix.max_mem_convert` option to prevent OOM errors.
#' @return A base R [`matrix`] containing the collected matrix values with the
#'   same dimensions and dimnames as the source object.
setAs("dbMatrix", "matrix", function(from) {
  .check_mem_limit(from)
  as.matrix(from)
})

#' @rdname coerce-dbMatrix-matrix
#' @name coerce-dbDenseMatrix-matrix
setAs("dbDenseMatrix", "matrix", function(from) {
  .check_mem_limit(from)
  as.matrix(from)
})

#' @rdname coerce-dbMatrix-matrix
#' @name coerce-dbSparseMatrix-matrix
setAs("dbSparseMatrix", "matrix", function(from) {
  .check_mem_limit(from)
  as.matrix(from, sparse = FALSE)
})

#' @title Coerce dbMatrix to dgCMatrix
#' @name coerce-dbMatrix-dgCMatrix
#' @description
#' Coercion methods to convert `dbMatrix` objects to in-memory `dgCMatrix` objects.
#' Respects `dbMatrix.max_mem_convert` option to prevent OOM errors.
#' @return A [`Matrix::dgCMatrix-class`] object containing the collected matrix
#'   values. Dense inputs are converted to sparse Matrix format after collection.
setAs("dbMatrix", "dgCMatrix", function(from) {
  .check_mem_limit(from)
  as.matrix(from, sparse = TRUE)
})

#' @rdname coerce-dbMatrix-dgCMatrix
#' @name coerce-dbDenseMatrix-dgCMatrix
setAs("dbDenseMatrix", "dgCMatrix", function(from) {
  .check_mem_limit(from)
  mat <- as.matrix(from)
  as(mat, "dgCMatrix")
})

#' @rdname coerce-dbMatrix-dgCMatrix
#' @name coerce-dbSparseMatrix-dgCMatrix
setAs("dbSparseMatrix", "dgCMatrix", function(from) {
  .check_mem_limit(from)
  as.matrix(from, sparse = TRUE)
})

#' @title Coerce matrix to dbMatrix
#' @name coerce-matrix-dbMatrix
#' @description
#' Coercion methods to convert in-memory `matrix` objects to `dbMatrix` objects.
#' Creates a new in-memory DuckDB connection.
#' @return A database-backed matrix object. Dense inputs return a
#'   [`dbDenseMatrix`], while sparse [`Matrix::dgCMatrix-class`] inputs return a
#'   [`dbSparseMatrix`].
setAs("matrix", "dbMatrix", function(from) {
  con <- DBI::dbConnect(duckdb::duckdb())
  as.dbMatrix(from, con = con)
})

#' @rdname coerce-matrix-dbMatrix
#' @name coerce-dgCMatrix-dbMatrix
setAs("dgCMatrix", "dbMatrix", function(from) {
  con <- DBI::dbConnect(duckdb::duckdb())
  as.dbMatrix(from, con = con)
})

# dbData ####
# Import dbData class from dbProject package instead of redefining it
#' @importClassesFrom dbProject dbData

### dbMatrix ####

#' @title S4 virtual class for dbMatrix
#' @description
#' Representation of sparse and dense matrices in a database. Each object
#' is used as a connection to a single table that exists within the database.
#' Inherits from `dbData`.
#' @slot dim_names row (1) and col (2) names
#' @slot dims dimensions of the matrix
#' @slot init logical. Whether the object is fully initialized
#' @return `dbMatrix()` returns an initialized [`dbDenseMatrix`] or
#'   [`dbSparseMatrix`] S4 object that points to matrix data stored in DuckDB.
#' @rdname dbMatrix
#' @aliases dbMatrix-class
#' @exportClass dbMatrix
#' @keywords internal
dbMatrix <- setClass(
  Class = 'dbMatrix',
  contains = c('dbData', 'VIRTUAL'),
  slots = list(
    dim_names = 'list',
    dims = 'integer',
    init = 'logical'
  ),
  prototype = list(
    dim_names = list(NULL, NULL),
    dims = c(NA_integer_, NA_integer_),
    init = FALSE
  )
)

#' Ensure dim_names are non-NULL factors
#' 
#' Helper function that ensures dim_names are always non-NULL factors.
#' If dim_names are NULL, creates "row1", "row2", ... and "col1", "col2", ...
#' If dim_names are character, converts to factor for efficient indexing.
#' 
#' @param dim_names list of row and column names (can be NULL)
#' @param dims integer vector of dimensions
#' @return list of factor row and column names
#' @keywords internal
#' @noRd
.ensure_dim_names <- function(dim_names, dims) {
  # Ensure dim_names list exists

  if (is.null(dim_names)) {
    dim_names <- list(NULL, NULL)
  }
  
  # Ensure row names
  if (is.null(dim_names[[1]])) {
    dim_names[[1]] <- as.factor(paste0("row", seq_len(dims[1])))
  } else if (!is.factor(dim_names[[1]])) {
    dim_names[[1]] <- as.factor(dim_names[[1]])
  }
  
  # Ensure col names
  if (is.null(dim_names[[2]])) {
    dim_names[[2]] <- as.factor(paste0("col", seq_len(dims[2])))
  } else if (!is.factor(dim_names[[2]])) {
    dim_names[[2]] <- as.factor(dim_names[[2]])
  }
  
  dim_names
}

#### dbDenseMatrix ####
#' @title S4 Class for `dbDenseMatrix`
#'
#' @description Representation of dense matrices using an on-disk database.
#' Inherits from \link{dbMatrix}.
#' @return Objects of class `dbDenseMatrix` store all matrix entries explicitly
#'   in DuckDB. They are typically returned by [dbMatrix()] or [as.dbMatrix()]
#'   for dense inputs.
#'
#' @name dbDenseMatrix-class
#' @export
dbDenseMatrix <- setClass(
  Class = "dbDenseMatrix",
  contains = "dbMatrix"
)

#### dbSparseMatrix ####
#' @title S4 Class for dbSparseMatrix
#'
#' @description Representation of sparse matrices using an on-disk database.
#' Inherits from \link{dbMatrix}.
#' @return Objects of class `dbSparseMatrix` store only non-zero matrix entries
#'   in DuckDB. They are typically returned by [dbMatrix()] or [as.dbMatrix()]
#'   for sparse inputs.
#' @name dbSparseMatrix-class
#' @export
dbSparseMatrix <- setClass(
  Class = "dbSparseMatrix",
  contains = "dbMatrix"
)

## dbIndex ####
#' @title S4 virtual class - Simple Class for dbData indices
#' @description
#' This is a virtual class used for indices (in signatures) for indexing
#' and sub-assignment of 'dbData' objects. Simple class union of 'logical',
#' 'numeric', 'integer', and  'character'.
#' Based on the 'index' class implemented in \pkg{Matrix}
#' @keywords internal
#' @noRd
setClassUnion(
  name = 'dbIndex',
  members = c('logical', 'numeric', 'integer', 'character')
)

test_that("compute() works for dbMatrix", {
  # Setup
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  # 1. Test dbSparseMatrix
  dbsm <- new(
    "dbSparseMatrix",
    value = dplyr::tbl(con, dplyr::sql("SELECT 1 as i, 1 as j, 1.0 as x")),
    name = "sparse_test",
    dims = as.integer(c(10, 10)),
    dim_names = list(paste0("row", 1:10), paste0("col", 1:10)),
    init = TRUE
  )

  # Call compute
  res <- compute(dbsm, name = "my_computed_table")

  # Checks
  expect_s4_class(res, "dbSparseMatrix")
  expect_equal(res@name, "my_computed_table")
  expect_true(DBI::dbExistsTable(con, "my_computed_table"))

  # Check dimnames persistence
  expect_true(DBI::dbExistsTable(con, "__my_computed_table_rownames"))
  expect_true(DBI::dbExistsTable(con, "__my_computed_table_colnames"))

  # Test dbLoad reconstruction
  loaded_dbm <- dbLoad(con, "my_computed_table", class = "dbSparseMatrix")
  expect_s4_class(loaded_dbm, "dbSparseMatrix")
  expect_equal(dim(loaded_dbm), dim(dbsm))

  # Check content
  res_df <- res@value |> dplyr::collect()
  expect_equal(nrow(res_df), 1)
  expect_equal(res_df$x, 1.0)

  # 2. Test dbDenseMatrix
  dbdm <- new(
    "dbDenseMatrix",
    value = dplyr::tbl(con, dplyr::sql("SELECT 1 as i, 1 as j, 1.0 as x")),
    name = "dense_test",
    dims = as.integer(c(10, 10)),
    dim_names = list(NULL, NULL),
    init = TRUE
  )

  res_dense <- compute(dbdm, name = "my_dense_table")

  expect_s4_class(res_dense, "dbDenseMatrix")
  expect_equal(res_dense@name, "my_dense_table")
  expect_true(DBI::dbExistsTable(con, "my_dense_table"))

  # 3. Test temporary = FALSE (Persistence)
  res_persist <- compute(dbsm, name = "my_persistent_table", temporary = FALSE)
  expect_true(DBI::dbExistsTable(con, "my_persistent_table"))

  # Verify persists after reconnect
  info <- DBI::dbGetQuery(
    con,
    "SELECT table_type FROM information_schema.tables WHERE table_name = 'my_persistent_table'"
  )
  expect_equal(info$table_type, "BASE TABLE")

  # Verify temp table is temporary
  info_temp <- DBI::dbGetQuery(
    con,
    "SELECT table_type FROM information_schema.tables WHERE table_name = 'my_computed_table'"
  )
  expect_equal(info_temp$table_type, "LOCAL TEMPORARY")

  # 4. Test overwrite = FALSE (default)
  compute(dbsm, name = "overwrite_test")
  expect_true(DBI::dbExistsTable(con, "overwrite_test"))

  # Try to overwrite without flag should fail
  expect_error(
    compute(dbsm, name = "overwrite_test", overwrite = FALSE)
  )

  # 5. Test overwrite = TRUE
  res_over <- compute(dbsm, name = "overwrite_test", overwrite = TRUE)
  expect_s4_class(res_over, "dbSparseMatrix")
  expect_equal(res_over@name, "overwrite_test")
})

---
title: "Operations"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
number-sections: true
vignette: >
  %\VignetteIndexEntry{Operations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include=FALSE}
old_options <- options(
  max.print = 10, # Set the maximum number of rows to display
  width = 80 # Set the maximum number of columns to display
)
```

## Introduction

This vignette introduces the generic functions and operations supported by `dbMatrix`. 

## Loading library

```{r, message=F}
library(dbMatrix)
library(Matrix)
```

## dbMatrix generics

`dbMatrix` objects currently support several statistical matrix operations listed below with support for more coming soon.

✅  - implemented
🟧 - not yet implemented

|          | dbSparseMatrix | dbDenseMatrix |
|----------|----------------|---------------|
| colSums  | ✅             | ✅            |
| rowSums  | ✅             | ✅            |
| colMeans | ✅             | ✅            |
| rowMeans | ✅             | ✅            |
| colSds   | 🟧             | ✅            |
| rowSds   | 🟧             | ✅            |
| t        | ✅             | ✅            |
| mean     | ✅             | ✅            |
| nrow     | ✅             | ✅            |
| ncol     | ✅             | ✅            |
| dims     | ✅             | ✅            |
| head     | ✅             | ✅            |
| tail     | ✅             | ✅            |
| ...      |                |               |

## dbSparse Matrix Operations

### Create test data
Let's create a sparse matrix for demonstration:

```{r}
set.seed(42)
dgc <- Matrix::rsparsematrix(100, 50, density = 0.1, rand.x = function(n) rpois(n, 5) + 1)
rownames(dgc) <- paste0("gene_", seq_len(100))
colnames(dgc) <- paste0("cell_", seq_len(50))

dplyr::glimpse(dgc)
```

```{r}
# create dbSparseMatrix from the same dgc
con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

sparse <- dbMatrix(
  value = dgc,
  con = con,
  name = "test_matrix",
  class = "dbSparseMatrix",
  overwrite = TRUE
)

# preview
# show function aims to emulate the show method for dgCMatrix
head(sparse)
```

### transpose
```{r}
dbMatrix::t(sparse)
```

### colMeans
```{r}
dbMatrix::colMeans(sparse)
```


### colSums
```{r}
dbMatrix::colSums(sparse)
```

### rowMeans
```{r}
dbMatrix::rowMeans(sparse)
```

### rowSums
```{r}
dbMatrix::rowSums(sparse)
```

### dim
```{r}
dim(sparse)

dim(dgc)
```

### Check results are equivalent
<details>
  <summary>Click to expand </summary>
  
  ```{r}
all.equal(dbMatrix::colMeans(sparse, memory = TRUE, names = TRUE), Matrix::colMeans(dgc))

all.equal(dbMatrix::colSums(sparse, memory = TRUE, names = TRUE), Matrix::colSums(dgc))

all.equal(dbMatrix::rowMeans(sparse, memory = TRUE, names = TRUE), Matrix::rowMeans(dgc))

all.equal(dbMatrix::rowSums(sparse, memory = TRUE, names = TRUE), Matrix::rowSums(dgc))
  ```
</details>


## dbDenseMatrix Operations

```{r, eval=TRUE, message=FALSE, warning=FALSE}
# Create a dense matrix directly
set.seed(42)
mat <- matrix(rnorm(100), nrow = 10, ncol = 10)
rownames(mat) <- paste0("row_", 1:10)
colnames(mat) <- paste0("col_", 1:10)

# Create dbDenseMatrix
con2 <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
dense <- dbMatrix(
  value = mat,
  con = con2,
  name = "dense_matrix",
  class = "dbDenseMatrix",
  overwrite = TRUE
)

# preview
dense
```

### transpose
```{r}
dbMatrix::t(dense)
```

### colMeans
```{r}
dbMatrix::colMeans(dense)
```

### colSums
```{r}
dbMatrix::colSums(dense)
```

### rowMeans
```{r}
dbMatrix::rowMeans(dense)
```

### rowSums
```{r}
dbMatrix::rowSums(dense)
```

### mean
```{r}
dbMatrix::mean(dense)
```

### dim
```{r}
dim(dense)
```

## Cleanup
```{r}
DBI::dbDisconnect(con, shutdown = TRUE)
DBI::dbDisconnect(con2, shutdown = TRUE)
options(old_options)
```

## Session Info
```{r eval=TRUE, message=FALSE}
sessionInfo()
```

---
title: "Overview"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Overview}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(dbMatrix)
```

## dbMatrix
`dbMatrix` is a core package in the [`dbverse`](https://dbverse-org.github.io/dbverse/) library that inherits from the `dbData` base class and consists of two subclasses, `dbSparseMatrix` and `dbDenseMatrix`.

`dbMatrix` objects emulate in-memory dense and sparse matrices in an embedded database powered by DuckDB. 

## Minimal example

```{r}
con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

mat <- matrix(1:9, nrow = 3, ncol = 3)
dbmat <- dbMatrix(
  value = mat,
  con = con,
  name = "overview_matrix",
  class = "dbDenseMatrix",
  overwrite = TRUE
)

dbmat

DBI::dbDisconnect(con, shutdown = TRUE)
```


## dbSparseMatrix
![](../man/figures/dbMatrix-overview.png)

Sparse matrices are represented as triplet vectors with `i`, `j`,
and `x` numerical vectors. `i` and `j` refer to the row and column indices, respectively, with specific `x` values or counts. Note: The `i` and `j` vectors are one-based following conventions from the `Matrix` R package and the `TSparseMatrix` class.

The `dbSparseMatrix` class stores the triplet vector representation with its associated dimension names and dimensions if available. If none are provided, they are automatically generated as factors (enums) with `row#` or `col#` labels for rownames and colnames, respectively. The `dbSparseMatrix` class does not contain zeros in the triplet vector representation. Zero values are inferred based on the provided dimensions.

Support for compressed sparse column matrices of the `dgCMatrix` class from the `Matrix` R package is currently implemented. For more information about sparse matrices, see [the wiki article here ](https://en.wikipedia.org/wiki/Sparse_matrix). 


## dbDenseMatrix

We use the same approach as `dbSparseMatrix` to represent a dense matrix as a
triplet `i`, `j`, and `x` vector. The only difference is that the `dbDenseMatrix` class consists of all values in the matrix, including zeros.

---
title: "Arithmetic"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Arithmetic}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
# Enable densification for vignette examples
old_options <- options(dbMatrix.allow_densify = TRUE)
```

```{r setup}
library(dbMatrix)
```


## dbMatrix arithmetic

`dbMatrix` objects support `Arith` and `Ops` operations. We will demonstrate how to perform arithmetic operations on `dbSparseMatrix` objects.

**Note:** Some operations with zero values are not yet supported with dbMatrix objects. In addition, certain arithmetic operations between `dbMatrix` objects are also not yet supported. We welcome user feedback and reporting issues on the [Github page ](https://github.com/dbverse-org/dbmatrix-r/).

### Create test data
Let's create a simple sparse matrix for demonstration:

```{r}
# Create a sparse matrix
set.seed(42)
dgc <- Matrix::rsparsematrix(100, 50, density = 0.1, rand.x = function(n) rpois(n, 5) + 1)
rownames(dgc) <- paste0("gene_", seq_len(100))
colnames(dgc) <- paste0("cell_", seq_len(50))

dplyr::glimpse(dgc)
```

The matrix contains 100 rows (genes) and 50 columns (cells). Like most single-cell RNA-seq data, the matrix is sparse.


### Create a dbMatrix object
Let's create a `dbSparseMatrix` object from the above `dgc` object.
```{r}
# Note: by default the constructor creates a dbMatrix object in-memory
con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")

dbsm <- dbMatrix(
  value = dgc,
  con = con,
  name = "test_matrix",
  class = "dbSparseMatrix",
  overwrite = TRUE
)

# preview the object
dbsm
```

### Scalar Arithmetic
`dbMatrix` emulates scalar arithmetic in the `Matrix` package. 

Note: Addition or subtraction with non-zero addends on a `dbSparseMatrix` results in a `dbDenseMatrix`.
```{r} 
dbsm + 1

dbsm * 100
```

### Matrix Arithmetic
`dbMatrix` also supports matrix arithmetic for `dbMatrix` objects that are [conformable](https://en.wikipedia.org/wiki/Conformable_matrix).
```{r}
dbsm + dbsm
```

### Matrix Multiplication

#### Hadamard product
```{r}
dbsm * dbsm
```

#### Matrix product
TODO

### Cleanup
```{r}
DBI::dbDisconnect(con, shutdown = TRUE)
options(old_options)
```

### Session Info
```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{get_tblName}
\alias{get_tblName}
\title{get_tblName}
\usage{
get_tblName(dbMatrix)
}
\arguments{
\item{dbMatrix}{A database-backed object inheriting from \code{dbData}.}
}
\value{
A character scalar giving the DuckDB table name associated with the
object.
}
\description{
get_tblName
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R, R/ingest.R
\name{as.dbMatrix}
\alias{as.dbMatrix}
\title{Convert \code{\link[Matrix:Matrix]{Matrix::Matrix}} to \code{\link{dbMatrix}}}
\usage{
as.dbMatrix(x, con = NULL, name = "dbMatrix", overwrite = FALSE, ...)

as.dbMatrix(x, con = NULL, name = "dbMatrix", overwrite = FALSE, ...)
}
\arguments{
\item{x}{Object to convert (e.g., matrix, dgCMatrix)}

\item{con}{DBI or duckdb connection object}

\item{name}{Table name to assign within database}

\item{overwrite}{Whether to overwrite if table already exists}

\item{...}{Additional arguments passed to methods}
}
\value{
A \code{\link{dbDenseMatrix}} for dense matrix inputs or a \code{\link{dbSparseMatrix}}
for sparse matrix inputs. The returned object keeps the input dimensions
and dimnames while storing matrix values in DuckDB.
}
\description{
Converts in-memory \code{\link{matrix}}, \code{\link[Matrix:dgeMatrix-class]{Matrix::dgeMatrix}}, or
\code{\link[Matrix:dgCMatrix-class]{Matrix::dgCMatrix}} into a \code{\link{dbMatrix}} object.

Generic function to convert in-memory objects to \code{dbMatrix} objects.
}
\details{
If no \code{con} is provided, a temporary in-memory database connection is created.
If no \code{name} is provided, a unique table name is generated.
}
\concept{dbMatrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R, R/dbMatrix.R
\name{dbMatrix}
\alias{dbMatrix}
\alias{dbMatrix-class}
\title{S4 virtual class for dbMatrix}
\usage{
dbMatrix(
  value,
  class = NULL,
  con = NULL,
  overwrite = FALSE,
  name = "dbMatrix",
  dims = NULL,
  dim_names = NULL,
  mtx_rowname_file_path,
  mtx_rowname_col_idx = 1,
  mtx_colname_file_path,
  mtx_colname_col_idx = 1,
  ...
)

dbMatrix(
  value,
  class = NULL,
  con = NULL,
  overwrite = FALSE,
  name = "dbMatrix",
  dims = NULL,
  dim_names = NULL,
  mtx_rowname_file_path,
  mtx_rowname_col_idx = 1,
  mtx_colname_file_path,
  mtx_colname_col_idx = 1,
  ...
)
}
\arguments{
\item{value}{data to be added to the database. See details for supported data types \code{(required)}}

\item{class}{class of the dbMatrix: \code{dbDenseMatrix} or \code{dbSparseMatrix} \code{(required)}}

\item{con}{DBI or duckdb connection object \code{(required)}}

\item{overwrite}{whether to overwrite if table already exists in database \code{(required)}}

\item{name}{table name to assign within database \code{(required, default: "dbMatrix")}}

\item{dims}{dimensions of the matrix \code{(optional: [int, int])}}

\item{dim_names}{dimension names of the matrix \code{(optional: list(enum, enum))}}

\item{mtx_rowname_file_path}{path to .mtx rowname file to be read into \code{(optional)}
database. by default, no header is assumed.}

\item{mtx_rowname_col_idx}{column index of row name file \code{(optional)}}

\item{mtx_colname_file_path}{path to .mtx colname file to be read into
database. by default, no header is assumed. \code{(optional)}}

\item{mtx_colname_col_idx}{column index of column name file \code{(optional)}}

\item{...}{additional params to pass to \code{dplyr::copy_to}}
}
\value{
\code{dbMatrix()} returns an initialized \code{\link{dbDenseMatrix}} or
\code{\link{dbSparseMatrix}} S4 object that points to matrix data stored in DuckDB.

A \code{\link{dbDenseMatrix}} or \code{\link{dbSparseMatrix}} object, depending on
\code{class}, pointing to matrix data stored in DuckDB. The object records the
matrix dimensions and dimension names.
}
\description{
Representation of sparse and dense matrices in a database. Each object
is used as a connection to a single table that exists within the database.
Inherits from \code{dbData}.

Create an S4 \code{dbMatrix} object in sparse or dense triplet vector format.
}
\details{
This function reads in data into a pre-existing DuckDB database.
Supported \code{value} data types:
\itemize{
\item \code{\link[Matrix:dgCMatrix-class]{Matrix::dgCMatrix}} In-memory sparse matrix from the \code{\link[Matrix:Matrix]{Matrix::Matrix}} package
\item \code{\link[Matrix:dgTMatrix-class]{Matrix::dgTMatrix}} In-memory triplet vector or COO matrix
\item \code{\link[Matrix:Matrix]{matrix}} In-memory dense matrix from base R
\item \code{.mtx} Path to \href{https://math.nist.gov/MatrixMarket/formats.html}{.mtx} file
\item \code{.csv} Path to .csv file
\item \code{tbl_duckdb_connection} Table in \code{\link[duckdb:duckdb]{duckdb::duckdb}} database in ijx format from
existing \code{\link{dbMatrix}} object. \code{dims} and \code{dim_names} must be
specified if \code{value} is \code{tbl_duckdb_connection}.
}
}
\section{Slots}{

\describe{
\item{\code{dim_names}}{row (1) and col (2) names}

\item{\code{dims}}{dimensions of the matrix}

\item{\code{init}}{logical. Whether the object is fully initialized}
}}

\examples{
dgc <- readRDS(system.file("extdata", "dgc.rds", package = "dbMatrix"))
con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
dbSparse <- dbMatrix(
  value = dgc,
  con = con,
  name = "sparse_matrix",
  class = "dbSparseMatrix",
  overwrite = TRUE
)
dbSparse
}
\concept{dbMatrix}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{rowMeans,dbMatrix-method}
\alias{rowMeans,dbMatrix-method}
\alias{colMeans,dbMatrix-method}
\title{Row (column) means for \code{\link{dbMatrix}} objects}
\usage{
\S4method{rowMeans}{dbMatrix}(x, na.rm = FALSE, dims = 1, ...)

\S4method{colMeans}{dbMatrix}(x, na.rm = FALSE, dims = 1, ...)
}
\arguments{
\item{x}{An NxK matrix-like object, a numeric data frame, or an array-like
object of two or more dimensions.}

\item{na.rm}{Always TRUE for \code{\link{dbMatrix}} queries. Included for compatibility
with the generic.}

\item{dims}{Always 1 for \code{\link{dbMatrix}} queries. Included for compatibility with
the generic.}

\item{...}{Additional arguments passed to specific methods.}
}
\value{
A named numeric vector containing one mean per row or column of \code{x}.
}
\description{
Calculates the mean for each row (column) of a matrix-like object.
}
\concept{summary}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{.to_db_dense}
\alias{.to_db_dense}
\title{Convert a dbSparseMatrix to dbDenseMatrix}
\usage{
.to_db_dense(x, chunk_size = NULL)
}
\arguments{
\item{x}{A \code{\link{dbSparseMatrix}} object}

\item{chunk_size}{integer. Number of columns to process per chunk during densification.
If NULL (default), the function first checks the global option \code{dbMatrix.chunk_size}.
If that is also NULL, it calculates a chunk size such that the estimated memory usage
of each chunk does not exceed \code{dbMatrix.max_mem_convert} (default 8GB).
If the total size is within the limit, a single chunk is used.}
}
\value{
A \code{\link{dbDenseMatrix}} object
}
\description{
Internal function to convert a \code{\link{dbSparseMatrix}} to
\code{\link{dbDenseMatrix}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{head,dbMatrix-method}
\alias{head,dbMatrix-method}
\alias{tail,dbMatrix-method}
\title{Return the First or Last Parts of an Object}
\usage{
\S4method{head}{dbMatrix}(x, n = 6L, ...)

\S4method{tail}{dbMatrix}(x, n = 6L, ...)
}
\arguments{
\item{x}{an object}

\item{n}{an integer vector of length up to \code{dim(x)} (or 1,
    for non-dimensioned objects).  A \code{logical} is silently coerced to
    integer.  Values specify the indices to be
    selected in the corresponding dimension (or along the length) of the
    object. A positive value of \code{n[i]} includes the first/last
    \code{n[i]} indices in that dimension, while a negative value
    excludes the last/first \code{abs(n[i])}, including all remaining
    indices. \code{NA} or non-specified values (when \code{length(n) <
      length(dim(x))}) select all indices in that dimension. Must
    contain at least one non-missing value.}

\item{...}{arguments to be passed to or from other methods.}
}
\value{
A \code{\link{dbMatrix}} object containing the first or last \code{n} rows of \code{x},
with updated dimensions and row names.
}
\description{
Returns the first or last parts of a vector, matrix, array, table, data frame
  or function.  Since \code{head()} and \code{tail()} are generic
  functions, they have been extended to other classes, including
  \code{"\link[stats]{ts}"} from \pkg{stats}.
}
\concept{matrix_props}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precompute.R
\name{precompute}
\alias{precompute}
\title{Compute a dense COO table in a database connection}
\usage{
precompute(conn, m, n, verbose = FALSE)
}
\arguments{
\item{conn}{duckdb database connection}

\item{m}{number of rows of precomputed dbMatrix table}

\item{n}{number of columns of precomputed dbMatrix table}

\item{verbose}{logical, print progress messages. default: FALSE.}
}
\value{
A \code{tbl_dbi} object referencing the newly created precomputed lookup
table in DuckDB.
}
\description{
Precomputes a COO list table in a specificied database connection in column-
major order.
This can speed up operations that involve breaking
sparsity of a \code{dbSparseMatrix},
such as in cases when performing + or - arithmetic operations.
}
\details{
The \code{m} and \code{n} parameters must exceed the
maximum row and column indices of the \code{dbMatrix} in order to be used for
densifying any \code{dbMatrix}. If these params are less than the maximum
row and column indices, a new precomputed table will be automatically
generated with the name 'precomp_mXn'.

In such cases, run this function again with a larger
\code{n_rows} and \code{num_cols}, or to manually remove the precomputed
table set \code{options(dbMatrix.precomp = NULL)} in the R console.
}
\concept{dbMatrix}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{.join_dbm_vect}
\alias{.join_dbm_vect}
\title{Join a \code{dbSparseMatrix} with a \code{dbMatrix} object}
\usage{
.join_dbm_vect(dbm, vec_matrix, op, swap_arith_order = FALSE)
}
\arguments{
\item{dbm}{A \code{dbSparseMatrix} object.}

\item{vec_matrix}{A \code{dbMatrix} object with 1D row or col.}

\item{op}{A character string representing the operation to be performed.}

\item{swap_arith_order}{order of the arguments for the operation. default: NULL}
}
\value{
A \code{\link{dbMatrix}} object containing the result of applying \code{op} between
\code{dbm} and \code{vec_matrix}.
}
\description{
Join a \code{dbSparseMatrix} with a \code{dbMatrix} object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svd.R
\name{db_svd}
\alias{db_svd}
\title{Perform Streaming SVD on a dbMatrix}
\usage{
db_svd(
  dbm,
  k = 10,
  center = TRUE,
  scale = FALSE,
  center_rows = NULL,
  memory_limit = getOption("dbMatrix.svd_memory", 8 * 1024^3),
  return_format = c("svd", "pca")
)
}
\arguments{
\item{dbm}{A dbSparseMatrix object}

\item{k}{Number of singular values to compute}

\item{center}{Logical, center rows (default TRUE)}

\item{scale}{Logical, scale rows (default FALSE)}

\item{center_rows}{Logical, center rows vs columns (default TRUE for standard PCA)}

\item{memory_limit}{Bytes for Fast Path. Default 8 GB.}

\item{return_format}{"svd" (d, u, v) or "pca" (eigenvalues, loadings, coords)}
}
\value{
List with SVD or PCA components
}
\description{
Perform Streaming SVD on a dbMatrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{Math,dbMatrix-method}
\alias{Math,dbMatrix-method}
\title{Math Operations for \code{\link{dbMatrix}} Objects}
\usage{
\S4method{Math}{dbMatrix}(x)
}
\arguments{
\item{x}{A \code{\link{dbMatrix}} object.}
}
\value{
A \code{\link{dbMatrix}} object with the mathematical operation applied to each element.
}
\description{
Implements the \code{Math} \code{\link{S4groupGeneric}} functions
for \code{\link{dbMatrix}} objects. This includes various mathematical operations such as
logarithms, exponentials, trigonometric functions, and other transformations.
}
\details{
This method provides implementations for the following Math functions:

\emph{Arithmetic and rounding}:
\itemize{
\item \code{abs()}, \code{sign()}, \code{sqrt()}, \code{ceiling()}, \code{floor()}, \code{trunc()}
}

\emph{Cumulative operations}:
\itemize{
\item \code{cummax()}, \code{cummin()}, \code{cumprod()}, \code{cumsum()}
\item \strong{Note: \code{cumprod()} is not supported}
}

\emph{Logarithmic}:
\itemize{
\item \code{log()}, \code{log10()}, \code{log2()}, \code{log1p()}
}

\strong{DuckDB Log Function Mappings}:\tabular{lll}{
   R Function \tab DuckDB Function \tab Notes \cr
   \code{log(x)} \tab \code{LN(x)} \tab Natural logarithm \cr
   \code{log10(x)} \tab \code{LOG10(x)} \tab Base-10 logarithm \cr
   \code{log2(x)} \tab \code{LOG2(x)} \tab Base-2 logarithm \cr
   \code{log1p(x)} \tab \code{LN(x + 1)} \tab log(1+x), computed as LN \cr
}


\strong{Sparsity-Preserving Log}:
For \code{dbSparseMatrix} with pending operations, \code{log(x + 1)} operations preserve sparsity
since \code{log(0 + 1) = 0}. The multiplicative component is applied first,
then the log transformation is applied to sparse values only.

\emph{Trigonometric}:
\itemize{
\item \code{cos()}, \code{sin()}, \code{tan()}, \code{acos()}, \code{asin()}, \code{atan()}
\item \code{cosh()}, \code{sinh()}, \code{tanh()}, \code{acosh()}, \code{asinh()}, \code{atanh()}
\item \code{cospi()}, \code{sinpi()}, \code{tanpi()}
\item \strong{Note: \code{acosh()} \code{asinh()} \code{atanh()} are not supported}
}

\emph{Exponential}:
\itemize{
\item \code{exp()}, \code{expm1()}
\item \strong{Note: \code{expm1()} is not supported}
}

\emph{Special functions}:
\itemize{
\item \code{gamma()}, \code{lgamma()}, \code{digamma()}, \code{trigamma()}
\item \strong{Note: \code{digamma()} \code{trigamma()} are not supported}
}

The function applies the specified mathematical operation to each element
of the \code{\link{dbMatrix}} object.
}
\examples{
mat <- matrix(1, nrow = 3, ncol = 3)
dbmat <- as.dbMatrix(mat)
log(dbmat)
sqrt(dbmat)
sin(dbmat)

}
\concept{transform}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{.eval_op_densify}
\alias{.eval_op_densify}
\title{Evaluate if a dbSparseMatrix should be densified}
\usage{
.eval_op_densify(generic_char, vec_matrix)
}
\arguments{
\item{generic_char}{A character string representing the operation to be performed.}

\item{vec_matrix}{A \code{dbMatrix} object with 1D row or col.}
}
\value{
\code{TRUE} if the operation should densify the sparse matrix before
evaluation, otherwise \code{FALSE}.
}
\description{
Evaluate if a dbSparseMatrix should be densified
}
\details{
Evaluates if a \code{dbSparseMatrix} should be
densified for \verb{[Arith]} operations and specific scalar values for
operations in the order of dbSparseMatrix, vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{coerce-matrix-dbMatrix}
\alias{coerce-matrix-dbMatrix}
\alias{coerce-dgCMatrix-dbMatrix}
\title{Coerce matrix to dbMatrix}
\value{
A database-backed matrix object. Dense inputs return a
\code{\link{dbDenseMatrix}}, while sparse \code{\link[Matrix:dgCMatrix-class]{Matrix::dgCMatrix}} inputs return a
\code{\link{dbSparseMatrix}}.
}
\description{
Coercion methods to convert in-memory \code{matrix} objects to \code{dbMatrix} objects.
Creates a new in-memory DuckDB connection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{is.na,dbMatrix-method}
\alias{is.na,dbMatrix-method}
\title{Element-wise is.na for dbMatrix}
\usage{
\S4method{is.na}{dbMatrix}(x)
}
\arguments{
\item{x}{A dbMatrix object.}
}
\value{
A dbMatrix with same dimensions, containing 1 where the original
value was NA and 0 otherwise.
}
\description{
Returns a dbMatrix with numeric values indicating NA positions (1 = NA, 0 = not NA).
}
\examples{
mat <- matrix(c(1, NA, 3, NA), nrow = 2)
dbmat <- as.dbMatrix(mat)
is.na(dbmat)

}
\concept{transform}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\docType{class}
\name{dbDenseMatrix-class}
\alias{dbDenseMatrix-class}
\alias{dbDenseMatrix}
\title{S4 Class for \code{dbDenseMatrix}}
\value{
Objects of class \code{dbDenseMatrix} store all matrix entries explicitly
in DuckDB. They are typically returned by \code{\link[=dbMatrix]{dbMatrix()}} or \code{\link[=as.dbMatrix]{as.dbMatrix()}}
for dense inputs.
}
\description{
Representation of dense matrices using an on-disk database.
Inherits from \link{dbMatrix}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{print_array}
\alias{print_array}
\title{Generate array for pretty printing of matrix values}
\usage{
print_array(
  i = NULL,
  j = NULL,
  x = NULL,
  dims,
  rownames = rep("", dims[1]),
  class = c("sparse", "dense"),
  fill = ".",
  digits = 5L
)
}
\arguments{
\item{i, j, x}{matched vectors of integers in i and j, with value in x}

\item{dims}{dimensions of the array (integer vector of 2)}

\item{fill}{fill character}

\item{digits}{default = 5. If numeric, round to this number of digits}
}
\value{
No return value. Called for its side effect of printing a formatted
matrix preview to the console.
}
\description{
Generate array for pretty printing of matrix values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{as.matrix.dbMatrix}
\alias{as.matrix.dbMatrix}
\title{Convert \code{\link{dbMatrix}} to in-memory matrix}
\usage{
\method{as.matrix}{dbMatrix}(x, ..., sparse = FALSE, names = TRUE)
}
\arguments{
\item{x}{A \code{\link{dbMatrix}} object (dbSparseMatrix or dbDenseMatrix)}

\item{...}{Additional arguments (not used)}

\item{sparse}{Logical indicating if the output should be a sparse matrix \code{default:FALSE}}

\item{names}{Logical indicating if the output should have dimnames. \code{default:FALSE}}
}
\value{
A \code{\link[Matrix:dgCMatrix-class]{Matrix::dgCMatrix}} or \code{\link{matrix}}
}
\description{
Converts a \code{\link{dbMatrix}} object into an in-memory matrix or sparse matrix.
}
\details{
This method converts a \code{\link{dbMatrix}} object into an in-memory
\code{\link[Matrix:dgCMatrix-class]{Matrix::dgCMatrix}} (sparse = TRUE) or \code{matrix()} (default, sparse = FALSE).

\strong{Warning: This function can cause memory issues for large \code{\link{dbMatrix}} objects.}

Set \code{sparse = TRUE} to convert to a sparse matrix.
Set \code{names = TRUE} to keep dimnames.
}
\concept{dbMatrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/names.R
\name{rownames.dbMatrix}
\alias{rownames.dbMatrix}
\alias{rownames<-.dbMatrix}
\alias{colnames.dbMatrix}
\alias{colnames<-.dbMatrix}
\alias{dimnames,dbMatrix-method}
\alias{dimnames<-,dbMatrix,list-method}
\title{Retrieve and Set Row (Column) Dimension Names of dbMatrix Objects}
\usage{
rownames.dbMatrix(x, do.NULL = TRUE, prefix = "row")

\method{rownames}{dbMatrix}(x) <- value

colnames.dbMatrix(x, do.NULL = TRUE, prefix = "col")

\method{colnames}{dbMatrix}(x) <- value

\S4method{dimnames}{dbMatrix}(x)

\S4method{dimnames}{dbMatrix,list}(x) <- value
}
\arguments{
\item{x}{a matrix-like \R object, with at least two dimensions for
    \code{colnames}.}

\item{do.NULL}{Not used for this method. Included for compatibility with the
generic.}

\item{prefix}{Not used for this method. Included for compatibility with the
generic.}

\item{value}{a valid value for that component of
    \code{\link[base]{dimnames}(x)}.  For a matrix or array this is either
    \code{NULL} or a character vector of non-zero length equal to the
    appropriate dimension.}
}
\value{
\code{rownames()} and \code{colnames()} return character vectors of dimension
names. \code{dimnames()} returns a length-2 list containing row and column name
vectors. The replacement forms return the modified \code{\link{dbMatrix}} object.
}
\description{
Retrieve and Set Row (Column) Dimension Names of dbMatrix Objects
}
\concept{matrix_props}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{t,dbMatrix-method}
\alias{t,dbMatrix-method}
\title{Matrix Transpose}
\usage{
\S4method{t}{dbMatrix}(x)
}
\arguments{
\item{x}{\code{\link{dbMatrix}} object}
}
\value{
\code{\link{dbMatrix}} object
}
\description{
Given a \code{\link{dbMatrix}} \code{x}, \code{t} returns the transpose of \code{x}.
}
\concept{transform}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global.R
\name{dbMatrix_options}
\alias{dbMatrix_options}
\alias{dbMatrix-options}
\title{dbMatrix Package Global Options}
\value{
No return value. This documentation page describes package options.
}
\description{
The following global options can be modified to control the
behavior of the \code{\link{dbMatrix}} package.
}
\details{
Use \code{options()} to set the below options.
}
\section{Options}{

\itemize{
\item \code{dbMatrix.digits}: integer. Number of digits to round to in the show
function of dbMatrix objects. Default is 7.
\item \code{dbMatrix.max_mem_convert}: numeric. Maximum size (in bytes) allowed for
implicit conversion of \code{dbMatrix} to in-memory matrix. Default is 8 * 1024^3 (8GB).
\item \code{dbMatrix.chunk_size}: integer. Number of columns to process per chunk during
densification (\code{.to_db_dense}). Smaller chunks reduce memory usage but may increase
execution time. Default is \code{NULL} (automatically calculated based on \code{dbMatrix.max_mem_convert}).
\item \code{dbMatrix.max_chunks}: integer. Maximum number of chunks allowed during
densification. This prevents query parser errors caused by excessive \verb{UNION ALL}
branches. Default is 10000.
\item \code{dbMatrix.verbose}: logical. If \code{TRUE} (default), prints informative messages
during implicit coercion.
\item \code{dbMatrix.precomp_db}: character. Path to an external DuckDB file containing
precomputed table(s). If set, \code{dbMatrix} will automatically attach this
database (read-only) and look for a suitable precomputed table to speed up densification.
\item \code{dbMatrix.allow_densify}: logical. If \code{FALSE} (default), automatic sparse-to-dense
conversion is disabled. This prevents unexpected disk spilling and memory issues
when operations would require densification (e.g., division by zero, scalar addition
to sparse matrix). Set to \code{TRUE} to enable on-disk dense conversion. \strong{Warning}:
Dense conversion can cause massive disk usage for large matrices.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{compute.dbMatrix}
\alias{compute.dbMatrix}
\title{Force computation of a dbMatrix}
\usage{
\method{compute}{dbMatrix}(
  x,
  name = NULL,
  temporary = TRUE,
  dimnames = TRUE,
  overwrite = FALSE,
  ...
)
}
\arguments{
\item{x}{A \code{dbMatrix} object}

\item{name}{Name of the table to create. If NULL, a random name is generated.}

\item{temporary}{Logical. If TRUE (default), create a temporary table.}

\item{dimnames}{default = TRUE. If TRUE, the rownames and colnames will be
saved in the database. This allows full reconstruction of the dbMatrix object
using \code{\link[dbProject:dbLoad]{dbProject::dbLoad()}}.}

\item{overwrite}{Logical. If TRUE, overwrite the table if it already exists.
Default is FALSE.}

\item{...}{Additional arguments passed to methods (ignored).}
}
\value{
A \code{dbMatrix} object pointing to the new table.
}
\description{
Explicitly compute a dbMatrix and save it to a table in the database.
This overrides the default \code{dplyr::compute} to use a direct \verb{CREATE TABLE AS}
statement, which is more robust for large tables in DuckDB.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{get_MM_dim}
\alias{get_MM_dim}
\title{get_MM_dim}
\usage{
get_MM_dim(mtx_file_path)
}
\arguments{
\item{mtx_file_path}{path to .mtx file to be read into database}
}
\value{
integer vector of dimensions
}
\description{
Internal function to read dimensions of a .mtx file
}
\details{
Scans for the header of an mtx file (starting with \%) and takes one more line
representing the dimensions and number of nonzero values.

Note: the header size can vary depending on the .mtx file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{get_MM_dimnames}
\alias{get_MM_dimnames}
\title{get_MM_dimnames}
\usage{
get_MM_dimnames(
  mtx_file_path,
  mtx_rowname_file_path,
  mtx_rowname_col_idx = 1,
  mtx_colname_file_path,
  mtx_colname_col_idx = 1,
  ...
)
}
\arguments{
\item{mtx_file_path}{path to .mtx file to be read into database}

\item{mtx_rowname_file_path}{path to .mtx rowname file to be read into
database. by default, no header is assumed.}

\item{mtx_rowname_col_idx}{column index of row name file}

\item{mtx_colname_file_path}{path to .mtx colname file to be read into
database. by default, no header is assumed.}

\item{mtx_colname_col_idx}{column index of column name file}

\item{...}{additional params to pass to \code{\link[data.table:fread]{data.table::fread()}}}
}
\value{
list of row and column name character vectors
}
\description{
Internal function to read row and column names of a .mtx file
}
\details{
Can be used to read row and column names from .mtx files. Note: these files
must not contain a header (colnames).

The mtx_rowname_col_idx and mtx_colname_col_idx can be used to specify the column
index of the row and column name files, respectively. By default, the first
column is used for both.

TODO: Support for reading in only rownames or colnames.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/names.R
\name{names,dbDenseMatrix-method}
\alias{names,dbDenseMatrix-method}
\title{The names of a dbMatrix Object}
\usage{
\S4method{names}{dbDenseMatrix}(x)
}
\arguments{
\item{x}{A dbMatrix object}
}
\value{
A character vector of the names of the 1D dbMatrix object (1D matrices only)
}
\description{
The names of a dbMatrix Object
}
\concept{matrix_props}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{\%in\%,dbDenseMatrix,ANY-method}
\alias{\%in\%,dbDenseMatrix,ANY-method}
\alias{\%in\%,ANY,dbDenseMatrix-method}
\alias{\%in\%,dbSparseMatrix,ANY-method}
\title{Value Matching}
\usage{
\S4method{\%in\%}{dbDenseMatrix,ANY}(x, table)

\S4method{\%in\%}{ANY,dbDenseMatrix}(x, table)

\S4method{\%in\%}{dbSparseMatrix,ANY}(x, table)
}
\arguments{
\item{x}{A dbMatrix object or any other object}

\item{table}{Any object or a dbMatrix object}
}
\value{
A logical vector of the same length as \code{x}, indicating which elements of \code{x} are in \code{table}.
}
\description{
Implements the \code{\%in\%} operator for dbMatrix objects. This operator checks if
elements from the left operand are contained in the right operand, returning
a logical vector.
}
\details{
This is a method for the standard \code{\%in\%} operator for dbMatrix objects.
It follows R's standard behavior for the \code{\%in\%} operator:
\itemize{
\item When \code{x} is a dbDenseMatrix, it returns a logical vector with the same length as the
total number of elements in the matrix.
\item When \code{table} is a dbDenseMatrix, it allows checking if elements in \code{x} are in the matrix.
\item For dbSparseMatrix objects, it throws an error to match the behavior of dgCMatrix.
}
}
\examples{
con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
mat <- matrix(1:9, nrow = 3, ncol = 3)
dbmat <- dbMatrix(
  value = mat,
  con = con,
  name = "example_matrix",
  class = "dbDenseMatrix",
  overwrite = TRUE
)

dbmat \%in\% c(1, 3, 5, 7, 9)

c(1, 3, 5, 7, 9) \%in\% dbmat
DBI::dbDisconnect(con, shutdown = TRUE)
}
\concept{transform}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{get_dbdir}
\alias{get_dbdir}
\title{get_dbdir}
\usage{
get_dbdir(dbMatrix)
}
\arguments{
\item{dbMatrix}{A database-backed object inheriting from \code{dbData}.}
}
\value{
A character scalar giving the DuckDB database directory used by the
object.
}
\description{
get_dbdir
}
\keyword{database}
\keyword{directory}
\keyword{internal}
\keyword{of}
\keyword{path}
\keyword{string}
\keyword{the}
\keyword{to}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{Summary,dbMatrix-method}
\alias{Summary,dbMatrix-method}
\title{Summary Methods for \code{\link{dbMatrix}} Objects}
\usage{
\S4method{Summary}{dbMatrix}(x, ..., na.rm = TRUE)
}
\arguments{
\item{x}{A dbMatrix object.}

\item{...}{Additional arguments (not used, but included for compatibility with the generic).}

\item{na.rm}{Logical. If TRUE, remove NA values before computation. Always set to TRUE for this implementation.}
}
\value{
The result of applying the respective summary function to the dbMatrix object.
The type of the return value depends on the specific function called.
}
\description{
Implements the \code{\link{S4groupGeneric}} group generic functions for dbMatrix objects.
}
\details{
This method provides implementations for the following \code{\link{S4groupGeneric}} functions:
\itemize{
\item \code{max()}: Maximum value
\item \code{min()}: Minimum value
\item \code{range()}: \emph{Not supported}
\item \code{prod()}: Product of all values
\item \code{sum()}: Sum of all values
\item \code{any()}: Returns TRUE if any value is TRUE
\item \code{all()}: Returns TRUE if all values are TRUE
}
}
\examples{
mat <- matrix(1, nrow = 3, ncol = 3)
dbmat <- as.dbMatrix(mat)
max(dbmat)
min(dbmat)
prod(dbmat)
sum(dbmat)
any(dbmat > 0)
all(dbmat > 0)

}
\concept{summary}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{map_ijx_dimnames}
\alias{map_ijx_dimnames}
\title{Map dimnames to i,j indices}
\usage{
map_ijx_dimnames(dbMatrix, colName_i, colName_j)
}
\arguments{
\item{dbMatrix}{dbMatrix object}

\item{colName_i}{name of column rownames to add to database}

\item{colName_j}{name of column colnames to add to database
default: 'FALSE'.'}
}
\value{
A lazy \code{tbl_dbi} with \code{i}, \code{j}, \code{x}, and the mapped row/column name
columns.
}
\description{
Map dimnames to i,j indices
}
\details{
Constructs a table in a database that contains the accompanying dimnames
for a dbMatrix. The resulting columns in the table:
\itemize{
\item i (row index)
\item colName_i (rownames),
\item j (col index)
\item j_names (colnames)
\item x (counts of i,j occcurences)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{nrow.dbMatrix}
\alias{nrow.dbMatrix}
\alias{ncol.dbMatrix}
\title{The Number of Rows/Columns of a dbMatrix Object}
\usage{
nrow.dbMatrix(x)

ncol.dbMatrix(x)
}
\arguments{
\item{x}{\code{\link{dbMatrix}} object}
}
\value{
A length-one integer giving the number of rows or columns in \code{x}.
}
\description{
\code{nrow} and \code{ncol} return the number of rows or columns present in \code{x}.
}
\concept{matrix_props}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{[,dbMatrix,dbIndex,missing,ANY-method}
\alias{[,dbMatrix,dbIndex,missing,ANY-method}
\alias{[,dbMatrix,missing,dbIndex,ANY-method}
\alias{[,dbMatrix,dbIndex,dbIndex,ANY-method}
\alias{[,dbMatrix,dbMatrix,missing,ANY-method}
\alias{[<-,dbMatrix,dbMatrix,missing,ANY-method}
\alias{[,dbMatrix,dbDenseMatrix,missing,ANY-method}
\alias{[,dbMatrix,missing,dbDenseMatrix,ANY-method}
\alias{[,dbMatrix,dbDenseMatrix,dbDenseMatrix,ANY-method}
\title{Extract or replace values in database-backed matrices}
\usage{
\S4method{[}{dbMatrix,dbIndex,missing,ANY}(x, i, j, ..., drop = TRUE)

\S4method{[}{dbMatrix,missing,dbIndex,ANY}(x, i, j, ..., drop = TRUE)

\S4method{[}{dbMatrix,dbIndex,dbIndex,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{dbMatrix,dbMatrix,missing,ANY}(x, i, j, ..., drop = TRUE)

\S4method{[}{dbMatrix,dbMatrix,missing,ANY}(x, i, j) <- value

\S4method{[}{dbMatrix,dbDenseMatrix,missing,ANY}(x, i, j, ..., drop = TRUE)

\S4method{[}{dbMatrix,missing,dbDenseMatrix,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{dbMatrix,dbDenseMatrix,dbDenseMatrix,ANY}(x, i, j, ..., drop = FALSE)
}
\arguments{
\item{x}{A \code{dbMatrix} object.}

\item{i}{Row, logical matrix, or matrix-style index.}

\item{j}{Column index.}

\item{...}{Additional arguments.}

\item{drop}{Ignored; included for matrix API compatibility.}

\item{value}{Replacement value.}
}
\value{
A subsetted or modified \code{dbMatrix}, or an extracted vector for
matrix-style indexing.
}
\description{
Methods for subsetting and replacing values in \code{dbMatrix}
objects.
}
\concept{dbMatrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{coerce-dbMatrix-matrix}
\alias{coerce-dbMatrix-matrix}
\alias{coerce-dbDenseMatrix-matrix}
\alias{coerce-dbSparseMatrix-matrix}
\title{Coerce dbMatrix to matrix}
\value{
A base R \code{\link{matrix}} containing the collected matrix values with the
same dimensions and dimnames as the source object.
}
\description{
Coercion methods to convert \code{dbMatrix} objects to in-memory \code{matrix} objects.
Respects \code{dbMatrix.max_mem_convert} option to prevent OOM errors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input_validation.R
\name{.check_value}
\alias{.check_value}
\title{Input validation for data arg}
\usage{
.check_value(value)
}
\arguments{
\item{value}{A \code{Matrix}, \code{matrix}, or \code{tbl_duckdb_connection} object}
}
\value{
No return value. Called for input validation and throws an error if
\code{value} is invalid.
}
\description{
Input validation for data arg
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{to_named_ijx_tbl}
\alias{to_named_ijx_tbl}
\title{Convert dbMatrix to named ijx table}
\usage{
to_named_ijx_tbl(
  x,
  row_col = "row_name",
  col_col = "col_name",
  compute = FALSE
)
}
\arguments{
\item{x}{A dbMatrix object (dbSparseMatrix or dbDenseMatrix)}

\item{row_col}{Name for the row-name column (default: "row_name")}

\item{col_col}{Name for the column-name column (default: "col_name")}

\item{compute}{Whether to materialize as temp table (default: FALSE)}
}
\value{
A lazy tbl with columns: row_col, col_col, x
}
\description{
Converts a \code{dbMatrix} to a lazy long table where row and column indices are
replaced by dimension names.
}
\concept{dbMatrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{to_ijx_disk}
\alias{to_ijx_disk}
\title{to_ijx_disk}
\usage{
to_ijx_disk(con, name)
}
\arguments{
\item{con}{duckdb connection}

\item{name}{name of table to convert to ijx on disk}
}
\value{
remote table in long format unpivoted from wide format matrix
}
\description{
to_ijx_disk
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{rowVars,dbDenseMatrix-method}
\alias{rowVars,dbDenseMatrix-method}
\alias{rowVars,dbSparseMatrix-method}
\alias{colVars,dbDenseMatrix-method}
\alias{colVars,dbSparseMatrix-method}
\title{Row (column) variances for \code{\link{dbMatrix}} objects}
\usage{
\S4method{rowVars}{dbDenseMatrix}(
  x,
  rows = NULL,
  cols = NULL,
  na.rm = TRUE,
  center = NULL,
  ...,
  useNames = TRUE
)

\S4method{rowVars}{dbSparseMatrix}(
  x,
  rows = NULL,
  cols = NULL,
  na.rm = TRUE,
  center = NULL,
  ...,
  useNames = TRUE
)

\S4method{colVars}{dbDenseMatrix}(
  x,
  rows = NULL,
  cols = NULL,
  na.rm = TRUE,
  center = NULL,
  ...,
  useNames = TRUE
)

\S4method{colVars}{dbSparseMatrix}(
  x,
  rows = NULL,
  cols = NULL,
  na.rm = TRUE,
  center = NULL,
  ...,
  useNames = TRUE
)
}
\arguments{
\item{x}{A \code{\link{dbMatrix}} object.}

\item{rows}{Always NULL for \code{\link{dbMatrix}} queries. Included for compatibility
with the generic.}

\item{cols}{Always NULL for \code{\link{dbMatrix}} queries. Included for compatibility
with the generic.}

\item{na.rm}{Always TRUE for \code{\link{dbMatrix}} queries. Included for compatibility
with the generic.}

\item{center}{Always NULL for \code{\link{dbMatrix}} queries. Included for compatibility
with the generic.}

\item{...}{Additional arguments (not used, but included for compatibility with the generic).}

\item{useNames}{Always TRUE for \code{\link{dbMatrix}} queries. Included for compatibility
with the generic.}
}
\value{
A named numeric vector containing one sample variance per row or
column of \code{x}.
}
\description{
Calculates the variance for each row (column) of a matrix-like object.
}
\concept{summary}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R, R/utils.R, R/writeMM.R
\name{Arith,dbMatrix,ANY-method}
\alias{Arith,dbMatrix,ANY-method}
\alias{Arith,ANY,dbMatrix-method}
\alias{Arith,dbMatrix,dbMatrix-method}
\alias{Ops,dbMatrix,ANY-method}
\alias{Ops,ANY,dbMatrix-method}
\alias{Ops,dbMatrix,dbMatrix-method}
\alias{dbLoad,DBIConnection-method}
\alias{writeMM,dbMatrix-method}
\title{Arith dbMatrix, e2}
\usage{
\S4method{Arith}{dbMatrix,ANY}(e1, e2)

\S4method{Arith}{ANY,dbMatrix}(e1, e2)

\S4method{Arith}{dbMatrix,dbMatrix}(e1, e2)

\S4method{Ops}{dbMatrix,ANY}(e1, e2)

\S4method{Ops}{ANY,dbMatrix}(e1, e2)

\S4method{Ops}{dbMatrix,dbMatrix}(e1, e2)

\S4method{dbLoad}{DBIConnection}(conn, name, class)

\S4method{writeMM}{dbMatrix}(obj, file, ...)
}
\arguments{
\item{e1}{First operand.}

\item{e2}{Second operand.}

\item{conn}{DBIConnection object}

\item{name}{valid name value (character)}

\item{class}{character, class of the dbMatrix object (e.g. "dbDenseMatrix" or "dbSparseMatrix")}

\item{obj}{dbMatrix object}

\item{file}{path to file}

\item{...}{additional arguments}
}
\value{
\itemize{
\item Arithmetic and logical group methods return a \code{\link{dbMatrix}} object of
the appropriate dense or sparse subclass, with the same dimensions as the
input and transformed values stored in DuckDB.
\item \code{dbLoad()} returns a \code{\link{dbDenseMatrix}} or \code{\link{dbSparseMatrix}} pointing
to an existing DuckDB table.
\item \code{writeMM()} writes a Matrix Market file to \code{file} and returns
\code{invisible(TRUE)} on success.
}
}
\description{
See \code{methods::Arith} for more details.

See \code{methods::Arith} for more details.

See \code{methods::Arith} for more details.

See \code{methods::Ops} for more details.

See \code{methods::Ops} for more details.

See \code{methods::Ops} for more details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{length,dbMatrix-method}
\alias{length,dbMatrix-method}
\title{Length of a \code{\link{dbMatrix}} Object}
\usage{
\S4method{length}{dbMatrix}(x)
}
\arguments{
\item{x}{\code{\link{dbMatrix}} object}
}
\value{
A length-one integer giving the number of stored elements in \code{x}.
}
\description{
Get or set the length of vectors (including lists) and factors,
and of any other R object for which a method has been defined.
}
\concept{matrix_props}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbMatrix.R
\name{dbMatrix_from_tbl}
\alias{dbMatrix_from_tbl}
\title{dbMatrix_from_tbl}
\usage{
dbMatrix_from_tbl(
  tbl,
  rownames_colName,
  colnames_colName,
  value_colName = NULL,
  name = "dbMatrix",
  overwrite = FALSE,
  row_names = NULL,
  col_names = NULL,
  i_col = NULL,
  j_col = NULL
)
}
\arguments{
\item{tbl}{\code{tbl_duckdb_connection} table in DuckDB database in long format}

\item{rownames_colName}{\code{character} column name of rownames in tbl \code{(required)}}

\item{colnames_colName}{\code{character} column name of colnames in tbl \code{(required)}}

\item{value_colName}{\code{character} column name containing pre-aggregated integer counts.
If \code{NULL} (default), counts occurrences of each row-column pair. \code{(optional)}}

\item{name}{table name to assign within database \code{(required, default: "dbMatrix")}}

\item{overwrite}{whether to overwrite if table already exists in database \code{(required)}}

\item{row_names}{\code{character} vector of pre-computed row names (sorted). If \code{NULL}
(default), row names are extracted from the table. \code{(optional)}}

\item{col_names}{\code{character} vector of pre-computed column names (sorted). If \code{NULL}
(default), column names are extracted from the table. \code{(optional)}}

\item{i_col}{\code{character} column name containing pre-computed row indices (1-based integers).
If provided with \code{j_col}, skips index encoding for optimal performance. \code{(optional)}}

\item{j_col}{\code{character} column name containing pre-computed column indices (1-based integers).
If provided with \code{i_col}, skips index encoding for optimal performance. \code{(optional)}}
}
\value{
\code{dbMatrix} object
}
\description{
Constructs a \code{dbSparseMatrix} object from a \code{tbl_duckdb_connection} object.
}
\details{
The \code{tbl_duckdb_connection} object must contain dimension names as columns in long format.

If \code{value_colName} is provided, the function uses pre-aggregated counts from that column.
This is useful when the input table already contains aggregated counts (e.g., from a GROUP BY + SUM operation).
If \code{value_colName} is \code{NULL} (default), the function counts occurrences of each row-column pair.

When \code{row_names} and/or \code{col_names} are provided, the function uses these directly
instead of querying distinct values from the table. This can significantly improve performance
when the input table is a complex lazy query (e.g., result of spatial joins).

When \code{i_col} and \code{j_col} are provided, the function uses these pre-computed integer
indices directly, skipping expensive string-to-index encoding. This is the fastest path.
}
\concept{dbMatrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate.R
\name{sim_duckdb}
\alias{sim_duckdb}
\alias{sim_dgc}
\alias{sim_denseMat}
\alias{sim_ijx_matrix}
\alias{sim_dbSparseMatrix}
\alias{sim_dbDenseMatrix}
\title{sim_dgc}
\usage{
sim_duckdb(value = datasets::iris, name = "test", con = NULL, memory = TRUE)

sim_dgc(num_rows = 50, num_cols = 50, n_vals = 50)

sim_denseMat(num_rows = 50, num_cols = 50)

sim_ijx_matrix(mat_type = NULL, num_rows = 50, num_cols = 50, seed_num = 42)

sim_dbSparseMatrix(
  num_rows = 50,
  num_cols = 50,
  seed_num = 42,
  name = "sparse_test",
  memory = FALSE
)

sim_dbDenseMatrix(
  num_rows = 50,
  num_cols = 50,
  seed_num = 42,
  name = "dense_test",
  memory = FALSE
)
}
\arguments{
\item{num_rows}{The number of rows in the matrix (default: 50)}

\item{num_cols}{The number of columns in the matrix (default: 50)}

\item{seed_num}{The seed number for reproducibility (default: 42)}
}
\value{
A dgCMatrix object
}
\description{
Simulate a dbSparseMatrix in memory

Simulate a dbDenseMatrix in memory.
}
\details{
This function generates a simulated sparse matrix (dgCMatrix) with number
of rows and columns and sets n_vals random values to a non-zero value.

This function generates a simulated dense matrix object with a specified number
of rows and columns.

This function generates an ijx representation of a simulated dgCMatrix object
with a specified number of rows and columns and sets 50 random values to
a non-zero value.
}
\section{Functions}{
\itemize{
\item \code{sim_duckdb()}: Simulate a duckdb connection dplyr tbl_Pool in memory

\item \code{sim_dgc()}: Simulate a dgcMatrix

\item \code{sim_denseMat()}: Simulate a dense matrix

\item \code{sim_ijx_matrix()}: Simulate a duckdb connection dplyr tbl_Pool in memory

\item \code{sim_dbSparseMatrix()}: Simulate a dbSparseMatrix in memory

\item \code{sim_dbDenseMatrix()}: Simulate a dbDenseMatrix in memory

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{dim,dbMatrix-method}
\alias{dim,dbMatrix-method}
\title{Dimensions of an Object}
\usage{
\S4method{dim}{dbMatrix}(x)
}
\arguments{
\item{x}{\code{\link{dbMatrix}} object}
}
\value{
An integer vector of length 2 giving the number of rows and columns
in \code{x}.
}
\description{
Retrieve the dimension of an object.
}
\concept{matrix_props}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{get_con}
\alias{get_con}
\title{get_con}
\usage{
get_con(dbMatrix)
}
\arguments{
\item{dbMatrix}{A database-backed object inheriting from \code{dbData}.}
}
\value{
A live DBI connection associated with the database-backed object.
}
\description{
get_con
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{coerce-dbMatrix-dgCMatrix}
\alias{coerce-dbMatrix-dgCMatrix}
\alias{coerce-dbDenseMatrix-dgCMatrix}
\alias{coerce-dbSparseMatrix-dgCMatrix}
\title{Coerce dbMatrix to dgCMatrix}
\value{
A \code{\link[Matrix:dgCMatrix-class]{Matrix::dgCMatrix}} object containing the collected matrix
values. Dense inputs are converted to sparse Matrix format after collection.
}
\description{
Coercion methods to convert \code{dbMatrix} objects to in-memory \code{dgCMatrix} objects.
Respects \code{dbMatrix.max_mem_convert} option to prevent OOM errors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operations.R
\name{mean,dbDenseMatrix-method}
\alias{mean,dbDenseMatrix-method}
\alias{mean,dbSparseMatrix-method}
\title{Arithmetic Mean for \code{\link{dbMatrix}} objects}
\usage{
\S4method{mean}{dbDenseMatrix}(x, ...)

\S4method{mean}{dbSparseMatrix}(x, ...)
}
\arguments{
\item{x}{\code{\link{dbMatrix}} object}

\item{...}{further arguments passed to or from other methods.}
}
\value{
A length-one numeric vector giving the arithmetic mean of all entries
in \code{x}.
}
\description{
Generic function for the (trimmed) arithmetic mean.
}
\concept{summary}
